/**
 * @file Command.h
 * @brief Command handling - Menu and shortcuts (declarations)
 * 
 * @author Ferenc Moricz
 * $Revision: 1.11 $
 * $Date: 2003/09/28 23:35:39 $
 */

#ifndef COMMAND
#define COMMAND

#include "Debug.h"

#include <map>
#include <vector>
using namespace std;

#include "Preferences.h"

#include <qevent.h>
#include <qintdict.h>
#include <qlabel.h>
#include <qlistview.h>
#include <qpopupmenu.h>

class CommandMenuItem
{                                                                              
  public:
	CommandMenuItem(QPopupMenu *pParent, int iMenuID, QString *sCommand);
	QPopupMenu *pParent;
	int iMenuID;
	QString *sCommand;
};

class Command : public QObject
{
Q_OBJECT
friend class Preferences;
  public:
	Command() { pStyleListMenu = NULL; }
	bool loadShortCuts( const char * = NULL, QDict<int> * = NULL );
	QPopupMenu *loadMenu( const char * = NULL, const char * = NULL, QDict<int> * = NULL );
	char *command(const char *comstr);
	bool shortEvent( QKeyEvent *e );
	QPopupMenu *pStyleListMenu;

	// i18n - automatic update of the labels' text during language change
	void addLabel(QLabel *pLabel, QString *pText);
	void updateLabels();
	void updateMenuItems();

  public slots:
	void menuEvent( int );

  protected:
	vector<QLabel *> oLabels;
	vector<QString *> oLabelTexts;
	QIntDict<QString> shortcut_command;
	map<int, CommandMenuItem *> menu_command;

	EditProperties shortcut;
	EditProperties menu;
};

#endif

