/*
 * EditBuffer.h
 *
 * The main editor widget's includes
 * 
 * Author: Ferenc Moricz
 * $Revision: 1.50 $
 * $Date: 2004/01/03 04:11:59 $
 *
 */
 
#ifndef EDITBUFFER
#define EDITBUFFER

#include "Debug.h"

#include <qlabel.h>
#include <qfile.h>
#include <qfileinfo.h>
#include <qvbox.h>
#include <qscrollbar.h>
#include <qstatusbar.h>
#include <qtimer.h>
#include <qlineedit.h>
#include <qcombobox.h>

#include <math.h>
//#ifndef max
//#define max(a, b) ( (a) > (b) ? (a) : (b) )
//#define min(a, b) ( (a) < (b) ? (a) : (b) )
//#endif

#include "EditBufferText.h"
#include "UndoRedo.h"
#include "Debug.h"

extern char FREDDYVERSION[];
class ModuleLoader;

class EditSession;

class EditBufferLineBar : public QFrame
{
  public:
	EditBufferLineBar( QWidget *parent, EditBuffer *buffer );
	void reconfigure();
	virtual QSize sizeHint () const;

  protected:

	QFont oNumberFont;
	QFontMetrics oNumberFontMt;
	bool dbuf;
	static QPixmap *pm;
	int fontascent;
	int iBarWidth;

	virtual void paintEvent( QPaintEvent *pe );
	virtual void resizeEvent( QResizeEvent * );
	EditBuffer *buffer;
};

class EditBuffer : public QVBox
{ 
Q_OBJECT
friend class EditBufferText;
  public:
	UndoRedo oUndo;

	EditBuffer( QWidget * =0, const QString * =0 );
	const QString *name;

	void loadFile(QFile &, int iIsUTF8=-1);
	void saveFile();
	char *command( const char *command, const char *arg1=NULL,
				   const char *arg2=NULL, const char *arg3=NULL );
	void cmd_scrollBar( void );
	void cmd_scrollTo( int, int = -1, bool = true );
	void cmd_scrollToCursor( int = -1, bool = true );
	void cmd_setCursor( int, int );
	void cmd_selection_copy( QString *pCopiedString = NULL );
	void cmd_selection_paste( bool bClipBoard = TRUE );
	void cmd_selection_clear( void );
	void cmd_selection_wrap( const char *sType = "left" );
	void cmd_find( bool bOpen = true );
	void cmd_checkFile( void );
	void cmd_gotoLine( void );

	void cmd_undo(bool bMicro = false);
	void cmd_redo(bool bMicro = false);

	void atomic_insertString(QString *sString, int iColumn, int iLine);
	void atomic_deleteString(int iColumn, int iLine, int iLength);
	void atomic_createLine(int iLine);
	void atomic_deleteLine(int iLine);
	void atomic_moveCursor(int iColumn, int iLine);
	void atomic_scroll(int iColumn, int iLine);
	void atomic_setSelection(QString &sSelectionName, int iColumn1, int iLine1, int iColumn2, int iLine2);

	inline QFile * getFile( void ) { return file; }
	inline void setFile(QFile *file) { if (this->file) delete(this->file); this->file = file; }

	bool bSyntaxHighlighting;

	static class ModuleLoader *module;
	class EditBufferText *text;
	vector<QWidget *, allocator<QWidget *> > childs;
	QScrollBar *hscroll, *vscroll;
	QHBox *bufferview;
	QVBox *bufferview_v;
	QWidget *findbox;
	EditBufferLineBar *linebar;
	int tabsize;

	enum FileType { FILETYPE_DOS, FILETYPE_UNIX } eFileType;
	bool bUTF8, bLoading;
	QComboBox *pFileTypeCombo, *pTabCombo, *pUTFCombo;

	QLineEdit *pCommandTool;

	void setModified( bool bModified );
	inline bool isModified() { return bFileModified; }
	void reconfigure();

  protected:
	QFile *file;
	QDateTime file_lastmodified;
	bool bFileModified;
	int hscrollValue;
	EditSession *pSession;

	QHBox *position;
	QLabel *position_x, *position_y;
	QTimer scrolltimer;
	QStatusBar *statusbar;
	bool event( QEvent *e );

  public slots:
	virtual void scrollEvent( int );
	virtual void scrollRelease( void );
	virtual void hscrollEvent( int );
	virtual void commandToolEnter( );
};

#endif

