/*
 * EditBufferText.h
 *
 * The lines' holder
 * 
 * Author: Ferenc Moricz
 * $Revision: 1.36 $
 * $Date: 2004/01/03 04:11:59 $
 *
 */
 
#ifndef EDITBUFFERTEXT
#define EDITBUFFERTEXT

#include "EditLine.h"
#include "EditProperties.h"
#include "EditLineSyntax.h"
#include <qwidget.h>
#include <qpixmap.h>

#include <vector>
using namespace std;

class EditBuffer;
class Editor;

class FreddyCursor : public QPoint
{
  public:
	void setY( int iY );
	inline FreddyCursor &FreddyCursor::operator =(QPoint &oPoint)
	{ (*(QPoint *)this) = oPoint; return *this; }
};

class EditBufferText : public QWidget
{  
Q_OBJECT
friend class EditBuffer;
friend class Editor;
friend class EditBufferFind;
friend class EditBufferLineBar;
  public:
	EditBufferText( EditBuffer *, QWidget * );
	void setSelecting(bool bSelecting);

	EditBuffer *up;
	static EditProperties *properties;
	 
	vector<EditLine *, allocator<EditLine *> > lines;
	int firstVisible, nLines, nPhysLines;
	int fontwidth, fontheight, fontascent;
	FreddyCursor cursor;

	/* selections */
	QPoint select[2];
	static QString sSelectionColorName, sSelectionCopiedColorName, sSelectionPastedColorName;
	QString *sCurrentSelectionColorName;
	int selectsmaller;
	bool selecting, mouseselecting, cursorpainted;
	const QColor *selectcolor;
	static EditLineSyntax *syntax;
	int iMaxLineLength;

	const QColor *textpen, *tabpen, *keypen, *commentpen, *stringpen;
	const QColor *cursorpen, *preprocpen, *bg, *textbg;
	const QString *style;
	bool dbuf;
	QPixmap pm;

	void reconfigure();

  protected:
	virtual bool event( QEvent * );
	int repaintfrom, repaintto;
	virtual void paintEvent( QPaintEvent * );
	void displayText(QPainter &p, QString &sText, int iBegin, int iLength);

	void scroll( int );
	void paintCursor( bool = true );
	int iMouseClick;
	QTimer *mousetimer, *pMouseClickTimer;
	QWidget *pAskedForFocus;

	void askForFocus(QWidget *pWidget);
	void cmd_text_insertchar( const QString &, bool bWrap = false );
	void cmd_text_filltocursor( void );

	void *pFilterID;
	bool bUseColorer;

	virtual void resizeEvent( QResizeEvent * );

	virtual void mousePressEvent( QMouseEvent * );
	virtual void mouseReleaseEvent( QMouseEvent * );
	virtual void mouseMoveEvent( QMouseEvent * );
	virtual void mouseDoubleClickEvent( QMouseEvent * );
#if QT_VERSION > 280
	virtual void contextMenuEvent( QContextMenuEvent * );
#endif

	virtual void dragEnterEvent( QDragEnterEvent * );
	virtual void dropEvent( QDropEvent * );

  public slots:
	virtual void keyPressEvent( QKeyEvent * );
	virtual void focusInEvent( QFocusEvent * );
	virtual void focusOutEvent( QFocusEvent * ); 
	virtual void mouseTimeout( );
	virtual void mouseClickTimeout( );
};

#endif
