/**
 * @file EditProperties.h
 * @brief Property management system, declarations
 * 
 * @author Ferenc Moricz
 * $Revision: 1.22 $
 * $Date: 2004/02/22 00:12:48 $
 *
 */

#ifndef EDITPROPERTIES
#define EDITPROPERTIES

#include "Debug.h"

#include <iostream>
#include <map>
#include <vector>
using namespace std;

#include <qstring.h>

#include <qcheckbox.h>
#include <qcolor.h>
#include <qdict.h> 
#include <qfont.h>
#include <qwidget.h>

class PropertyItem
{
friend class EditProperties;
friend class Editor;

  public:
	PropertyItem( const char * );
	void setValue( const char * );

	QString *getString() { return string; }
	QString *getComment() { return pComment; }

	void setFontValue();

  protected:
	QString *pComment;
	QString *string;
	int *number;
	bool *logical;
	QColor *color;
	QFont *pFont;
	bool bIsDefault;
};

/*template <class _Tp>
struct less_qstring : public binary_function<_Tp,_Tp,bool>
{
	bool operator()(const _Tp& __x, const _Tp& __y)
		const { return (__x.compare(__y) < 0 ? -1 : (__x.compare(__y) == 0 ? 0 : 1));}
};
*/

typedef vector<PropertyItem, allocator<PropertyItem> > tPropVect;

class EditProperties : public map< QString, tPropVect, less<QString> >
{                                                                                                                                      
  public:
	static EditProperties *pMain;

	EditProperties();

	bool loadConfig( const char * = NULL, bool bReload = false );
	bool saveConfig( const char *sFileName = NULL, vector<QString> *pItems = NULL,
	                 bool bFlush = FALSE );

	void toFirst( QString );
	bool hasMore( QString );

	const QString *getComment( QString );

	vector<PropertyItem> *getPropertyList(QString sPropertyName);

	/* general data (handled as string) */
	const QString *getValue( QString, 
						const QString * = NULL, bool bCreate = FALSE );
	bool setValue( QString, const QString * = NULL, bool = FALSE );

	/* numerical value */
	int getNumericalValue( QString, int = 0, bool bCreate = FALSE );
	bool setNumericalValue( QString, int = 0, bool = FALSE );

	/* boolean value */
	bool setBoolValue( QString, bool = FALSE, bool = FALSE );
	bool getBoolValue( QString, bool = FALSE, bool bCreate = FALSE );

	/* color value */
	const QColor *getColorValue( QString, 
						const QColor * = NULL, bool bCreate = FALSE );
	bool setColorValue( QString, const QColor * = NULL, bool = FALSE );

	/* font value */
	const QFont *getFontValue( QString, 
						const QFont * = NULL, bool bCreate = FALSE );
	bool setFontValue( QString &key, const QFont *value = NULL, bool next = FALSE);
	const QFont *loadFontValue( QString &sConfigString );

  protected:
	PropertyItem *getProperty( QString &oKey, bool bCreate );
	void token(char *, int, int &, int &);
	QString *config;
};

#endif

