/*
 * Editor.h
 *
 * Editor's main program's includes
 * 
 * Author: Ferenc Moricz
 * $Revision: 1.50 $
 * $Date: 2004/02/22 00:12:48 $
 *
 */
 
#ifndef EDITOR
#define EDITOR

#ifdef _SYSTEM_WINDOWS_
extern char *INSTALL_ETC;
extern char *INSTALL_SHARE;
extern char *INSTALL_DOC;
#endif

// option: using a colorer filter or using the integrated syntax highlighter
// #define COLORER_FILTER

extern char FreddyVersion[];

#include "Debug.h"
        
#include "EditSession.h"
#include "EditProperties.h"
#include "Module.h"
#include "Command.h"

#include <qapplication.h>
#include <qdir.h>
#include <qmainwindow.h>
#include <qsplitter.h>
#include <qstyle.h>
#include <qvbox.h>
#include <qwidgetstack.h>

class Command;
class ModuleLoader;

/* Common types and templates */

#include <functional>
#include <qstring.h>

/* Editor - socket communication between instances */

class EditorMainWindow : public QMainWindow
{
  public:
	EditorMainWindow();
  protected:
	virtual void closeEvent(QCloseEvent *e) { QApplication::exit(0); }
};

class Editor : public QObject
{
friend class Command;
friend class Preferences;
friend class EditorServer;
friend class EditorSocket;
friend class EditProperties;

Q_OBJECT
  public:
	Editor(int argc, char *argv[], QApplication *);

	void configGetDirectories(void);
	enum WhichConfigFile{ FREDDY, SHORTCUTS, MENU, COMMANDLINE, SAVEDSTATE };
	void configLoadFile(WhichConfigFile, bool bReload = false);

	void loadState();
	void saveState();
	void saveModifiedFiles();

	void showCommandLineHelp(void);

	QStringList *processCommandLine(int argc, char *argv[]);

	EditBuffer *loadFile(QFile *file, bool = false);

	EditSession *newSession( const QString *name = NULL);
	void closeSession( EditSession * = NULL);
	void renameSession( QString * = NULL, EditSession * = NULL);
	void checkActualSession( void );

	char *command( const char *command, const char *arg1=NULL,
	               const char *arg2=NULL, const char *arg3=NULL );

	static Editor *editor;
	static EditorMainWindow *pMainWindow;
	Command *com;
	ModuleLoader *moduleloader;

	QLineEdit *pCommandTool;

	void startUp(void);
	void reconfigure(void);

	QFrame *getContainer() { return container; };
	QWidgetStack *getStack() { return stack; };

	QPopupMenu *pContextMenu, *pTabMenu;
	bool testmode, bSavedStateOpened;

	QStringList *pCommandFileList;

  public slots:
	void newFile();
	void openFile(QStringList *);
	void saveFile();
	void saveasFile(const char * = NULL);
	void closeFile();

	void openCommandlineFile(QStringList *);

	void sessionMenu(int);

  protected:
	QWidgetStack *stack;
	QFrame *container;

	Preferences *pPreferences;

	QPopupMenu *sessionmenu;
	int sessionmenuindex, sessionid;
	vector<QWidget*, allocator<QWidget *> > sessions;

	QApplication *pApplication;
	QTranslator *pTranslator;

	class EditorServer *pEditorServer;
	class EditorSocket *pEditorClient;

	QString sMenuPath;
	QString style;

	QDir configdir_user, configdir_default;
	EditProperties savedstate;  /// the state when previous session finished
	EditProperties oPropCmdline;
};

#include <qserversocket.h>
#include <qsocket.h>

/* Editor - socket communication between instances */
class EditorSocket : public QSocket
{
Q_OBJECT
  public:
	int iState;

	EditorSocket();

  public slots:
	void slotConnected( void );
	void slotClosed( void );
	void slotRead( void );
	void slotWrote( int );
	void slotError( int );
};

class EditorServer : public QServerSocket
{
Q_OBJECT
  public:
	EditorServer( int );

	static QSocket *locateServer( int );

	virtual void newConnection( int socket );
};

#endif

