/*
 * Modules.h
 *
 * Loadable module support for Freddy
 * 
 * Author: Ferenc Moricz
 * $Revision: 1.12 $
 * $Date: 2003/10/21 02:24:12 $
 *
 */

#ifndef FREDDY_MODULE
#define FREDDY_MODULE


#include <vector>
using namespace std;

#include "Debug.h"

#include <qfile.h>
#include <qstring.h>

class Module;

#ifdef _SYSTEM_LINUX_
#define MODULES_API 
#else
#include <windows.h>
#define MODULES_API __declspec(dllexport)
#endif

typedef MODULES_API Module *(ModuleInit)(Module *);
typedef ModuleInit *pModuleInit;
/* defaults for unused params: qsParam1: NULL, qsParam2: NULL, iParam1:-1, iParam2: -1 */
typedef MODULES_API QString *(ModuleFunction)(Module *pThis, const QString &qsName,
                  QString *qsParam1 = NULL, QString *qsParam2 = NULL, int iParam1 = -1, int iParam2 = -1);
typedef ModuleFunction *pModuleFunction;
/* defaults for unused params: sArg1: NULL, sArg2: NULL, pArg: NULL */
typedef void *(ModuleQuery)( const char *sCommand, const char *sArg1 = NULL,
                  const char *sArg2 = NULL, const void *pArg = NULL);
typedef ModuleQuery *pModuleQuery;

class Module
{
  public:
	QString *qsName;
	vector<QString, allocator<QString> > functions;

	pModuleInit pInit;
	pModuleFunction pFunction;
	pModuleQuery pQuery;

#ifdef _SYSTEM_WINDOWS_
	HINSTANCE iModuleInstance;
#endif
};

class ModuleLoader
{
  public:
	ModuleLoader();

	void init(void);

	void loadModules(const char *sFile, int iLevel);
	void broadcastEvent(const QString &eventName,
                  QString *qsParam1 = NULL, QString *qsParam2 = NULL,
                  int iParam1 = -1, int iParam2 = -1);

	char *command( const char *plugin, const char *command, const char *arg1=NULL,
				   const char *arg2=NULL, const char *arg3=NULL );

	static void *moduleQuery( const char *sCommand, const char *sArg1=NULL,
                  const char *sArg2=NULL, const void *pArg=NULL );

	vector<Module *, allocator< Module * > > modules;
	static Module *pFilterModule;
};

#endif
