/**
 * @file Preferences.h
 * @brief Preferences window and tools
 * 
 * @author Ferenc Moricz
 * $Revision: 1.6 $
 * $Date: 2003/08/03 11:21:27 $
 *
 */

#ifndef PREFERENCES
#define PREFERENCES

#include "EditProperties.h"

#include <vector>
using namespace std;

#include <qcheckbox.h>
#include <qcombobox.h>
#include <qhbox.h>
#include <qlistview.h>
#include <qpushbutton.h>
#include <qtextbrowser.h>
#include <qwidgetstack.h>

class Command;

class PreferenceItem : public QObject
{
Q_OBJECT
  public:
	PreferenceItem();

	enum { CHECK, TEXT, SELECT, COLOR, FONT } eType;

	QString *sProperty;
	bool bChanged, bChangedApplied;

	QPushButton *pButton;

	QColor *pColor;
	QCheckBox *pCheck;
	QLineEdit *pLineEdit;
	QFont *pFont;
	QComboBox *pSelectBox;

  public slots:
	virtual void slotItem();
	virtual void textChanged(const QString &sNewText);
	virtual void stateChanged(int iState);
};

class PreferenceItemContainer : public QHBox
{
  public:
	PreferenceItemContainer(QWidget *pParent);
	void filter( QWidget *pWidget );
	QString sProperty, *sComment;
	class Preferences *pPreferences;
	QWidget *pActiveWidget;
  protected:
	virtual bool event(QEvent *e);
	bool eventFilter( QObject *, QEvent * );
};

class PreferencePage
{
  public:
	void revert(void);

	QWidget *pPageWidget;
	QString *sComment;
	QListViewItem *pPageTreeItem;
	vector<PreferenceItem *> oItems;
};

class Preferences : public QObject
{
Q_OBJECT
friend class PreferenceItemContainer;
  public:
	Preferences( void );
	void resetForSave( void );
	static void drawPixMap(QPushButton *pButton, QColor *pColor);
	void show(void);
	void close(void);

  protected:
	QDialog *pPreferences;
	QDialog *pPrefDialogWindow;
	QHBox *pPrefDialog;
	QWidget *pSplitter;
	QWidgetStack *pPages;
	map<QString, PreferencePage *, less<QString> > oPrefPages;
	QListView * pCategoryTree;
	vector<QString> oChangedItems;
	QTextBrowser *pHelpWindow;

  public slots:
	virtual void slotPrefOK( void );
	virtual void slotPrefApply( void );
	virtual void slotPrefSaveConfig( void );
	virtual void slotPrefCancel( void );
	virtual void slotPrefHelp( bool bCheck );
	virtual void slotTreeClicked( QListViewItem *pItem );
};

#endif

