#ifndef QEXTTABBAR_H
#define QEXTTABBAR_H

#ifndef QT_H
#include "qwidget.h"
#include "qptrlist.h"
#endif // QT_H

class ExtQTabBar;
class QIconSet;

class ExtQTab : public Qt
{
    friend class ExtQTabBar;
    friend class ExtQTabWidget;

public:
    ExtQTab();
    virtual ~ExtQTab();
    ExtQTab( const QString& text );
    ExtQTab( const QIconSet& icon, const QString& text = QString::null );

    void setText( const QString& text);
    QString text() const { return label; }
    void setIconSet( const QIconSet& icon );
    QIconSet* iconSet() const { return iconset; }
    void setRect( const QRect& rect ) { r = rect; }
    QRect rect() const { return r; }
    void setEnabled( bool enable ) { enabled = enable; }
    bool isEnabled() const { return enabled; }
    void setIdentifier( int i ) { id = i; }
    int identifier() const { return id; }

private:
    void setTabBar( ExtQTabBar *tb );
    QString label;
    QRect r; // the bounding rectangle of this (may overlap with others)
    bool enabled;
    int id;
    QIconSet* iconset; // optional iconset
    ExtQTabBar *tb;
};


struct ExtQTabPrivate;
//class *QAccel;

#include <qpopupmenu.h>
class ExtQTabBar: public QWidget
{
    Q_OBJECT
    Q_ENUMS( Shape )
    Q_PROPERTY( Shape shape READ shape WRITE setShape )
    Q_PROPERTY( int currentTab READ currentTab WRITE setCurrentTab )
    Q_PROPERTY( int count READ count )
    Q_PROPERTY( int keyboardFocusTab READ keyboardFocusTab )

public:
    ExtQTabBar( QWidget* parent=0, const char* name=0 );
    ~ExtQTabBar();

    enum Shape { RoundedAbove, RoundedBelow,
		 TriangularAbove, TriangularBelow };

    Shape shape() const;
    virtual void setShape( Shape );

    void show();

    virtual int addTab( ExtQTab * );
    virtual int insertTab( ExtQTab *, int index = -1 );
    virtual void removeTab( ExtQTab * );

    virtual void moveTab( ExtQTab *, int index = -1 );     /* < EXT > */

    virtual void setTabEnabled( int, bool );
    bool isTabEnabled( int ) const;


    QSize sizeHint() const;
    QSize minimumSizeHint() const;

    int currentTab() const;
    int keyboardFocusTab() const;

    ExtQTab * tab( int ) const;
    ExtQTab * tabAt( int ) const;
    int indexOf( int ) const;
    int count() const;

    virtual void layoutTabs();
    virtual ExtQTab * selectTab( const QPoint & p ) const;

    void 	removeToolTip( int index );
    void     	setToolTip( int index, const QString & tip );
    QString 	toolTip( int index ) const;

    static QPopupMenu *pTabMenu;

public slots:
    virtual void setCurrentTab( int );
    virtual void setCurrentTab( ExtQTab * );

signals:
    void  selected( int );
    void tabOrderChanged();

protected:
    virtual void paint( QPainter *, ExtQTab *, bool ) const; // ### not const
    virtual void paintLabel( QPainter*, const QRect&, ExtQTab*, bool ) const;

    void focusInEvent( QFocusEvent *e );
    void focusOutEvent( QFocusEvent *e );

    void resizeEvent( QResizeEvent * );
    void paintEvent( QPaintEvent * );
    void mousePressEvent ( QMouseEvent * );
    void mouseMoveEvent ( QMouseEvent * );
    virtual void mouseReleaseEvent ( QMouseEvent * );
    void contextMenuEvent( QContextMenuEvent * );
    void keyPressEvent( QKeyEvent * );
    void styleChange( QStyle& );
    void fontChange ( const QFont & );

    virtual void dragMoveEvent( QDragMoveEvent * );
    virtual void dropEvent( QDropEvent * );
    
    QPtrList<ExtQTab> * tabList();

private slots:
    void scrollTabs();

private:
    QPtrList<ExtQTab> * l;
    QPtrList<ExtQTab> * lstatic;
    void makeVisible( ExtQTab* t  );
    void updateArrowButtons();
    ExtQTabPrivate * d;

    bool selectreleased;                /* < EXT > */

    friend class ExtQTabBarToolTip;
    friend class ExtQTab;

private:	// Disabled copy constructor and operator=
#if defined(Q_DISABLE_COPY)
    ExtQTabBar( const ExtQTabBar & );
    ExtQTabBar& operator=( const ExtQTabBar & );
#endif
};

#endif // EXTQTABBAR_H

#ifndef QTABWIDGET_H
#define QTABWIDGET_H

#ifndef QT_H
#include "qwidget.h"
#include "qiconset.h"
#endif // QT_H

class ExtQTabBar;
class ExtQTabWidgetData;
class ExtQTab;
class QWidgetStack;

class ExtQTabWidget : public QWidget
{
    Q_OBJECT
    Q_ENUMS( TabPosition )
    Q_ENUMS( TabShape )
    Q_PROPERTY( TabPosition tabPosition READ tabPosition WRITE setTabPosition )
    Q_PROPERTY( TabShape tabShape READ tabShape WRITE setTabShape )
    Q_PROPERTY( int margin READ margin WRITE setMargin )
    Q_PROPERTY( int currentPage READ currentPageIndex WRITE setCurrentPage )
    Q_PROPERTY( int count READ count )
    Q_OVERRIDE( bool autoMask DESIGNABLE true SCRIPTABLE true )

public:
    ExtQTabWidget( QWidget *parent = 0, const char *name = 0, WFlags f = 0 );
    ~ExtQTabWidget();

    virtual void addTab( QWidget *, const QString & );
    virtual void addTab( QWidget *child, const QIconSet& iconset,
			 const QString &label );
    virtual void addTab( QWidget *, ExtQTab* );

    virtual void insertTab( QWidget *, const QString &, int index = -1 );
    virtual void insertTab( QWidget *child, const QIconSet& iconset,
			    const QString &label, int index = -1 );
    virtual void insertTab( QWidget *, ExtQTab*, int index = -1 );

    virtual void moveTab( QWidget *, int index = -1 );

    void changeTab( QWidget *, const QString &);
    void changeTab( QWidget *child, const QIconSet& iconset,
		    const QString &label );

    bool isTabEnabled(  QWidget * ) const;
    void setTabEnabled( QWidget *, bool );

    QString tabLabel( QWidget * ) const;
    void setTabLabel( QWidget *p, const QString &l );

    QIconSet tabIconSet( QWidget * w ) const;
    void setTabIconSet( QWidget * w, const QIconSet & iconset );

    void removeTabToolTip( QWidget * w );
    void setTabToolTip( QWidget * w, const QString & tip );
    QString tabToolTip( QWidget * w ) const;

    QWidget * currentPage() const;
    QWidget *page( int ) const;
    QString label( int ) const;
    int currentPageIndex() const;
    int indexOf( QWidget * ) const;

    QSize sizeHint() const;
    QSize minimumSizeHint() const;

    enum TabPosition { Top, Bottom };
    TabPosition tabPosition() const;
    void setTabPosition( TabPosition );

    enum TabShape { Rounded, Triangular };
    TabShape tabShape() const;
    void setTabShape( TabShape s );

    int margin() const;
    void setMargin( int );

    int count() const;

public slots:
    void setCurrentPage( int );
    virtual void showPage( QWidget * );
    virtual void removePage( QWidget * );

protected:
    void showEvent( QShowEvent * );
    void resizeEvent( QResizeEvent * );
    void setTabBar( ExtQTabBar * );
    ExtQTabBar* tabBar() const;
    void styleChange( QStyle& );
    void updateMask();
    bool eventFilter( QObject *, QEvent * );

signals:
    void currentChanged( QWidget * );
#ifndef Q_QDOC
    void selected( const QString& );
#endif

private slots:
    void showTab( int i );

private:
    ExtQTabWidgetData *d;
    void setUpLayout( bool = FALSE );

    friend class ExtQTabDialog;

private:	// Disabled copy constructor and operator=
#if defined(Q_DISABLE_COPY)
    ExtQTabWidget( const ExtQTabWidget & );
    ExtQTabWidget& operator=( const ExtQTabWidget & );
#endif
};

#endif

