#ifndef EXTQTABBAR_H
#define EXTQTABBAR_H

#ifndef QT_H
#include "qwidget.h"
#include "qpainter.h"
#include "qlist.h"
#include "qiconset.h"
#endif // QT_H

#ifndef QT_NO_TABBAR

class ExtQTab
{
public:
    ExtQTab(): enabled( TRUE ), id( 0 ), iconset( 0 ) {}
    virtual ~ExtQTab();
    ExtQTab( const QString& text )
	: label( text ), enabled( TRUE ), id( 0 ), iconset( 0 ) {}
    ExtQTab( const QIconSet& icon, const QString& text = QString::null )
	: label( text ), enabled( TRUE ), id( 0 ), iconset( new QIconSet(icon) ) {}
#if 1
    void setText( const QString& text) { label = text; }
    QString text() const { return label; }
    void setIconSet( const QIconSet& icon ) { iconset = new QIconSet( icon ); }
    QIconSet* iconSet() const { return iconset; }
    void setRect( const QRect& rect ) { r = rect; }
    QRect rect() const { return r; }
    void setEnabled( bool enable ) { enabled = enable; }
    bool isEnabled() const { return enabled; }
    void setIdentifier( int i ) { id = i; }
    int identitifer() const { return id; }
#endif
// private: (public for compatibility, ### change 3.0)
    QString label;
    QRect r;    // the bounding rectangle of this - may overlap with others
    bool enabled;
    int id;
    QIconSet* iconset;     // optional iconset
};


struct ExtQTabPrivate;


class ExtQTabBar: public QWidget
{
    Q_OBJECT
    Q_ENUMS( Shape )
    Q_PROPERTY( Shape shape READ shape WRITE setShape )
    Q_PROPERTY( int currentTab READ currentTab WRITE setCurrentTab )
    Q_PROPERTY( int count READ count )
    Q_PROPERTY( int keyboardFocusTab READ keyboardFocusTab )

public:
    ExtQTabBar( QWidget *parent = 0, const char *name = 0 );
   ~ExtQTabBar();

    enum Shape { RoundedAbove, RoundedBelow,
		 TriangularAbove, TriangularBelow };

    Shape shape() const;
    virtual void setShape( Shape );

    void show();

    virtual int addTab( ExtQTab * );
    virtual int insertTab( ExtQTab *, int index = -1 );
    virtual void removeTab( ExtQTab * );

    virtual void moveTab( ExtQTab *, int index = -1 );

    virtual void setTabEnabled( int, bool );
    bool isTabEnabled( int ) const;

    QSize sizeHint() const;
    QSize minimumSizeHint() const;
    QSizePolicy sizePolicy() const;

    int currentTab() const;
    int keyboardFocusTab() const;

    ExtQTab * tab( int );
    ExtQTab * tabAt( int );
    int indexOf( int ) const;
    int count() const;

    virtual void layoutTabs();

    static QPopupMenu *pTabMenu;

public slots:
    virtual void setCurrentTab( int );
    virtual void setCurrentTab( ExtQTab * );

signals:
    void  selected( int );
    void tabOrderChanged();

protected:
    virtual void paint( QPainter *, ExtQTab *, bool ) const; // ### not const
    virtual void paintLabel( QPainter*, const QRect&, ExtQTab*, bool ) const;

    void focusInEvent( QFocusEvent *e );
    void focusOutEvent( QFocusEvent *e );

    virtual ExtQTab * selectTab( const QPoint & p ) const;
    void updateMask();

    void resizeEvent( QResizeEvent * );
    void paintEvent( QPaintEvent * );
    void mousePressEvent ( QMouseEvent * );
    void mouseReleaseEvent ( QMouseEvent * );
    virtual void mouseMoveEvent ( QMouseEvent *e );
    void dragMoveEvent( QDragMoveEvent *e );
    virtual void dropEvent ( QDropEvent * );
    void keyPressEvent( QKeyEvent * );
    void styleChange( QStyle& );

    QList<ExtQTab> * tabList();

private slots:
    void scrollTabs();

private:
    QList<ExtQTab> * l;
    QList<ExtQTab> * lstatic;
    void makeVisible( ExtQTab* t  );
    void updateArrowButtons();
    ExtQTabPrivate * d;

    bool selectreleased;

private:	// Disabled copy constructor and operator=
#if defined(Q_DISABLE_COPY)
    ExtQTabBar( const ExtQTabBar & );
    ExtQTabBar& operator=( const ExtQTabBar & );
#endif
};


#endif // QT_NO_TABBAR

#endif // QTABBAR_H

#ifndef QEXTTABWIDGET_H
#define QEXTTABWIDGET_H

#ifndef QT_H
#include "qwidget.h"
#include "qiconset.h"
#endif // QT_H

#ifndef QT_NO_TABWIDGET

class ExtQTabBar;
class ExtQTabWidgetData;
class ExtQTab;
class QWidgetStack;


class ExtQTabWidget : public QWidget
{
    Q_OBJECT
    Q_ENUMS( TabPosition )
    Q_ENUMS( TabShape )
    Q_PROPERTY( TabPosition tabPosition READ tabPosition WRITE setTabPosition )
    Q_PROPERTY( TabShape tabShape READ tabShape WRITE setTabShape )
    Q_PROPERTY( int margin READ margin WRITE setMargin )
    Q_PROPERTY( int currentPage READ currentPageIndex WRITE setCurrentPage )
    
public:
    
    ExtQTabWidget( QWidget *parent, const char *name, WFlags f);
    ExtQTabWidget( QWidget *parent=0, const char *name=0);
   ~ExtQTabWidget();

    void addTab( QWidget *, const QString &);
    void addTab( QWidget *child, const QIconSet& iconset, const QString &label);
    void addTab( QWidget *, ExtQTab* );

    void insertTab( QWidget *, const QString &, int index = -1);
    void insertTab( QWidget *child, const QIconSet& iconset, const QString &label, int index = -1);
    void insertTab( QWidget *, ExtQTab*, int index = -1 );

    void changeTab( QWidget *, const QString &);
    void changeTab( QWidget *child, const QIconSet& iconset, const QString &label);

    virtual void moveTab( QWidget *, int index = -1 );

    bool isTabEnabled(  QWidget * ) const;
    void setTabEnabled( QWidget *, bool );

    void showPage( QWidget * );
    void removePage( QWidget * );
    QString tabLabel( QWidget * ); //### const in 3.0
    void setTabLabel( QWidget *p, const QString &l );
 
    QWidget * currentPage() const;
    int currentPageIndex() const;
    void setCurrentPage( int );
	void setCurrentPageIndex( int );

    QSize sizeHint() const;
    QSize minimumSizeHint() const;

    enum TabPosition { Top, Bottom };
    TabPosition tabPosition() const;
    void setTabPosition( TabPosition );

    enum TabShape { Rounded, Triangular };
    TabShape tabShape() const;
    void setTabShape( TabShape s );
    
    int margin() const;
    void setMargin( int );

    QSizePolicy sizePolicy() const;

protected:
    void showEvent( QShowEvent * );
    void resizeEvent( QResizeEvent * );
    void setTabBar( ExtQTabBar* );
    ExtQTabBar* tabBar() const;
    void styleChange( QStyle& );
    void updateMask();
    bool eventFilter( QObject *, QEvent * );

signals:
    void currentChanged( QWidget * );
    void selected( const QString& ); // obsolete

private slots:
    void showTab( int i );

private:
    ExtQTabWidgetData *d;
    void setUpLayout(bool = FALSE);
    void init();

#if 1 //def TOTAL_LOSER_COMPILER
    friend class ExtQTabDialog;
#else
    friend void ExtQTabDialog::setTabBar( ExtQTabBar* );
    friend void ExtQTabBar* ExtQTabDialog::tabBar() const;
#endif

private:	// Disabled copy constructor and operator=
#if defined(Q_DISABLE_COPY)
    ExtQTabWidget( const ExtQTabWidget & );
    ExtQTabWidget& operator=( const ExtQTabWidget & );
#endif
};

#endif // QT_NO_TABWIDGET

#endif // QEXTTABWIDGET_H


