/**
 * @file ctags.cpp
 * @brief CTAGS module for Freddy
 * 
 * @author Ferenc Moricz
 * Created:   2001/05/16
 *
 * $Revision: 1.12 $
 * $Date: 2004/02/22 00:12:49 $
 */

#include "ctags.h"
#include "Editor.h"

#include <qcache.h>
#include <qdir.h>
#include <qfileinfo.h>
#include <qpainter.h>

#include <stdio.h>

Module *pMod;

QPtrDict<ETags_CTags> ctagcache;
QDict<ETags_CTags> ctags;
ETags_TaggedFileDict taggedfiledict;
ETags_CTags *pCurrentDict = NULL;

/*static ETags_CTags *getCTagsFile(EditBuffer *buffer);
static void loadCTagsFile(ETags_CTags *tags, EditBuffer *buffer);*/

// .module_init(.)
/**
 * CTag's module init function
 */
Module *module_init(Module *pModule)
{
	pMod = pModule;
	pMod->qsName = new QString("ctags");
	return pMod;
}

#include <qfiledialog.h>
#include <qmessagebox.h>
QString *module_function(Module *pModule, const QString &qsFunction,
            QString *qsParam1, QString *qsParam2, int iParam1, int iParam2)
{
	/* get string that is below the cursor */
	char *pWord = (char *)pMod->pQuery("cb.buffer.currentword");

	/* go to the appropriate line of the appropriate buffer */

	// get the tag from the CTAG file
	//! ETags_CTags *t = NULL;
	EditBuffer *pBuffer = (EditBuffer *)pMod->pQuery("cb.get.buffer");
	/*ETags_CTags *t = ctagcache.find(buffer);*/

	/* search for the definition of a function */
	if ( ((qsFunction == "1") || (qsFunction == "2")) &&
	    (pWord != NULL) && (pCurrentDict != NULL))
	{
		QList<ETags_CTagsTag> *tag = pCurrentDict->find(pWord);
		if (tag==NULL)
		{
			DEBUG_WARNING("ctags plugin - tag '%s' not found!", pWord);
		} else
		for(int i=0; i<(signed)tag->count(); i++)
		{                                   // search for the tag in the list
			ETags_CTagsTag *ttag = tag->at(i);
			if (ttag == NULL) return NULL;
			/*printf("Tag(%d) found in: %s, ", i, ttag->path->latin1());
			printf("ind:%c mask:%s\n", ttag->kind, ttag->mask->latin1());*/

			/* search for the open the tag with type 'function' */
			if ( ((qsFunction == "2") && (ttag->kind == 'f')) ||
			     ((qsFunction == "1") && (ttag->kind == 'p')) )
			{
				printf("name: %s, mask: %s path:%s\n", ttag->name->latin1(),  ttag->mask->latin1(),
				     ttag->path->latin1());
				QFileInfo fi;           /* get absolute path to ttag->path */
				fi.setFile(*(pCurrentDict->ctagdir), *(ttag->path));

				/* search for the buffer where this file is loaded */
				if (!pMod->pQuery("cb.show.editor.filebypath", fi.absFilePath()))
				{               /* not found; ask user whether to load it */
					QString sMessage = (*(ttag->path));
					sMessage.append(" is not opened.<br><b>Open</b> it?");
					QMessageBox mb("Freddy", sMessage,
					     QMessageBox::Warning, QMessageBox::Yes|QMessageBox::Default,
					     QMessageBox::No|QMessageBox::Escape, QMessageBox::NoButton,
					     NULL, "", true);
					mb.setFocus();

					if ( mb.exec() == QMessageBox::Yes )
					{           /* user said yes -> open the file */
						pMod->pQuery("cb.editor.openfile", fi.absFilePath());
					}
				}

				/* Go to the line which contains the tag */
				{
					int iAllLines = (int)pMod->pQuery("cb.get.buffer.lines.count");

					/* searching for the appropriate line (var: iLine) */
					QString sLineNumber, sColNumber;
					for(int iLine = 0; iLine < iAllLines; iLine ++)
					{
						sLineNumber.setNum( iLine );
						QString *pLine = (QString *)pMod->pQuery("cb.get.buffer.lines.at", sLineNumber.latin1());
						if (pLine && *(ttag->mask) == (*pLine))    /* we found the line ! */
						{
							/* searching for the first character of the name (var: iFinalPos) */
							QString sWithClass = ":: *";
							sWithClass.append(pWord);
							QRegExp sWithRegExp(sWithClass);
							int iPurePos = pLine->find(pWord), iClassPos = pLine->find(sWithRegExp), iFinalPos;
							if (iClassPos != -1) iClassPos = pLine->find(pWord, iClassPos);
							iFinalPos = (iClassPos != -1 ? iClassPos : iPurePos);
							printf(" pw: %s, s: %s\n", pWord, pLine->latin1());
							printf("$$$ %d %d %d\n", iPurePos, iClassPos, iFinalPos);

							/* displaying the result - placing the cursor after the name and selecting the name */
							sColNumber.setNum(iFinalPos);
							pMod->pQuery("cb.set.buffer.cursor.position", sColNumber.latin1(), sLineNumber.latin1());
							pMod->pQuery("cb.set.buffer.selectmode", "1");
							sColNumber.setNum(iFinalPos + strlen(pWord));
							pMod->pQuery("cb.set.buffer.cursor.position", sColNumber.latin1(), sLineNumber.latin1());
							pMod->pQuery("cb.do.buffer.find.cursor", sColNumber.latin1(), sLineNumber.latin1());
							pMod->pQuery("cb.set.buffer.selectmode", "0");
							return NULL;
						}
					}
				}
			}

/* orig searching
				QDir dir = fi.dir(true);

				EditSession *stack = (EditSession *)global_editor->getStack()->visibleWidget();
				for(int i=0; i<(int)stack->count(); i++)
				{
					EditBuffer *buffer = (EditBuffer *)stack->page(i);
					QFile *f = buffer->getFile();
					if (f != NULL) fi.setFile(*f);
					if ( (fi.exists()) &&
					     (fi.dir(true).canonicalPath() == dir.canonicalPath()) )
					{
						printf("OK, megvan!\n");
						stack->setCurrentPage(i);
						QRegExp mask(*(ttag->mask));
						for(int line=0; line < (int)buffer->
						               text->lines.count(); line++)
							if (*(ttag->mask) == *(buffer->
						               text->lines.at(line)))
							{
								buffer->cmd_scrollTo(line);
								buffer->cmd_setCursor(0, 0);
							}
					}
				}
			break;
*/
		}
	}

#if 0
	/* search for the declaration of a function */
	else if ((qsFunction == "2") && (pWord != NULL))
	{
		QList<ETags_CTagsTag> *tag = t->find(pLine->mid(i1, i2-i1+1));
		if (tag==NULL) printf("Tag not found!\n");
		else for(int i=0; i<(signed)tag->count(); i++)
		{                                   // search for the tag in the list
			ETags_CTagsTag *ttag = tag->at(i);
			if (ttag == NULL) return NULL;
			//printf("Tag(%d) found in: %s, ", i, ttag->path->latin1());
			//printf("kind:%c mask:%s\n", ttag->kind, ttag->mask->latin1());

			if (ttag->kind == 'p')
			{
				ETags_DisplayText *dp = new ETags_DisplayText(ttag->mask, buffer);
				dp->show();
				dp->setFocus();
				break;
			}
		}
	}
#endif

	/* create/initialize the mapbar */
	else if (qsFunction == "3")
	{
		/* if the mapbar plugin is not loaded we can do nothing */
		Module *pMapBar = (Module *)pMod->pQuery("cb.module.find", "mapbar");

		if (pMapBar)
		{
			/* testing if the mapbar is already present */
			QString sISHERE = "ishere";
			QString *qsRes = pMapBar->pFunction(pMapBar, sISHERE);

			if ((*qsRes != "yes") && (*qsRes != "true"))
			{
				/* mapbar is not created - creating one */
				pMapBar->pFunction(pMapBar, QString("create"));
				qsRes = pMapBar->pFunction(pMapBar, QString("ishere"));
			}
			if (*qsRes == "yes")
			{
				DEBUG_MESSAGE("Mapbar is already here");
			}
		}
	}

	else if (qsFunction == "4")
	{
		Module *pMapBar = (Module *)pMod->pQuery("cb.module.find", "mapbar");
		if ((pMapBar != NULL) && (pBuffer != NULL))
		{
			QString sCommand = "repaint";
			ETags_create(pMapBar, pBuffer);
			pMapBar->pFunction(pMapBar, sCommand);
		}
	}

	else if (qsFunction == "5")
	{
		/* show "open file" window */
		QString sFileName = QFileDialog::getOpenFileName( "/home/moricz/program/", 
		          "CTag file (tags)",
	              NULL, "Open CTAGS file...", "Open: Select a CTAGS file to load" );
		if (!sFileName.isNull())
		{
			ETags_CTags *pNewTags = new ETags_CTags();
			ctags.insert(sFileName, pNewTags);
			printf("loadctags!!!\n");
			pNewTags->loadCTagsFile( sFileName );
			pCurrentDict = pNewTags;
			/* ctagcache.insert(buffer, ret); */
		}
	}

	return NULL;
}

ETags_CTags::ETags_CTags(QFile *file) : QDict< QList<ETags_CTagsTag> > () 
{
	this->file = file;
}

#include <qdir.h>
#include <qfileinfo.h>

/*
 * getCTagsFile()  [ EditBuffer *buffer ]
 *                 ->ETags_CTags *
 *
 * Search for the appropriate CTAGS file (with name "tags") for the buffer
 */

#if 0
static ETags_CTags *getCTagsFile(EditBuffer *buffer)
{
	ETags_CTags *ret = ctagcache.find(buffer);

	if (ret == NULL)
	{
		QFileInfo fi(*(buffer->getFile()));
		QDir d(fi.dirPath(true));

		/* find the ctag file and insert it to the lists */

		while(d.exists())              // go up and up and look for the file
		{
			if (d.exists("tags",false))// does a tags file exist here?
			{
				QFile *file = new QFile(d.filePath("tags", true));

				if ((ret = ctags.find(d.filePath("tags", true))) == NULL)
				{                       // put the file into the cache
					ret = new ETags_CTags(file);
					ret->ctagdir = new QDir(d);
					ctags.insert(d.filePath("tags", true), ret);
					loadCTagsFile(ret, buffer);
				}

				ctagcache.insert(buffer, ret);

				break;
			}
			else if (!d.cdUp()) break;
		}
	}

	return ret;
}
#endif

// ETags_CTags::loadCTagsFile(.)
/**
 *  Opens a CTAGS file
 *  @bug buffer overrun (long lines))
 *  @callgraph
 */
void ETags_CTags::loadCTagsFile( QString &sFileName )
{
	char buf[512], *pe1, *pe2, *tkind;
	ETags_TaggedFile *tagged = NULL;
	QFile oFile(sFileName);

	/* I. opening the 'tags' file */
	/* I.1. checking if the file is present */
	if (oFile.exists())
	{
		/* I.2. opening the file */
		if (!oFile.open( IO_ReadOnly )) return;

		QFileInfo oFileFI(oFile);
		ctagdir = new QDir( oFileFI.dirPath( true ) );
		QFileInfo fi("/");
		QDir dir("/");
		/* II. reading the file - cycle until the end */
		while( !oFile.atEnd() )         
		{
			/* II.1 reading one line */
			oFile.readLine( buf, 512 );

			/* II.2 checking if it is a comment or it is too short */
			if (buf[0]=='!') continue;
			if ( ((pe1 = strchr(buf, '\t')) != NULL) &&
			     ((pe2 = strchr(pe1+1, '\t')) !=NULL) )
			{
				/* II.3 we're sure it is a real tag - creating an object for it */
				ETags_CTagsTag *tag = new ETags_CTagsTag();
				*pe1 = '\0';
				*pe2 = '\0';
				tag->name = new QString(buf);

				tagged = taggedfiledict.searchFile(new QString(pe1+1));
				if (tagged->abspath == NULL)  /// if it is new, set abspath
				{
					fi.setFile(*ctagdir, pe1+1);
					dir.setPath(fi.dirPath(true));
					tagged->abspath = new QString(dir.canonicalPath());
					tagged->abspath->append("/");
					tagged->abspath->append(fi.fileName());
					//!!taggedfiledict.insert(*tagged);
				}
				tag->path = tagged->path;
				tagged->tags.append(tag);

				tkind = strstr(pe2+1, ";\"\t");   // "      /// kind of tag
				if ((tkind != NULL) && (tkind[3]>='a') &&
				    (tkind[3]<='z') && ((tkind[4]<'a')||(tkind[4]>'z')))
					tag->kind = tkind[3];

				pe1 = pe2+1;            /// mask
				if ((pe2 = strstr(pe1, ";\"\t")) != NULL) *(pe2-2) = '\0'; //"
				tag->mask = new QString(pe1+2);

				                        /// add the tag into the dictionary
				QList<ETags_CTagsTag> *list;
				if ((list=find(*tag->name)) == NULL)
				{
					list = new QList<ETags_CTagsTag>();
					insert(*tag->name, list);
				}
				list->append(tag);
			}
		}
		oFile.close( );
	}
}

ETags_TaggedFileDict::ETags_TaggedFileDict() : QDict< QList<ETags_TaggedFile> > () 
{
}

// .ETags_TaggedFileDict::searchFile(..)
/**
 * Tries to find the record of the given filename in the dictionary (if 
 *    'create' is true, it creates it)
 * @param which the file name to search for
 * @param create should we create a new record for the filename
 * @return the record for the given file
 */
ETags_TaggedFile *ETags_TaggedFileDict::searchFile(QString *which, bool create)
{
	QList<ETags_TaggedFile> *taggedfiles;
	ETags_TaggedFile *tagged, *ret = NULL;

	/* I. searching for the file in the dict */
	QFileInfo fi(QDir("/home/moricz/program/"), *which); /// search for the filelist
	taggedfiles = find(fi.fileName());

	/* I.a we not found the file */
	if (taggedfiles == NULL)
	{
		/* I.a.b 'create' is false - returning with NULL */
		if (!create) return NULL;
		/* I.a.a 'create' is true -> creating a new filelist */
		taggedfiles = new QList<ETags_TaggedFile>;
		insert(fi.fileName(), taggedfiles);
	}

	for(int i=0; i < (int)taggedfiles->count(); i++)/// search for the file
	{
		tagged = taggedfiles->at(i);
		if ( (create && (*(tagged->path) == *which)) ||
		     (!create && (*(tagged->abspath) == *which)) )
		{
			ret = tagged;
			break;
		}
	}

	if (ret == NULL)                    /// if file not found, insert one
	{
		if (!create) return NULL;
		ret = new ETags_TaggedFile();
		ret->path = new QString(*which);
		taggedfiles->append(ret);
	}

	return ret;
}

ETags_DisplayText::ETags_DisplayText(QString *text, QWidget *p) : QDialog(p)
{
	this->text = text;
}

QSize ETags_DisplayText::sizeHint() const
{
	return (*new QSize(400, 15));
}

void ETags_DisplayText::paintEvent(QPaintEvent *pe)
{
	QPainter p(this);

	p.begin( this );
	p.drawText(1, 12, *text);
	p.end();
}

bool ETags_DisplayText::event( QEvent *e )
{
	if ( (e->type() == QEvent::KeyPress) ||
	     (e->type() == QEvent::KeyRelease) )
	{
		parentWidget()->setFocus();
		close(true);
		return true;
	}

	return QWidget::event( e );
}

/* 
 * MapBar::create() - fills a mapbar with members
 */

void ETags_create(Module *pMapBar, EditBuffer *pBuffer)
{
	ETags_TaggedFile *tagged;
	QFile *pBufferFile = (QFile *)pMod->pQuery("cb.get.buffer.filepath", "");
	QFileInfo fi;
	if (pBufferFile != NULL) fi.setFile(*pBufferFile);

	printf("filepath: %s\n", fi.absFilePath().latin1());
	tagged = taggedfiledict.searchFile(new QString(fi.absFilePath()), false);
	if (tagged != NULL)
	{
		for(int i=0; i<(int)tagged->tags.count(); i++)
			printf("  %d.%s\n", i, tagged->tags.at(i)->name->latin1());
	}

	if (tagged != NULL)
	{
		for(int i = 0; i<(int)tagged->tags.count(); i++)
		{
			ETags_CTagsTag *stag = tagged->tags.at(i);
			QRegExp mask(*(stag->mask));

		for(int line=0; line < (int)pBuffer->text->lines.size(); line++)
				if (*(stag->mask) == *(pBuffer->text->lines[line]))
				{
					stag->line = line;
					QString *acln = pBuffer->text->lines[line];
					stag->what = new QString();
					int n = acln->find(*(stag->name));
					stag->what->append(acln->left(n));
					stag->what->append("<b><font color=#0000ff>");
					stag->what->append(*(stag->name));
					stag->what->append("</font></b>");
					stag->what->append(acln->mid(n+stag->name->length()));
				}
		}

		                        /// Sort tags by line number
		for(int i=0; i<(int)tagged->tags.count()-1; i++)
			for(int j=i; j<(int)tagged->tags.count(); j++)
			{
				ETags_CTagsTag *stag1 = tagged->tags.at(i);
				ETags_CTagsTag *stag2 = tagged->tags.at(j);
				if (stag1->line > stag2->line)
				{
					tagged->tags.remove(i);
					tagged->tags.insert(i, stag2);
					tagged->tags.remove(j);
					tagged->tags.insert(j, stag1);
				}
			}

		if (pBuffer != NULL)                                                            
		for(unsigned int i=0; i < tagged->tags.count(); i++)
		{
			ETags_CTagsTag *stag = tagged->tags.at(i);
			if (!stag) continue;
			unsigned int iTagLength;

			if (i + 1 < tagged->tags.count())
				iTagLength = tagged->tags.at(i + 1)->line - 1;
			else iTagLength = pBuffer->text->lines.size();

			if ( (stag->line >= (int)pBuffer->text->lines.size()) ||
			     ((int)(stag->line + iTagLength) >= (int)pBuffer->text->lines.size()) )
				continue;

			for(unsigned int j = stag->line; j < iTagLength; j++)
				if ( (pBuffer->text->lines[j]) &&
				     (pBuffer->text->lines[j]->startsWith("}")) ) iTagLength = j;
			iTagLength -= stag->line - 1;

			pMapBar->pFunction(pMapBar, "additem", stag->name, NULL, stag->line, iTagLength);

			DEBUG_PRINT("%s, line=%d\n", stag->name->latin1(),
			                    stag->line);
		}
//		lines = buffer->text->lines.count();
//		pModule->show();
	} else DEBUG_ERROR("There is no information about this file!");
}

