/**
 * @file mapbar.cpp
 * @brief mapbar module - implementation
 *
 * MapBar. A scrollbar-like widget to display the actual file's structure.
 * 
 * @author Ferenc Moricz
 * Created:   2002/02/03
 *
 * $Date: 2003/11/18 02:32:58 $
 * $Revision: 1.10 $
 */

#include "mapbar.h"

const int MB_LEFTSPACE = 8;

#include <qpainter.h>
#include <qpen.h>
#include <qptrdict.h>

QPtrDict<Module_MapBar> *global_widgetdict = NULL;
Module *pMod;

// .module_init(.)
/**
 * MapBar's module init function.
 * @param pModule 
 */

Module *module_init(Module *pModule)
{
	pModule->qsName = new QString("mapbar");
	if (global_widgetdict == NULL) global_widgetdict = new QPtrDict<Module_MapBar>();
	return pMod = pModule;
}

#include <qwidget.h>
/**
 * MapBar's service function.
 */
QString *module_function(Module *pModule, const QString &qsFunction,
            QString *qsParam1, QString *qsParam2, int iParam1, int iParam2)
{
	EditBuffer *pBuffer = (EditBuffer *)pMod->pQuery("cb.get.ptr.buffer");
	if (pBuffer == NULL) return NULL;

	if (qsFunction == "event.cursor")
	{
		Module_MapBar *pMapBar;
		if ((pMapBar = global_widgetdict->find(pBuffer)) != NULL)
		{
			pMapBar->iY = iParam2;
			pMapBar->repaint(false);
		}
	} else
	if ((qsFunction == "3") || (qsFunction == "create"))
	{
		if (pMod->pQuery("cb.get.buffer.havescrollbar"))
		{
			Module_MapBar *mb = new Module_MapBar(NULL, NULL);
			pMod->pQuery("cb.set.widget.mapholder", NULL, NULL, mb);

			if (global_widgetdict != NULL)
				global_widgetdict->insert( pBuffer, mb );
			else DEBUG_ERROR("MapBar not initialized!");

//			mb->addItem(NULL, new QString("Kaka"), NULL, 0, 10);
//			mb->addItem(NULL, new QString("Cica"), NULL, 30, 25);
//			mb->addItem(NULL, new QString("Valami"), NULL, 57, 50);
		}
	}
	else if (qsFunction == "additem")
	{
		Module_MapBar *pMapBar;
		if ((pMapBar = global_widgetdict->find(pBuffer)) != NULL)
		{
			pMapBar->addItem(NULL, qsParam1, qsParam2, iParam1, iParam2);
		}
	}
	else if (qsFunction == "ishere")
	{
		if (global_widgetdict->find(pBuffer) != NULL)
			return new QString("true");
		else return new QString("false");
	}
	else if (qsFunction == "repaint")
	{
		Module_MapBar *pMapBar;
		if ((pMapBar = global_widgetdict->find(pBuffer)) != NULL)
		{
			pMapBar->fullRepaint();
		}
	}

	return NULL;
}

Module_MapBar::Module_MapBar(QWidget *parent, EditBuffer *buffer)
	: QWidget(parent, "mapbar", WRepaintNoErase)
{
	Items = *new QList<Module_MapBar_Item>;
	scrollbar = NULL;
	setSizePolicy( QSizePolicy(QSizePolicy::Fixed, QSizePolicy::Maximum) );
}

QSize Module_MapBar::sizeHint() const
{
	constPolish();
	return (*new QSize(16+MB_LEFTSPACE, 32767));
}

#include <qapplication.h>
#include <qstyle.h>
#include <qtooltip.h>

/**
 * Painting the bars and its labels (the background)
 *
 * @param pPixmap the pixmap to paint onto
 */
void Module_MapBar::paintBar(QPixmap *pPixmap)
{
	QPainter p;
	QPen pen(black);
	int nfunct, iFrom, iTo, odd=0;
	int bufferlines = (int)pMod->pQuery("cb.get.buffer.lines.count");
	int allines = bufferlines + (int)pMod->pQuery("cb.get.buffer.visiblelines.count") + 1;
	int iBottom=0, iTop=0;

	if (!pMod->pQuery("cb.get.buffer.havescrollbar"))
	{
		DEBUG_ERROR("ScrollBar doesn't exists!");

		iBottom = iTop = 10;
		return;
	} else
	{
#if 1
//#if QT_VERSION > 280
		QScrollBar *pVScroll = (QScrollBar *)pMod->pQuery("cb.get.buffer.getscrollbar");
		/*!!*/
		iTop = QApplication::style().querySubControlMetrics(
		          QStyle::CC_ScrollBar, pVScroll, QStyle::SC_ScrollBarGroove).y();
		iBottom = QApplication::style().querySubControlMetrics(
		          QStyle::CC_ScrollBar, pVScroll, QStyle::SC_ScrollBarGroove).height();

/*		iTop = QApplication::style().querySubControlMetrics(
		          QStyle::CC_ScrollBar, buffer->vscroll, QStyle::SC_ScrollBarGroove).y();
		iBottom = QApplication::style().querySubControlMetrics(
		          QStyle::CC_ScrollBar, buffer->vscroll, QStyle::SC_ScrollBarGroove).height();
*/
//		iBottom = (int)pMod->pQuery("cb.get.buffer.text.height") - (iTop + iBottom);
		iBottom = pVScroll->height() - (iTop + iBottom);
#else
		iBottom = iTop = 10;
#endif
	}

	pPixmap->fill(Qt::lightGray);
	p.begin( pPixmap );                    /* begin to paint */
	QColor *pTextBgColor =  (QColor *)pMod->pQuery("cb.get.properties.color",
	        (char *)"buffer.text.background.color", (char *)&gray );

	p.setPen(black);
	QFont *nfont = new QFont( "courier", 12 );
	p.setFont(*nfont);

	QWMatrix matrix;                    /* rotating text */
	matrix.rotate(90);
	p.setWorldMatrix(matrix, false);

	/* print labels */
	sw = ((float)height()-iTop-iBottom) / allines;

	/* print 'after-file' area */
	QBrush pBr(darkGray);
	pBr.setColor( *(QColor *)pMod->pQuery("cb.get.properties.color",
	        (char *)"module.widget.mapbar.afterend.color", (char *)&darkGray ));
	p.fillRect((int)(iTop + (bufferlines + 1) * sw), -MB_LEFTSPACE,
	        (int)(iTop + (int)pMod->pQuery("cb.get.buffer.visiblelines.count") * sw), -15, pBr);

	/* print scroll bar dead areas */
	QColor *pDeadAreaColor = (QColor *)pMod->pQuery("cb.get.properties.color",
	        (char *)"module.widget.mapbar.deadarea.color", NULL );
	pBr.setColor(pDeadAreaColor ? *pDeadAreaColor : *pTextBgColor);
	p.fillRect(0, 0, iTop, -15-MB_LEFTSPACE, pBr);
	p.fillRect(size().height()-iBottom, 0, iBottom, -15-MB_LEFTSPACE, pBr);

	/* 'out of bar area' is on the left of the mapbar's area */
	pBr.setColor(*pTextBgColor);
	p.fillRect(0, 0, size().height(), -MB_LEFTSPACE+3, pBr);
	p.translate(iTop, 0);

	for(nfunct=0; nfunct<(int)Items.count(); nfunct++)
	{
		Module_MapBar_Item *pItem = Items.at(nfunct);
		QBrush *br;

		iFrom = (int)((float)pItem->iLine * sw);
		iTo = (int)((float)(pItem->iLine + pItem->iLength) * sw);
		iTo = min( iTo, (int)((bufferlines - 1)*sw) );

#if QT_VERSION > 280
		p.setClipRect(iFrom, -15-MB_LEFTSPACE, iTo - iFrom + 2, 15, QPainter::CoordPainter);
#else
//		p.setClipRect(0, iFrom, 15, iTo - iFrom + 2);
#endif
		br=new QBrush(QColor(125, 185, 125));        /* fill background */
		p.fillRect(iFrom, 0-MB_LEFTSPACE, iTo-iFrom+1, -15-MB_LEFTSPACE, *br);
		p.setPen(QColor(225, 240, 225));             /* print frame */
		p.drawLine(iFrom, 0-MB_LEFTSPACE, iTo, 0-MB_LEFTSPACE);
		p.drawLine(iFrom, 0-MB_LEFTSPACE, iFrom, -15-MB_LEFTSPACE);
		p.setPen(white);
		p.drawLine(iFrom, 1-MB_LEFTSPACE, iTo, 1-MB_LEFTSPACE);
		p.drawLine(iFrom+1, 0-MB_LEFTSPACE, iFrom+1, -15-MB_LEFTSPACE);
		if (odd) p.setPen(QColor(100, 155, 100));
		else p.setPen(QColor(60, 135, 60));
		p.drawLine(iTo-1, 0-MB_LEFTSPACE, iTo-1, -15-MB_LEFTSPACE);
		p.drawLine(iFrom, -14-MB_LEFTSPACE, iTo, -14-MB_LEFTSPACE);
		if (odd) p.setPen(QColor(25, 75, 25));
		else p.setPen(QColor(10, 45, 10));
		p.drawLine(iTo, 0-MB_LEFTSPACE, iTo, -15-MB_LEFTSPACE);
		p.drawLine(iFrom, -15-MB_LEFTSPACE, iTo, -15-MB_LEFTSPACE);

		p.setPen(black);                // print text
		p.drawText(iFrom+3, -2-MB_LEFTSPACE, *(pItem->sName));
		QToolTip::add(this, QRect(0-MB_LEFTSPACE, iFrom+15, 15, iTo+14),
		                    *(pItem->sName));
	}
	delete(nfont);
}

void Module_MapBar::paintEvent(QPaintEvent *pe)
{
	QPainter p;
	QPen pen(black);
	int bufferlines = (int)pMod->pQuery("cb.get.buffer.lines.count");
	int allines = bufferlines + (int)pMod->pQuery("cb.get.buffer.visiblelines.count") + 1;
	int iBottom=0, iTop=0;

	if (!pMod->pQuery("cb.get.buffer.havescrollbar"))
	{
		DEBUG_ERROR("ScrollBar doesn't exists!");

		iBottom = iTop = 10;
		return;
	} else
	{
#if 0
//#if QT_VERSION > 280
		QScrollBar *pVScroll = (QScrollBar *)pMod->pQuery("cb.get.buffer.getscrollbar");
		/*!!*/
//		iTop = QApplication::style().querySubControlMetrics(
//		          QStyle::CC_ScrollBar, pVScroll, QStyle::SC_ScrollBarGroove).y();
//		iBottom = QApplication::style().querySubControlMetrics(
//		          QStyle::CC_ScrollBar, pVScroll, QStyle::SC_ScrollBarGroove).height();

/*		iTop = QApplication::style().querySubControlMetrics(
		          QStyle::CC_ScrollBar, buffer->vscroll, QStyle::SC_ScrollBarGroove).y();
		iBottom = QApplication::style().querySubControlMetrics(
		          QStyle::CC_ScrollBar, buffer->vscroll, QStyle::SC_ScrollBarGroove).height();
*/
//		iBottom = (int)pMod->pQuery("cb.get.buffer.text.height") - (iTop + iBottom);
		iBottom = pVScroll->height() - (iTop + iBottom);
#else
		iBottom = iTop = 10;
#endif
	}


	p.begin( this );                    /* begin to paint */
	if (!bPainted)
	{
		paintBar(&oBarPixmap);
		p.drawPixmap(0, 0, oBarPixmap, 0, 0, oBarPixmap.size().width(), oBarPixmap.size().height());
		bPainted = true;
	}

	sw = ((float)height()-iTop-iBottom) / allines;		/* print labels */

	/** @tbd more optimized repainting algorithm */
	if (iPreviousCursorLine != -1)
		p.drawPixmap(0, 0, oBarPixmap, 0, 0, oBarPixmap.size().width(), oBarPixmap.size().height());
		//p.drawPixmap(0, iPreviousCursorLine-5, oBarPixmap, 0, iPreviousCursorLine-5, MB_LEFTSPACE+16, 10);

	QWMatrix matrix;                    /* rotating text */
	matrix.rotate(90);
	p.setWorldMatrix(matrix, false);

	/* calculate the position of the cursor */
	int iCursor = iPreviousCursorLine = (int)((float)iY * sw);
	iCursor += iTop;

	/* paint the cursor */
	p.setClipping( false );
	p.fillRect( iCursor - 4, 0, 9, -MB_LEFTSPACE-2, QColor(0xff, 0, 0) );
	p.setPen(QColor(0x50, 0, 0));
	p.drawLine( iCursor+5, -1, iCursor+5, -MB_LEFTSPACE-5 );
	p.drawLine( iCursor-3, -MB_LEFTSPACE-4, iCursor+5, -MB_LEFTSPACE-4);
	p.drawLine( iCursor-2, -MB_LEFTSPACE-5, iCursor+5, -MB_LEFTSPACE-5);
	p.drawLine( iCursor-2, -MB_LEFTSPACE, iCursor+1, -MB_LEFTSPACE);
	p.drawLine( iCursor+1, -MB_LEFTSPACE, iCursor+1, -MB_LEFTSPACE+5);
	p.setPen(QColor(0xff, 0, 0));
	p.drawLine( iCursor, -MB_LEFTSPACE-2, iCursor, -MB_LEFTSPACE-15 );
	p.drawLine( iCursor+1, -MB_LEFTSPACE-2, iCursor+1, -MB_LEFTSPACE-15 );
	p.end();
}

// Module_MapBar::resizeEvent(.)
/**
 * Resize-event handler
 * When the size of the mapbar is changed we have to resize the oBarPixmap
 * QPixmap also and we have to repaint the widget. That's why this event-handler
 * is here.
 *
 * @param pEvent the incoming resize-event
 */
void Module_MapBar::resizeEvent( QResizeEvent *pEvent )
{
	oBarPixmap.resize(16+MB_LEFTSPACE, pEvent->size().height());
	fullRepaint();
}

#include <qwhatsthis.h>
/*void Module_MapBar::mousePressEvent(QMouseEvent *e)
{
	int nfunct;
	int allines = buffer->text->lines.count();
	float sw = ((float)height()-30)/(allines-1);
	for(nfunct=-1; nfunct<(int)Items.count(); nfunct++)
	{
		Module_MapBar_Item *pItem = Items.at(nfunct);
		int hfrom = (int)((nfunct == -1 ? 0 : pItem->iLine) * sw);
		int iTo = (int)((pItem->iLine + pItem->iLength > allines ?
		        allines-1 : pItem->iLine + pItem->iLength - 1 )*sw);
		if ((e->y() < iTo+15) && (hfrom+15 <= e->y())) break;
	}
	QWhatsThis::display(*Items.at(nfunct)->sWhat, QPoint(-30,e->y()), this);
}*/

void Module_MapBar::addItem(Module_MapBar_Item_Class *pClass,
                            QString *sName, QString *sWhat, int iLine, int iLength)
{
	Module_MapBar_Item *pItem = new Module_MapBar_Item();

	pItem->sName = sName;
	pItem->iLine = iLine;
	pItem->iLength = iLength;
	pItem->pClass = pClass;

	addItem( pItem );
}

void Module_MapBar::addItem(Module_MapBar_Item *pItem)
{
	/* binary search for the place of the item to be inserted */
	int iPos1, iPos2, iHalf, iHalfItem;

	for( iPos1 = 0, iPos2 = Items.count(); iPos1 < iPos2; )
	{
		iHalf = ( iPos1 + iPos2 ) / 2;
		iHalfItem = Items.at(iHalf)->iLine;

		if (pItem->iLine > iHalfItem) iPos1 = iHalf + 1;
		else if (pItem->iLine < iHalfItem) iPos2 = iHalf - 1;
		else break;
	}

	Items.insert( iPos1, pItem );
}


// Module_MapBar::fullRepaint()
/**
 * Fully repaints the mapbar's boxes and labels
 */
void Module_MapBar::fullRepaint(void)
{
	bPainted = false;
	repaint();
}

