/*
 * mapbar - declaration
 *
 * MapBar. A scrollbar-like widget to display the actual file's structure.
 * 
 * Author: Ferenc Moricz
 * Created:   2002/02/03
 *
 * $Date: 2003/11/02 01:52:56 $
 * $Revision: 1.6 $
 */

#include <qdict.h>
#include <qptrlist.h>
#include <qstring.h>
#include <qscrollbar.h>
#include <qwhatsthis.h>
#include <qwidget.h>

#include "EditBuffer.h"
#include "Module.h"

extern "C" MODULES_API QString *module_function(Module *pModule, const QString &qsFunction,
                      QString *qsParam1, QString *qsParam2, int iParam1, int iParam2);

extern "C" MODULES_API Module *module_init(Module *pModule);


class Module_MapBar_Item_Class
{
  public:
	enum Module_MapBar_Item_Type { Section, Selection, Point };

	Module_MapBar_Item_Type eType;
};

class Module_MapBar_Item
{
  public:
	int iLine, iLength;
	QString *sName;
	QString *sWhat;

	Module_MapBar_Item_Class *pClass;
};

class Module_MapBar : public QWidget
{
  public:
	Module_MapBar(QWidget *pWidget) : QWidget(pWidget)
		{ iPreviousCursorLine = -1; bPainted = false; }

	void addItem(Module_MapBar_Item_Class *pClass, QString *sName,
	                                 QString *sWhat, int iLine, int iLength);
	void addItem(Module_MapBar_Item *pItem);
	void fullRepaint(void);

	int lines, iY;
	int scrollbegin, scrollend;

	QScrollBar *scrollbar;

	Module_MapBar(QWidget *parent, EditBuffer *buffer);
	virtual QSize sizeHint() const;

	QPopupMenu *pContextMenu;

	QColor oColAfterEnd, oColDeadArea, oColTextBackground;
	int iVisibleLines;

  protected:
	virtual void paintEvent (QPaintEvent *pe);
	virtual void resizeEvent( QResizeEvent * );

	void paintBar(QPixmap *pPixmap);

	QPixmap oBarPixmap;
	int iPreviousCursorLine;           /** helper variable for more optimized repaint*/
	bool bPainted;
	float sw;                          /** the ratio of the mapped and the real sizes */

	QWhatsThis *what;
	QFont oFont;
/*	virtual void mousePressEvent(QMouseEvent *e);*/

	QList<Module_MapBar_Item> Items;
};

