/**
 * @file Debug.c
 *
 * Freddy's debug definitions
 * 
 * @author Moricz, Ferenc
 * @date january 13., 2004.
 *
 */

#ifdef DEBUGLEVEL
#undef DEBUGLEVEL
#endif

#define DEBUGLEVEL 4

#include "Debug.h"

#include <qstring.h>

#include <stdio.h>
#include <stdarg.h>

nsdebug::Debug oDebugMain;
nsdebug::Debug *nsdebug::Debug::pMain = &oDebugMain;

//static int iDebugLevel = 0;

/** Adds a new rule to the debug rule list
  * @param sRule the rule to be added
  * You can append new rules to the accept and deny rule list.
  * To add a accept rule prepend "+" or "" to the string.
  * To add a remove rule prepend a "-" to the string.
  */
namespace nsdebug {
void Debug::addRule(const char *sRule, int iRuleLevel)
{
	Rule oRule;

	oRule.sClass = sRule;

	/* 1. remove the ending '.' and '*' signs */
	while(oRule.sClass.endsWith("*") ||
	      oRule.sClass.endsWith("."))
			oRule.sClass.remove(oRule.sClass.length()-1, 1);

	if ((oRule.sClass.at(0).isNumber()) && (iRuleLevel == -1))
	{
		oRule.iLevel = (oRule.sClass.left(1).latin1()[0] - '0');
		oRule.sClass.remove(0, 1);
		if (oRule.sClass.at(0) == ':')
			oRule.sClass.remove(0, 1);
	}

	if (oRule.sClass[0] == '-')
		oRule.iLevel = 0;
	if (oRule.sClass[0] == '+')
		oRule.iLevel = 9;
	if ((oRule.sClass[0] == '+') || (oRule.sClass[0] == '-'))
		oRule.sClass.remove(0, 1);

	oVectRules.push_back(oRule);

	applyRules();
}

void Debug::removeRule(int iRule)
{
	oVectRules.erase(oVectRules.begin() + iRule);
}
}

namespace nsdebug {
Debug::Debug(void) : oTreeRoot("", NULL)
{
	iMissed = 0;
}

Debug::~Debug()
{
	if (iMissed)
	{
		printf("%d message(s) were missed.\n", iMissed);
		iMissed = 0;
	}
}

Class::Class(const char *sName, Class *pParent)
{
	/* 1. get our short name */
	this->sName = sName;

	/* 2. get the parent's full name, and the depth of the parent  */
	if (pParent != NULL)
	{
		sFullName = pParent->sFullName;
		iDepth = pParent->iDepth + 1;
	} else
	{
		sFullName = "";
		iDepth = 0;
	}

	/* 3. parent's full name + "." + our short name -> our long name */
	if (sFullName.length() > 0)
		sFullName.append(".");
	if (pParent) sFullName.append(sName);

}

Class::~Class()
{
	/* 1. destruct all of the childs */
	while(oVectChilds.size())
	{
		/* 1.1 get the first of the remaining childs */
		Class *pItem = oVectChilds.back();
		/* 1.2 remove item '1.1' from the child list */
		oVectChilds.pop_back();
		/* 1.3 destruct item '1.1' */
		delete(pItem);
	}
}

#if 0
/**
 * Inserts an item to the tree
 * @param sName the name of the new item
 * @return a pointer to the item whose name is 'sName'.
 *         If the name was inserted before the pointer will point to a  previously created item.
 */
Class *Debug::insertItem(char *sName)
{
	QString sIName = sName;
	Class *pRETURN = NULL, *pCurrent;

	/* 1. begin the search from the root */
	pCurrent = &oTreeRoot;

	/* 2. ITER go through the sections of the name to search */
	for(int iSection = 0; true; iSection++)
	{
		const QString &sSection = sIName.section('.', iSection, iSection);
		/* 2.1 COND is this is a null string, search for the tree element */
		if ((sSection != QString::null) && (sSection != ""))
		{
			int iSubClass;
			/* 2.1.t.1 search for the name in the childs of the current tree item */
			for(iSubClass = 0; iSubClass < (int)pCurrent->oVectChilds.size(); iSubClass++)
			{
				if (pCurrent->oVectChilds[iSubClass]->sName == sSection)
				{
					pCurrent = pCurrent->oVectChilds[iSubClass];
					break;
				}
			}

			/* 2.1.t.2 didn't found the name -> we create an item for it */
			if (iSubClass >= (int)pCurrent->oVectChilds.size())
			{
				Class *pNewClass;
				pCurrent->oVectChilds.push_back(pNewClass = (new Class(sSection.latin1(), pCurrent)));
				pCurrent = pNewClass;
			}
		} else 
		{
			/* 2.1.f COND if this is a null string we found the element */
			if (sSection == "")
			{
				pRETURN = pCurrent;
				break;
			}
		}
	}

	return pRETURN;
}
#endif

Class *Debug::getClass(const QString &sName, int *pLevel, bool bCreate)
{
	QString sQName = sName;
	Class *pRETURN = NULL;
	bool bFound;

	Class *pCurrent = &oTreeRoot;

	if ((pLevel != NULL) && (sQName.at(0).isNumber()))
	{
		*pLevel = (sQName.left(1).latin1()[0] - '0');
		sQName.remove(0, 1);
		if (sQName.at(0) == ':')
			sQName.remove(0, 1);
	}

	/* 2. ITER go through the sections of the name to search */
	for(int iSection = 0; true; iSection++)
	{
		const QString &sSection = sQName.section('.', iSection, iSection);
		/* 2.1 COND if this is a null string then we are at the end of the search */
		if ((sSection != QString::null) && (sSection != ""))
		{
			/* 2.1.t null string -> search for the tree element */
			int iSubClass;
			/* 2.1.t.1 search for the name in the childs of the current tree item */
			for(iSubClass = 0, bFound = false; iSubClass < (int)pCurrent->oVectChilds.size(); iSubClass++)
			{
				if (pCurrent->oVectChilds[iSubClass]->sName == sSection)
				{
					pCurrent = pCurrent->oVectChilds[iSubClass];
					bFound = true;
					break;
				}
			}

			/* 2.1.t.2 didn't found the name -> we create an item for it */
			if (!bFound)
			{
				if (bCreate)
				{
					Class *pNewClass;
					pCurrent->oVectChilds.push_back(pNewClass = (new Class(sSection.latin1(), pCurrent)));
					pCurrent = pNewClass;
					applyRules();
				} else
				{
					pRETURN = NULL;
					break;
				}
			}
		} else 
		{
			/* 2.1.f COND if this is a null string we found the element */
			if (sSection == "")
			{
				pRETURN = pCurrent;
				break;
			}
		}
	}

	return pRETURN;
}

void Debug::error(Class *pClass, int iLevel, char *sFormat, ...)
{
	if (iMissed)
	{
		printf("%d message(s) were missed.\n", iMissed);
		iMissed = 0;
	}

#ifdef DEBUG_DYNAMIC
	if (iDebugLevel >= 3)
#endif
	{
		va_list argp;

		fprintf(stderr, "ERROR (%s): ", pClass->getFullName()->latin1());
		va_start(argp, sFormat);
		vfprintf(stderr, sFormat, argp);
		va_end(argp);
		fprintf(stderr, "\n");
	}
}

void Debug::warning(Class *pClass, int iLevel, char *sFormat, ...)
{
	if (iMissed)
	{
		printf("%d message(s) were missed.\n", iMissed);
		iMissed = 0;
	}

#ifdef DEBUG_DYNAMIC
	if (iDebugLevel >= 3)
#endif
	{
		va_list argp;

		fprintf(stderr, "warning (%s): ", pClass->getFullName()->latin1());
		va_start(argp, sFormat);
		vfprintf(stderr, sFormat, argp);
		va_end(argp);
		fprintf(stderr, "\n");
	}
}

void Debug::message(Class *pClass, int iLevel, char *sFormat, ...)
{
	if (iMissed)
	{
		printf("%d message(s) were missed.\n", iMissed);
		iMissed = 0;
	}

#ifdef DEBUG_DYNAMIC
	if (iDebugLevel >= 3)
#endif
	{
		va_list argp;

		fprintf(stderr, "info (%s): ", pClass->getFullName()->latin1());
		va_start(argp, sFormat);
		vfprintf(stderr, sFormat, argp);
		va_end(argp);
		fprintf(stderr, "\n");
	}
}

void Debug::print(Class *pClass, int iLevel, char *sFormat, ...)
{
	if (iMissed)
	{
		printf("%d message(s) were missed.\n", iMissed);
		iMissed = 0;
	}

#ifdef DEBUG_DYNAMIC
	if (iDebugLevel >= 3)
#endif
	{
		va_list argp;

		va_start(argp, sFormat);
		vfprintf(stderr, sFormat, argp);
		va_end(argp);
	}
}

void Debug::applyRules(void)
{
	oTreeRoot.setLevelOfSubTree(9);

	/* 1.ITER go through all of the rules */
	for(int iRule = 0; iRule < (int)oVectRules.size(); iRule++)
	{
		/* 1.2 remove the unnecessary '*' and '.' signs. e.g "debug.*" -> "debug" */
		Class *pClass = getClass(oVectRules[iRule].sClass.latin1());
		if (pClass != NULL)
		{
			pClass->setLevelOfSubTree(oVectRules[iRule].iLevel);
		}
	}
}

void Class::setLevelOfSubTree(int iNewLevel)
{
	this->iLevel = iNewLevel;

	for(int iIdx = 0; iIdx < (int)oVectChilds.size(); iIdx++)
	{
		//of (oMew:oom=EST FK
		oVectChilds[iIdx]->setLevelOfSubTree(iNewLevel);
	}
}

/**
 * Print the list of the rules
 */
void Debug::printRules(void)
{
	/* 1f. child item -> display the item */
	for(int iN = 0; iN < oVectRules.size(); iN++)
	{
		int iLevel = oVectRules[iN].iLevel;
		printf("%d. (%c) %s\n", iN,  (iLevel == 0 ? '-' : (iLevel == 9 ? '+' : iLevel+'0')),
		        oVectRules[iN].sClass.latin1());
	}
}

/**
 * Print the structure of the tree
 * @param pRoot the parameter which is needed for the recursion
 */
void Debug::printTree(Class *pRoot)
{
	/* 1.COND are we at the root item? */
	/* 1t. root item -> display "(ROOT)" */
	if (pRoot == NULL)
	{
		printf("Structure of the debug class tree:\n(ROOT)\n");
		pRoot = &oTreeRoot;
	} else
	/* 1f. child item -> display the item */
	{
		for(int iN = 0; iN < pRoot->iDepth; iN++)
			printf("    ");
		printf("(%c)%s\n", (pRoot->iLevel!=0 ?
		      (pRoot->iLevel!=9 ?pRoot->iLevel+'0' :'+') :'-'), pRoot->sName.latin1());
	}

	/* 2.ITER go through all of the child items */
	for(int iSubClass = 0; iSubClass < (int)pRoot->oVectChilds.size(); iSubClass++)
	{
		printTree(pRoot->oVectChilds[iSubClass]);
	}
}
}

/**
 * Utility to check if a unit is allowed to produce debug output
 */
#if 0
inline int debug_isUnitAllowed(char *sUnitName)
{
	int bALLOWED = 0;
	int iN, iChN;
	char cRuleCh;

	/* 1. Go through all of the accept rules */
	for(iN = 0; iN < iAcceptRulesN; iN++)
	{
		/* 1.1 Examine the 'iN'th accept rule - do we fit to this rule?
		       ITER check this rule from character to character  */
		for(iChN = 0; 1; iChN ++)
		{
			cRuleCh = oAcceptRules[iN]->sRuleName[iChN];
			/* 1.1.1 COND if we are at the end of a rule - we fit to this rule
			         e.g "config.save" fits to rule "config.*"
			                                                ^ cRuleCh - '*' */
			if ((cRuleCh == '\0') || (cRuleCh == '*'))
			{
				bALLOWED = 1;
				break;
			}
			/* 1.1.2 COND if the target class differs from the rule - we don't fit to this rule
			         e.g "config.save" doesn't fit to rule "config.load" 
			                     ^ sUnitName[iChN] - 's'           ^ cRuleCh - 'l' */
			if (cRuleCh != sUnitName[iChN]) break;
		}
		if (bALLOWED)
			break;
	}

	/* 2. Go through all of the _deny_ rules (in case the unit seems to be acceptable) */
	if (bALLOWED)
	for(iN = 0; iN < iDenyRulesN; iN++)
	{
		/* 2.1 Examine the 'iN'th accept rule - do we fit to this rule?
		       ITER check this rule from character to character  */
		for(iChN = 0; 1; iChN ++)
		{
			cRuleCh = oDenyRules[iN]->sRuleName[iChN];
			/* 2.1.1 COND if we are at the end of a rule - we fit to this rule
			         e.g "config.save" fits to rule "config.*"
			                                                ^ cRuleCh - '*' */
			if ((cRuleCh == '\0') || (cRuleCh == '*'))
			{
				bALLOWED = 0;
				break;
			}
			/* 2.1.2 COND if the target class differs from the rule - we don't fit to this rule
			         e.g "config.save" doesn't fit to rule "config.load" 
			                     ^ sUnitName[iChN] - 's'           ^ cRuleCh - 'l' */
			if (cRuleCh != sUnitName[iChN]) break;
		}
		if (!bALLOWED)
			break;
	}

	/* 3. return with the result */
	return bALLOWED;
}

void debug_message(Class *pClass, char *sClass, char *sFormat, ...)
{
#ifdef DEBUG_DYNAMIC
	if (iDebugLevel >= 3)
#endif
	if (debug_isUnitAllowed(sClass))
	{
		va_list argp;

		fprintf(stderr, "error: ");
		va_start(argp, sFormat);
		vfprintf(stderr, sFormat, argp);
		va_end(argp);
		fprintf(stderr, "\n");
	}
}

void debug_print(char *sClass, char *sFormat, ...)
{
//	debug_message(
}
#endif

