/*
 * EditBufferFind.cpp
 *
 * Searching for a string in the current buffer
 *
 * Author: Ferenc Moricz
 * $Revision: 1.24 $
 * $Date: 2003/09/06 23:32:53 $
 *
 */

#include "EditBufferFind.h"

/* the text of the last search */
static QString *lastsearch = NULL;

bool EditBufferFind::eFindBottom = true;

#include <qpushbutton.h>
#include <qtooltip.h>
EditBufferFind::EditBufferFind( EditBufferText *buffer, QWidget *pParent ) 
				: QHBox( (QWidget *)pParent )
{
	this->pParent = pParent;
	this->buffer = buffer;
	new QLabel( tr("Find: "), this);
	pDialog = NULL;
	bFound = false;

	/* '^' button */
	pUpButton = new QPushButton("^", this, "^");
	QObject::connect( pUpButton, SIGNAL(clicked( void )),
				this, SLOT( upButtonEvent( void ) ) );
	QToolTip::add(pUpButton, "Switch to window mode");

	/* find editline */
	pSearchText=new QLineEdit( this );
	pSearchText->installEventFilter(this);

	/* '>' button */
	pMoreButton = new QPushButton(">", this, ">");
	QObject::connect( pMoreButton, SIGNAL(clicked( void )),
				this, SLOT( moreButtonEvent( void ) ) );

	pReplaceButton = NULL;
	pReplaceText = NULL;

	setMinimumWidth(320);
//	setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Preferred);
	pSearchText->setMinimumWidth(100);
	pSearchText->setMaximumWidth(180);

	palette = new QPalette( pSearchText->palette() );
	paletteerror = new QPalette( *palette );
	QColorGroup *qg = new QColorGroup( palette->active() );
	qg->setColor( QColorGroup::Base, red );
	paletteerror->setActive( *qg );
	paletteerror->setInactive( *qg );

	QObject::connect( pSearchText, SIGNAL(textChanged(const QString &)), 
					  this, SLOT( findCB(const QString &) ) );

	gotoPlace(eFindBottom);
	show();

	selectB=buffer->select[buffer->selectsmaller]; selectE=buffer->select[1-buffer->selectsmaller];
	selectcolor = buffer->selectcolor;
	firstVisible = buffer->firstVisible;
	cursor = *new QPoint( buffer->cursor );
	buffer->selectcolor=buffer->properties->getColorValue(
		"buffer.selection.findcolor", &green );

	if ( (buffer->select[0].y() != -1) && (buffer->select[0].y() == buffer->select[1].y()) )
	{
		EditLine *pLine = buffer->lines[ buffer->select[0].y() ];
		if (pLine != NULL)
		{
			int iFirst = pLine->tabbedPosition( buffer->select[buffer->selectsmaller].x(),
			                                    buffer->up->tabsize);
			int iLast = pLine->tabbedPosition( buffer->select[1 - buffer->selectsmaller].x(),
			                                    buffer->up->tabsize);
			bNoSearch = true;
			pSearchText->setText( pLine->mid( iFirst, iLast - iFirst ).latin1() );
			pSearchText->selectAll();
			bNoSearch = false;
		}
	} else
	if (lastsearch!=NULL)
	{
		bNoSearch = true;
		pSearchText->setText( *lastsearch );
		pSearchText->selectAll();
		bNoSearch = false;
	}
	buffer->select[0].setY(-1); 
	buffer->select[1].setY(-1);
	buffer->selectsmaller = 0;
	pSearchText->setFocus();
}

EditBufferFind::~EditBufferFind()
{
	buffer->up->cmd_find(0);
	pSearchText->removeEventFilter(this);
	if (buffer)
	{
		buffer->up->cmd_scrollToCursor();
		buffer->repaint(false);
		buffer->up->findbox = NULL;
		buffer->askForFocus(NULL);
		buffer->setFocus();
	}
}

void EditBufferFind::findClose( void )
{
	buffer->up->cmd_find(0);
	if (pDialog)
	{
		pDialog->close(true);
		pDialog = NULL;
	}
	else close( true );
}

bool EditBufferFind::returnEvent()
{
	bool bRet = false;
	if (!pDialog)
	{
		findClose();
		bRet = true;
	}
	return bRet;
}

/*
 * .EditBufferFind::eventFilter(..)
 *
 * Event filter for the edit line widgets; it is used to catch
 * special text events (e.g enter, escape, pageup, pagedown)
 * during text editing.
 *
 */ 
bool EditBufferFind::eventFilter( QObject * pWatchedObject, QEvent *pEvent )
{
	if (pWatchedObject == pSearchText)
	{
		if (pEvent->type() == QEvent::KeyPress)
		{
			QKeyEvent oKeyEvent = *(QKeyEvent *)pEvent;
			switch( oKeyEvent.key() )
			{
				case Qt::Key_Escape:
						return escEvent();
				case Qt::Key_Enter:
				case Qt::Key_Return:
						return returnEvent();
				case Qt::Key_PageUp:
						pageEvent(0);
						break;
				case Qt::Key_PageDown:
						pageEvent(1);
						break;
			}
		} else
		if ((pEvent->type() == QEvent::FocusOut) && (!bNoSearch))
		{
			if ((!pReplaceText) || (!pReplaceText->hasFocus()))
			{
				if (!pDialog)
				{
					findClose();
					return true;
				}
			}
		}
	}
	return false;
}

bool EditBufferFind::escEvent()
{
	buffer->select[0]=selectB;
	buffer->select[1]=selectE;
	buffer->selectcolor=selectcolor;
	buffer->cursor = cursor; 
	buffer->firstVisible=firstVisible;

	findClose();               /* close window */

	return true;
}

void EditBufferFind::pageEvent( int direction )
{
	/*r purge?? - was it unnecessary ? */
	int x = buffer->select[0].x();
	if ( x < 0 ) return;
	if (!direction) {
		/* we are at the 0th column -> go to the end of the previous line */
		if (--x == -1) x--;
		if ( x < 0 )
			buffer->cursor.setY( buffer->cursor.y()-1 );
	}
	else x++;
	buffer->select[0].setX(x);

//#if defined(_SYSTEM_LINUX_)
#if 0
	search( pSearchText->text(), direction);
#else
	find( pSearchText->text(), direction);
#endif
}

void EditBufferFind::findCB( const QString &str )
{
	if ( lastsearch==NULL ) lastsearch = new QString(); 
	*lastsearch = str;
	if (bNoSearch) return;
//#if defined(_SYSTEM_LINUX_)
#if 0
	search( str, 1 );
#else
	find( str, 1 );
#endif
}

#include "Editor.h"
void EditBufferFind::gotoPlace( bool bGotoBottom )
{
	bNoSearch = true;
	if (bGotoBottom)
	{
		reparent(pParent, 0, QPoint(0, 0), true );
		pUpButton->setText("^");
		buffer->askForFocus(pSearchText);
		if (pDialog)
		{
			buffer->up->cmd_find(0);
			pDialog->close(true);
			pDialog = NULL;
		}
	} else
	{
		pDialog = new QDialog(Editor::editor->pMainWindow, "Find", true, WDestructiveClose);
		reparent( pDialog, 0, QPoint(0, 0), true );
		pDialog->show();
		pUpButton->setText("\\/");
		pSearchText->setFocus();
	}

	eFindBottom = bGotoBottom;

	bNoSearch = false;
}

void EditBufferFind::upButtonEvent( void )
{
	if (eFindBottom)
		gotoPlace(0);
	else gotoPlace(1);
}

void EditBufferFind::moreButtonEvent( void )
{
	if (!pReplaceText)
	{
		pMoreButton->setText("<");
		pReplaceButton = new QPushButton("&Replace", this, "&Replace");
		QObject::connect( pReplaceButton, SIGNAL(clicked( void )),
				this, SLOT( replaceButtonEvent( void ) ) );
		pReplaceButton->show();
		pReplaceText = new QLineEdit(this);
		pReplaceText->show();
	} else
	{
		pMoreButton->setText(">");
		pReplaceButton->close(true);
		pReplaceButton = NULL;
		pReplaceText->close(true);
		pReplaceText = NULL;
	}
}

void EditBufferFind::replaceButtonEvent( void )
{
	if (bFound)
	{
		EditLine *l = buffer->lines[ buffer->select[0].y() ];
		l->replace(l->tabbedPosition( buffer->select[0].x(), buffer->up->tabsize ),
			pSearchText->text().length(), pReplaceText->text());
		buffer->repaintfrom=buffer->repaintto=-1;
		buffer->repaint( false );
		pageEvent(1);
	}
}

void EditBufferFind::find( const QString &str, int direction)
{
	bool wrapped=FALSE;

	int x = buffer->select[0].x();
	if ( x == -1 ) x = buffer->cursor.x();
	if ( x == -2 ) x++;
	for( int i=buffer->cursor.y(); true; i+=(direction ? 1:-1) ) {
		if ( i == -1 ) {
			i = buffer->lines.size()-1;
			wrapped = true;
		}
		else if ( i == (int)buffer->lines.size() ) {
			i = 0;
			wrapped = true;
		}
		if (( i == buffer->cursor.y() ) && wrapped ) {
			pSearchText->setPalette( *paletteerror );
			break;
		} 
		EditLine *l = buffer->lines[ i ];
		if (x>=0) x = l->tabbedPosition( x, buffer->up->tabsize, false );
		int found;
		if (direction) {
#if QT_VERSION > 280
			found = l->find( str, x);
#else
			found = l->find( str, x);
#endif
			x=0;
		} else {
#if QT_VERSION > 280
			found = l->findRev( str, x);
#else
			found = l->findRev( str, x);
//			found = l->searchRev( str, x);
#endif
			x=-1;
		}
		if ( found > -1 )
		{
			bFound = true;
			pSearchText->setPalette( *palette );
			buffer->cursor.setX(
				l->nontabbedPosition(found+str.length(), buffer->up->tabsize) );
			buffer->cursor.setY( i );
			buffer->select[0].setX( l->nontabbedPosition( found, buffer->up->tabsize ) );
			buffer->select[0].setY( i );
			buffer->select[1].setX( buffer->cursor.x() );
			buffer->select[1].setY( i );
			buffer->up->cmd_scrollToCursor();
			buffer->repaintfrom=buffer->repaintto=-1;
			buffer->repaint( false );
			break;
		} else bFound = false;
	}
}

