/*
 * EditBufferText.cpp
 *
 * The container of the lines
 *
 * Author: Ferenc Moricz
 * $Revision: 1.125 $
 * $Date: 2004/02/22 00:12:49 $
 *
 */

#include "EditBuffer.h"
#include "Editor.h"

#include <qpainter.h>
#include <qcursor.h>

#include <qwhatsthis.h>

#include <stdlib.h>
#include <stdio.h>

#define WOFFSET 5 

//extern vector<Module *, allocator<Module *> > aFilterModules;

EditProperties *EditBufferText::properties;
EditLineSyntax *EditBufferText::syntax = NULL;

QString EditBufferText::sSelectionColorName = "buffer.selection.color";
QString EditBufferText::sSelectionCopiedColorName = "buffer.selection.copiedcolor";
QString EditBufferText::sSelectionPastedColorName = "buffer.selection.pastedcolor";

QPoint mousepoint(0, 0);

void FreddyCursor::setY( int iY )
{
	if (Editor::editor->moduleloader)
		Editor::editor->moduleloader->broadcastEvent( "event.cursor", NULL, NULL, -1, iY);
	((QPoint *)this)->setY(iY);
}

EditBufferText::EditBufferText( EditBuffer *parent, QWidget *frame )
                : QWidget( frame, "", WResizeNoErase )
{
	up = parent;

	cursor = *new QPoint( 0, 0 );

	select[0].setX( -1 );
	select[0].setY( -1 );
	select[1].setX( -1 );
	select[1].setY( -1 );
	selectsmaller = 0;
	selecting = mouseselecting = FALSE;
	pAskedForFocus = NULL;
	sCurrentSelectionColorName = &sSelectionColorName;

	firstVisible = nLines = 0;
	cursor.setX( 0 );
	cursor.setY( 0 );

	if (syntax == NULL) syntax = new EditLineSyntax();
	pFilterID = NULL;
	bUseColorer = true;

	this->setCursor( QCursor( IbeamCursor ) );

	iMaxLineLength = 0;
	fontheight = 0;

	reconfigure();

	if (dbuf) pm.resize( size().width(), fontheight );

	repaintfrom=0; repaintto=-1;
	mousetimer = pMouseClickTimer = NULL;

	setFocusPolicy( StrongFocus );
	setAcceptDrops( true );
}

void EditBufferText::reconfigure()
{
	textpen = properties->getColorValue( "buffer.text.color", &black);
	tabpen = properties->getColorValue( "buffer.tab.color", &white);
	keypen = properties->getColorValue( "buffer.text.key.color", &blue );
	stringpen = properties->getColorValue( "buffer.text.string.color", &green );
	commentpen=properties->getColorValue( "buffer.text.comment.color", &blue );
	preprocpen=properties->getColorValue( "buffer.text.preprocessor.color", &green );
	bg = properties->getColorValue( "buffer.background.color", &gray);
	textbg = properties->getColorValue( "buffer.text.background.color", &lightGray );
	if (textbg != NULL) setBackgroundColor( *textbg );
	cursorpen = properties->getColorValue("buffer.cursor.color", &red);
	style = properties->getValue( "buffer.selection.style", new QString("lineend"));

	dbuf = properties->getBoolValue( "buffer.doublebuffering", false);

	if (dbuf) 
		pm.resize( size().width(), fontheight );

	bUseColorer = properties->getBoolValue( "module.colorer.enabled", true );

	QFont oFntDef("Fixed", 12);
	QFont fnt = *(properties->getFontValue("buffer.font", &oFntDef));
	fnt.setFixedPitch( true );
	setFont( fnt );
	QFontMetrics *fm = new QFontMetrics( font() );
	fontwidth = fm->width('M');
	fontheight = fm->height();
	fontheight += properties->getNumericalValue("buffer.font.gap", 2);
	fontascent = fm->ascent();

	iMouseClick = 0;

	delete(fm);

	nLines = size().height() / fontheight;
}

void EditBufferText::askForFocus(QWidget *pWidget)
{
	pAskedForFocus = pWidget;
}

/*
 * Events
 */

void EditBufferText::resizeEvent( QResizeEvent *e_resize )
{
	fflush(stdout);
	nLines = e_resize->size().height() / fontheight;
	//? why does it cause a crash when vscroll disappears during resizing?
	// up->cmd_scrollBar();
	if (dbuf) 
		pm.resize( size().width(), fontheight );
}

#if QT_VERSION > 280
#include <qpopupmenu.h>
void EditBufferText::contextMenuEvent( QContextMenuEvent *pEvent )
{
	if (!Editor::editor->pContextMenu) return;
	Editor::editor->pContextMenu->exec(QCursor::pos());
	pEvent->accept();
	return;
}
#endif

QPoint oMouseLastClick, oThisPos;
void EditBufferText::mousePressEvent( QMouseEvent *e )
{
	bool selectdisappear = false, bPosition = false;

	if (e->button() & QEvent::RightButton) return;
	if (e->button() & QEvent::MidButton) return;        /*? should be optional? */

	oThisPos.setX( (e->x()-5)/fontwidth + up->hscrollValue );
	oThisPos.setY( e->y()/fontheight+firstVisible );

	if (oMouseLastClick != oThisPos)
	{
		iMouseClick = 0;
		oMouseLastClick = oThisPos;
		bPosition = true;
	}
	if (iMouseClick == 0) bPosition = true;

	if (bPosition)
	{
//		printf("select[0]: %d %d\n", select[0].x(), select[0].y());
//		printf("select[1]: %d %d\n", select[1].x(), select[1].y());
		if (select[0]!=select[1]) selectdisappear = true;
		select[0].setX( oThisPos.x() );
		select[0].setY( oThisPos.y() );
		select[1].setX( oThisPos.x() );
		select[1].setY( oThisPos.y() );
//		repaint(false);
	}

	up->oUndo.doAtomicOperation(AtomicTextOperation::SETSELECTION, sSelectionColorName,
		cursor.x(), cursor.y(), cursor.x(), cursor.y());
	mouseselecting = 1;
	up->cmd_setCursor( e->x(), e->y() );
	grabMouse();

	/* count mouse clicks */
	if (pMouseClickTimer == NULL)
	{
		pMouseClickTimer = new QTimer( this, "mouseclicktimer" );
		QObject::connect( pMouseClickTimer, SIGNAL(timeout( void )),
							this, SLOT( mouseClickTimeout( void ) ) );
		iMouseClick = 0;
		pMouseClickTimer->start( 200 );
	} else pMouseClickTimer->start( 200 );
	if (!(iMouseClick % 2)) iMouseClick ++;

	if (iMouseClick == 3)
	{                                  /* double click */
		mouseDoubleClickEvent(e);
	} else                             /* triple click */
	if (iMouseClick == 5)
	{
		int iY = e->y()/fontheight+firstVisible;
		EditLine * s = lines[iY];
		if (s == NULL) return;
		select[0].setX( 0 );
		select[0].setY( iY );
		select[1].setX( s->nontabbedPosition(s->length()+1, up->tabsize) );
		select[1].setY( iY );
		selectsmaller = 0;
		mouseselecting = 0;
		repaint();
	} else
	if (iMouseClick == 7)              /* quadro click */
	{
		EditLine * s = lines[ lines.size() - 1 ];
		if (s == NULL) return;
		select[0].setX( 0 );
		select[0].setY( 0 );
		select[1].setX( s->nontabbedPosition(s->length()+1, up->tabsize) );
		select[1].setY( lines.size() - 1 );
		selectsmaller = 0;
//		repaintfrom = repaintto = -1;
		mouseselecting = 0;
		repaint();
	}
}

void EditBufferText::mouseReleaseEvent( QMouseEvent *e )
{
	mouseselecting = 0;
	releaseMouse();
	if (mousetimer != NULL)
	{
		mousetimer->stop();
		delete(mousetimer);
		mousetimer = NULL;
	}
	/* paste clipboard-selection */
	if (e->button() & MidButton)
	{
		up->cmd_selection_paste( FALSE );
		repaintfrom = repaintto = -1;
		repaint();
	}

	if (iMouseClick % 2) iMouseClick ++;
}

void EditBufferText::mouseClickTimeout( void )
{
	iMouseClick = 0;
	oMouseLastClick.setY( 0 );
	pMouseClickTimer->stop();
	delete( pMouseClickTimer );
	pMouseClickTimer = NULL;
}

void EditBufferText::mouseDoubleClickEvent( QMouseEvent *e )
{
	EditLine * s = lines[select[0].y()];
	if (s == NULL) return;
	select[0].setX((e->x()-5)/fontwidth);
	select[0].setY(e->y()/fontheight+firstVisible);
	select[1].setX((e->x()-5)/fontwidth);
	select[1].setY(e->y()/fontheight+firstVisible);
	int x1, x2;
	x1 = x2 = s->tabbedPosition( select[0].x(), up->tabsize );
	if (x1 >= (int)s->length()) return;
//! mixed string
//	while((s->at(x1)!=' ') && (s->at(x1)!='\t') && (x1 > -1)) x1--; x1++;
//	while((s->at(x2)!=' ') && (s->at(x2)!='\t') && (x2 < (int)s->length())) x2++;
	while((s->at(x1).isLetter()) || (s->at(x1).isNumber()) && (x1 > -1)) x1--; x1++;
	while((s->at(x2).isLetter()) || (s->at(x2).isNumber()) && (x2 < (int)s->length())) x2++;
	select[0].setX(s->nontabbedPosition(x1, up->tabsize));
	select[1].setX(s->nontabbedPosition(x2, up->tabsize));
	selectsmaller = 0;
	mouseselecting = 0;
	repaint();
/* WIN	e->accept();*/
}

void EditBufferText::mouseMoveEvent( QMouseEvent *e )
{
	if (mouseselecting)
	{
		mousepoint.setX(e->x());
		mousepoint.setY(e->y());
		if ((mousepoint.y() >= 0) || (mousepoint.y() <= height()))
		{
			if ( (select[0].y() < (int)lines.size()) &&
			     ( e->y() / fontheight < (int)lines.size()) )
			{
				select[1].setX( e->x() / fontwidth );
				select[1].setY( e->y() / fontheight + firstVisible );
				up->cmd_setCursor( e->x(), e->y() );
				if (mousetimer == NULL)
				{
					mousetimer = new QTimer( this, "mousetimer" );
					QObject::connect( mousetimer, SIGNAL(timeout( void )),
							                this, SLOT( mouseTimeout( void ) ) );
					mousetimer->start( 100 );
				}
			}
		}
	}
}

void EditBufferText::mouseTimeout( )
{
	if ((mousepoint.y() < 0) || (mousepoint.y() > height()))
	{
		int y = (mousepoint.y()<0? -mousepoint.y() : height()-mousepoint.y());

		float scroll =  y / (float)fontheight;  /* biadic scroll parameter */
		scroll = scroll*scroll + 1;
		if (mousepoint.y() < 0) y = cursor.y() - (int)scroll;
		else y = cursor.y() + (int)scroll;
		y = max( 0, y);
		y = min( y, (int)lines.size()-1 );

		cursor.setY(y);
		select[1].setX( (mousepoint.x()-5)/fontwidth );
		select[1].setY( y );
		up->cmd_scrollToCursor(-1, false);
	}
}

/*
 * EditBufferText::focusInEvent(.)
 * This function is used to pass the focus to the widget
 * which asked for it.
 */
void EditBufferText::focusInEvent ( QFocusEvent * )
{
	if (pAskedForFocus) pAskedForFocus->setFocus();
	pAskedForFocus = NULL;
}


void EditBufferText::focusOutEvent( QFocusEvent *e_focus )
{
}

static int moduletoexecute;
#include <qapplication.h>
#include <qclipboard.h>
bool EditBufferText::event( QEvent *e )
{
//	dmesg("buffer", "Text EVENT: EditBuffer type:%d", e->type()); 
    
	QKeyEvent * ke = (QKeyEvent *) e;

	if ( (e->type() == QEvent::MouseButtonDblClick) ||
	     (e->type() == QEvent::MouseButtonPress))
	{
		QMouseEvent *pMouseEvent = (QMouseEvent *) e;
		mousePressEvent( pMouseEvent );
#if QT_VERSION > 280
		pMouseEvent->accept();
#endif
	}

	/* Wheel scroll event */
	if ( (e->type() == QEvent::Wheel) && (up->vscroll!=NULL) )
	{
		QWheelEvent *pWheelEvent = (QWheelEvent *)e;

		/* |  VERTICAL mouse wheel event */
#if QT_VERSION > 280
		if (pWheelEvent->orientation() == Qt::Vertical)
#endif
		{
#ifdef _SYSTEM_WINDOWS_
			QWheelEvent *pWheel = (QWheelEvent *)e;
			QWheelEvent *pWheelNew = new QWheelEvent( pWheel->pos(),
				pWheel->delta()/3, pWheel->state());
	//r		printf("wheel: %d\n", pWheel->delta()); 
			QApplication::sendEvent( up->vscroll, pWheelNew );
#else
			QApplication::sendEvent( up->vscroll, e );
#endif
		}
		/* <-> HORIZONTAL mouse wheel event */
#if QT_VERSION > 280
		else if (pWheelEvent->orientation() == Qt::Horizontal)
		{
			QApplication::sendEvent( up->hscroll, e);
		}
#endif
	}

	if ( e->type() == QEvent::Paint ) up->cmd_checkFile(); // modified, reload?

	if (e->type() == QEvent::FocusOut) return false;

	if ((e->type() == QEvent::KeyPress) || (e->type() == QEvent::KeyRelease))
	switch ( ke->key() ) {
		case Key_Tab:
			if (ke->type() == QEvent::KeyPress ) keyPressEvent( ke );
			ke->accept();
			return TRUE;
		case Key_Shift:
			if (ke->type() == QEvent::KeyRelease) {
				selecting = false;
			}
			break;
		case Key_1:
		case Key_2:
		case Key_3:
		case Key_4:
		case Key_5:
		case Key_6:
		case Key_7:
		case Key_8:
		case Key_9:
			if (ke->type() != QEvent::KeyPress) break;
			if (ke->state() & ControlButton)
			{
				//r printf("modcont: %d\n", up->module->modules.count());

				moduletoexecute = min(ke->key() - Key_1, (int)(up->module->modules.size())-1);
				if (moduletoexecute >= 0)
				{
					dmesg("5:plugin", "Module to execute: %d - %s\n",
				             moduletoexecute, (up->module->modules)[moduletoexecute]->qsName->latin1());
				}
				else derr("5:plugin", "No module to execute!\n");
				return true;
			}
			else if (ke->state() & AltButton)
			{
				if ((int)up->module->modules.size() > moduletoexecute)
				{
					dmesg("5:plugin", "executing module %d\n", moduletoexecute);
					Module *pModule = (up->module->modules)[moduletoexecute];
					QString numstr(NULL);
					numstr.setNum( ke->key() - Key_1 + 1 );
					pModule->pFunction(pModule, numstr);
				}
				else dwarn("plugin", "No module to execute!");

				return true;
			}
			break;
		}

	return QWidget::event( e );
}

void EditBufferText::dragEnterEvent(QDragEnterEvent *e)
{
	QMimeSource *msrc = (QMimeSource *)e;
	printf("Mime format:%s\n", msrc->format());
	if (strstr(msrc->format(), "text") == msrc->format())
	printf("Text: %s\n", (const char *)msrc->encodedData("text/plain").data());
	e->accept();
}

void EditBufferText::dropEvent(QDropEvent *e)
{
	printf("Mime format:%s\n", e->format());
	if (strstr(e->format(), "text") == e->format())
	printf("Text: %s\n", (const char *)e->encodedData(e->format()).data());
}

void EditBufferText::keyPressEvent( QKeyEvent *e )
{
	int iPrevLength = lines.size(), iCurrLength, iCol;
	enum PM { REPAINT_NO = 0, REPAINT_FULL=1, REPAINT_PART=2, REPAINT_SCROLLTO=4 };
	int eRepaintMode = PM(REPAINT_NO);

	bool wasselected = false;

//	printf("Key: %x\n", e->key());
//	dmesg("6plugin" "Input Key: %x", e->key());

	/* 1. check if there is a shortcut handle for this event */
	if (Editor::editor->com->shortEvent( e ))
	{
		repaint(false);
		return;
	}

	/* 2. is this a new selection? */
	if ((e->state()&ShiftButton) && (!selecting) && (!mouseselecting) &&
		((e->key()==Key_Left) || (e->key()==Key_Right) ||
		(e->key()==Key_Up) || (e->key()==Key_Down) ||
		(e->key()==Key_Home) || (e->key()==Key_End) ||
		(e->key()==Key_PageUp) || (e->key()==Key_PageDown)))
	{
		/* 2.1 create new operation */
		up->oUndo.createOperation("NEW_SELECTION");
		up->oUndo.doAtomicOperation(AtomicTextOperation::SETSELECTION, *sCurrentSelectionColorName,
			select[0].x(), select[0].y(), select[1].x(), select[1].y());
		if (select[0].y()==-1)
			up->oUndo.doAtomicOperation(AtomicTextOperation::SETSELECTION, sSelectionColorName,
				cursor.x(), cursor.y(), select[1].x(), select[1].y());
		up->oUndo.doAtomicOperation(AtomicTextOperation::SETSELECTION, sSelectionColorName,
			select[0].x(), select[0].y(), cursor.x(), cursor.y());
		selecting = true;
		eRepaintMode |= PM(REPAINT_FULL);
		/* 2.2 close operation */
		up->oUndo.closeOperation();
	}

	/* 3. is this selecting period closed? */
	if ((!selecting)&&(!mouseselecting)&&(select[0].y()!=-1))
	{
		/* 3.1 create new operation */
		up->oUndo.createOperation("NEW_SELECTION_END");
		if ((e->key()==Key_Control) || (e->key()==Key_Alt)
		   || (e->key()==Key_Shift) || (e->key()==Key_Insert)
		   || (e->key()==Key_Tab) )
		{}
		else if (e->text().length()==0)
		{
			up->oUndo.doAtomicOperation(AtomicTextOperation::SETSELECTION, *sCurrentSelectionColorName,
				select[0].x(), select[0].y(), select[1].x(), select[1].y());
			up->oUndo.doAtomicOperation(AtomicTextOperation::SETSELECTION, sSelectionColorName,
				select[0].x(), -1, select[1].x(), -1);
			eRepaintMode |= PM(REPAINT_FULL);
		} else
		{
			up->cmd_selection_clear();
			wasselected = true;
		}
		/* 3.2 close operation */
		up->oUndo.closeOperation();
	}

	/* 4. keyboard event handler */
	switch (e->key())
	{

	  /*********/
	  /* LEFT  */
	  /*********/
	  case Key_Left:
		/* 1. create a new operation */
		up->oUndo.createOperation("KEY_MOVE_(LEFT)");

		if (selecting || mouseselecting)
			up->oUndo.doAtomicOperation(AtomicTextOperation::SETSELECTION, *sCurrentSelectionColorName,
				select[0].x(), select[0].y(), select[1].x(), select[1].y());

		/* 2.a is CTRL pressed? -> jumping from word to word */
		if (e->state() & QKeyEvent::ControlButton)
		{
			bool bScroll = false;
			int iCurrentX, iNextX;
			/* 2.a.1 LOOP: go until a beginning of a word is found */
			while(1)
			{
				/* 2.a.1.1 get the current line and current cursor position */
				EditLine *lineT=lines[ cursor.y() ];
				if (!bScroll) iCurrentX = lineT->tabbedPosition(cursor.x(), up->tabsize );
				else iCurrentX = -1;

				/* 2.a.1.2 get the current settings for the gotoword function */
				bool bWEnd = properties->getBoolValue("buffer.function.gotoword.wordendalso", true);
				bool bSplit = properties->getBoolValue("buffer.function.gotoword.splitatsign", true);

				/* 2.a.1.3 search for the previous 'beginning of a word' position in the line */
				if ((lineT != NULL) && ((iNextX = lineT->gotoWord(iCurrentX, false, bWEnd, bSplit))!= -1))
				{
					/* 2.a.1.3.a found previous word -> jump to this word */
					up->oUndo.doAtomicOperation(AtomicTextOperation::MOVECURSOR, QString::null,
						lineT->nontabbedPosition( iNextX, up->tabsize ), cursor.y(), cursor.x(), cursor.y() );
					break;
				}
				else 
				{
					/* 2.a.1.3.b not found previous word -> jump to previous line if exists */
					bScroll = true;
					up->oUndo.doAtomicOperation(AtomicTextOperation::MOVECURSOR, QString::null,
						lineT->nontabbedPosition( iNextX, up->tabsize ), cursor.y(), cursor.x(), cursor.y() );
					if (cursor.y() > 0) up->oUndo.doAtomicOperation(AtomicTextOperation::MOVECURSOR, QString::null,
						cursor.x(), cursor.y()-1, cursor.x(), cursor.y() );
					else break;
				}
			}
			/* repainting */
			if (!bScroll)
			{
				eRepaintMode |= PM(REPAINT_SCROLLTO);
				eRepaintMode |= PM(REPAINT_PART);
			} else eRepaintMode = PM(REPAINT_FULL);
		} else
		/* 2.b is CTRL not pressed? moving the cursor 1 character to the left */
		if (cursor.x() <= 0)
		{
			if ((properties->getBoolValue("buffer.edit.leftarrow_goesup", false))
			    && (cursor.y() > 0))
			{                           /* Key_End */
				cursor.setY(cursor.y() - 1);
				EditLine *lineT=lines[ cursor.y() ];
				if (lineT != NULL)
				{
					up->oUndo.doAtomicOperation(AtomicTextOperation::MOVECURSOR, QString::null,
						lineT->nontabbedPosition( lineT->length(), up->tabsize ), cursor.y(), cursor.x(), cursor.y());
					eRepaintMode |= PM(REPAINT_SCROLLTO);
					eRepaintMode |= PM(REPAINT_FULL);
				}
			}
		} else
		{
			up->oUndo.doAtomicOperation(AtomicTextOperation::MOVECURSOR, QString::null,
				cursor.x()-1, cursor.y(), cursor.x(), cursor.y());
			repaintfrom = repaintto = cursor.y();
			eRepaintMode |= PM(REPAINT_SCROLLTO);
			eRepaintMode |= PM(REPAINT_PART);
		}

		up->oUndo.closeOperation();
		break;

	  /**********/
	  /* RIGHT  */
	  /**********/
	  case Key_Right:
		up->oUndo.createOperation("KEY_MOVE_(RIGHT)");

		if (selecting || mouseselecting)
			up->oUndo.doAtomicOperation(AtomicTextOperation::SETSELECTION, *sCurrentSelectionColorName,
				select[0].x(), select[0].y(), select[1].x(), select[1].y());

		if (e->state() & QKeyEvent::ControlButton)
		{
			bool bScroll = false;
			int iCurrentX, iNextX;
			while(1)
			{
				/* get the current line and current cursor position */
				EditLine *lineT=lines[ cursor.y() ];
				if (!bScroll) iCurrentX = lineT->tabbedPosition(cursor.x(), up->tabsize );
				else iCurrentX = -1;

				/* get the current settings for the gotoword function */
				bool bWEnd = properties->getBoolValue("buffer.function.gotoword.wordendalso", true);
				bool bSplit = properties->getBoolValue("buffer.function.gotoword.splitatsign", true);

				if ((lineT != NULL) && ((iNextX = lineT->gotoWord(iCurrentX, true, bWEnd, bSplit))!= -1))
				{
					up->oUndo.doAtomicOperation(AtomicTextOperation::MOVECURSOR, QString::null,
						lineT->nontabbedPosition( iNextX, up->tabsize ), cursor.y(), cursor.x(), cursor.y());
					break;
				}
				else 
				{
					bScroll = true;
					if (cursor.y() < ((int)lines.size()-1))
						up->oUndo.doAtomicOperation(AtomicTextOperation::MOVECURSOR, QString::null,
							cursor.x(), cursor.y()+1, cursor.x(), cursor.y());
					else break;
				}
			}
			/* repainting */
			if (!bScroll)
			{
				eRepaintMode |= PM(REPAINT_SCROLLTO);
				eRepaintMode |= PM(REPAINT_PART);
			} else eRepaintMode = PM(REPAINT_FULL);
		} else
		{
			up->oUndo.doAtomicOperation(AtomicTextOperation::MOVECURSOR, QString::null,
				cursor.x()+1, cursor.y(), cursor.x(), cursor.y());
			repaintfrom = repaintto = cursor.y();
			eRepaintMode |= PM(REPAINT_SCROLLTO);
			eRepaintMode |= PM(REPAINT_PART);
		}
		up->oUndo.closeOperation();
		break;

	  /*******/
	  /* UP  */
	  /*******/
	  case Key_Up:
		if (cursor.y() <= 0) break;
		up->oUndo.createOperation("KEY_MOVE_(UP)");

		if (selecting || mouseselecting)
			up->oUndo.doAtomicOperation(AtomicTextOperation::SETSELECTION, *sCurrentSelectionColorName,
				select[0].x(), select[0].y(), select[1].x(), select[1].y());

		paintCursor( false );
		up->oUndo.doAtomicOperation(AtomicTextOperation::MOVECURSOR, QString::null,
			cursor.x(), cursor.y()-1, cursor.x(), cursor.y());
		repaintfrom = cursor.y();
		repaintto = cursor.y()+1;
		eRepaintMode |= PM(REPAINT_SCROLLTO);
		eRepaintMode |= PM(REPAINT_PART);
		up->oUndo.closeOperation();
		break;

	  /*********/
	  /* DOWN  */
	  /*********/
	  case Key_Down:
		if (cursor.y() >= (int)lines.size()-1) break;

		up->oUndo.createOperation("KEY_MOVE_(DOWN)");

		if (selecting || mouseselecting)
			up->oUndo.doAtomicOperation(AtomicTextOperation::SETSELECTION, *sCurrentSelectionColorName,
				select[0].x(), select[0].y(), select[1].x(), select[1].y());

		paintCursor( false );
		up->oUndo.doAtomicOperation(AtomicTextOperation::MOVECURSOR, QString::null,
			cursor.x(), cursor.y()+1, cursor.x(), cursor.y());
		repaintfrom = cursor.y()-1;
		repaintto = cursor.y();
		eRepaintMode |= PM(REPAINT_SCROLLTO);
		eRepaintMode |= PM(REPAINT_PART);
		up->oUndo.closeOperation();
		break;

	  /*********/
	  /* HOME  */
	  /*********/
	  case Key_Home:
		{
			EditLine *pLine;
			int iC = 0;

			/* 1. get the actual line */
			pLine = (cursor.y() < (int)lines.size() ? lines[cursor.y()] : NULL);

			/* 2. create operation */
			up->oUndo.createOperation("KEY_MOVE_(HOME)");

			/* 3. in case of "intellihome" go to the end of the spaces/tabs */
			if (properties->getBoolValue("buffer.edit.intellihome", true) && pLine)
			{
				/* 2.1 search for the end of the whitespaces at the beginning of the row */
				for(iC = 0; iC < (int)pLine->length(); iC++)
					if (!pLine->at(iC).isSpace()) break;
				iC = pLine->nontabbedPosition(iC, up->tabsize);

				/* 2.2 search for the end of the whitespaces at the beginning of the row */
				if (!((iC == 0) || ((iC != cursor.x()) &&
				  properties->getBoolValue("buffer.edit.intellihome.first",true))))
					iC = 0;
			}

			/* 4. go to the calculated position */
			if (pLine)
				up->oUndo.doAtomicOperation(AtomicTextOperation::MOVECURSOR, QString::null,
					iC, cursor.y(), cursor.x(), cursor.y());
			else up->oUndo.doAtomicOperation(AtomicTextOperation::MOVECURSOR, QString::null,
				0, cursor.y(), cursor.x(), cursor.y());

			/* 5. close operation */
			up->oUndo.closeOperation();

			repaintfrom = repaintto = cursor.y();
			eRepaintMode |= PM(REPAINT_SCROLLTO);
			eRepaintMode |= PM(REPAINT_PART);

			break;
		}

	  /********/
	  /* END  */
	  /********/
	  case Key_End:
		{
			/* 1. create a new operation */
			up->oUndo.createOperation("KEY_MOVE_(END)");
			if (cursor.y() < (int)lines.size())
			{
				EditLine *lineT=lines[ cursor.y() ];
				/* 2. is this line exists ? */
				if (lineT != NULL)
				{
					/* 2.a go to the end of the line */
					up->oUndo.doAtomicOperation(AtomicTextOperation::MOVECURSOR, QString::null,
							lineT->nontabbedPosition( lineT->length(), up->tabsize ), cursor.y(), cursor.x(), cursor.y());
					repaintfrom = repaintto = cursor.y();
					eRepaintMode |= PM(REPAINT_SCROLLTO);
					eRepaintMode |= PM(REPAINT_PART);
				}
			} else
			{
				up->oUndo.doAtomicOperation(AtomicTextOperation::MOVECURSOR, QString::null,
						0, cursor.y(), cursor.x(), cursor.y());
				repaintfrom = repaintto = cursor.y();
				eRepaintMode |= PM(REPAINT_PART);
			}
			/* 3. create a new operation */
			up->oUndo.closeOperation();
			break;
		}

	  /*********/
	  /* PGUP  */
	  /*********/
	  case Key_PageUp:
		/* 1. create a new operation */
		up->oUndo.createOperation("KEY_PGUP");
		/* 2. go down 'nLines' lines */
		up->oUndo.doAtomicOperation(AtomicTextOperation::MOVECURSOR, QString::null,
				cursor.x(), cursor.y()-nLines, cursor.x(), cursor.y());
		/* 3. if we are before the first line, go to the first line */
		if ( cursor.y()<0)
			up->oUndo.doAtomicOperation(AtomicTextOperation::MOVECURSOR, QString::null,
					cursor.x(), 0, cursor.x(), cursor.y());
		eRepaintMode |= PM(REPAINT_SCROLLTO);
		eRepaintMode |= PM(REPAINT_PART);
		/* 5. close the operation */
		up->oUndo.closeOperation();
		break;

	  /*********/
	  /* PGDN  */
	  /*********/
	  case Key_PageDown:
		/* 1. create a new operation */
		up->oUndo.createOperation("KEY_PGDN");
		/* 2. go down 'nLines' lines */
		up->oUndo.doAtomicOperation(AtomicTextOperation::MOVECURSOR, QString::null,
				cursor.x(), cursor.y()+nLines, cursor.x(), cursor.y());
		/* 3. if we are after the end of the buffer, go to the end of the buffer */
		if ( cursor.y()>=(int)lines.size())
			up->oUndo.doAtomicOperation(AtomicTextOperation::MOVECURSOR, QString::null,
					cursor.x(), lines.size()-1, cursor.x(), cursor.y());
		e->accept();
		/* 4. set repaint parameters */
		eRepaintMode |= PM(REPAINT_SCROLLTO);
		eRepaintMode |= PM(REPAINT_PART);
		/* 5. close the operation */
		up->oUndo.closeOperation();
		break;

	  /**********/
	  /* ENTER  */
	  /**********/
	  case Key_Return:
	  case Key_Enter:          /// <--/  E N T E R
	  {
		/* 1. create a new operation, set the buffer modified */
		up->oUndo.createOperation("KEY_ENTER");
		//up->setModified( true );

		/* 2. fill the line until the cursor */
		cmd_text_filltocursor();

		/* 3.1 calculate where to split the line */
		EditLine *lineT=lines[ cursor.y() ];
		int split=lineT->tabbedPosition( cursor.x(), up->tabsize );

		/* 3.2 store the second half of the line */
		QString *sLineSecHalf = new QString(lineT->right( lineT->length()-split )); /// Create new string

		/* 3.3 delete the second half of the current line */
		up->oUndo.doAtomicOperation(AtomicTextOperation::DELETESTRING, *sLineSecHalf,
				split, cursor.y());

		QString *sLineSecHalfMod = new QString(*sLineSecHalf);
		/* 4.1 fill the beginning of the stored second half with spaces, tabs from the previous line */
		for(iCol = 0; iCol < (int)lineT->length(); iCol++)
			if ((lineT->at(iCol) == ' ') || (lineT->at(iCol) == '\t'))
				sLineSecHalfMod->insert(iCol, lineT->at(iCol));
			else break;

		/* 4.2 create a new line from the second half of the current line */
		up->oUndo.doAtomicOperation(AtomicTextOperation::CREATELINE, QString::null, 0, cursor.y()+1);
		up->oUndo.doAtomicOperation(AtomicTextOperation::INSERTSTRING, *sLineSecHalfMod,
				0, cursor.y() + 1);

		/* 5. move the cursor to the beginning of the new line (after white spaces) */
		up->oUndo.doAtomicOperation(AtomicTextOperation::MOVECURSOR, QString::null,
				lineT->nontabbedPosition( iCol, up->tabsize ), cursor.y()+1, cursor.x(), cursor.y());

		/* 6. set repaint options */
		eRepaintMode |= PM(REPAINT_FULL);

		/* 7. close operation */
		up->oUndo.closeOperation();
		break;
	  }

	  /*************/
	  /* DELETE,   */
	  /* BACKSPACE */
	  /*************/
	  case Key_Delete:
	  case Key_BackSpace:
	  {
		/* Key Handler: Delete, Backspace */

		/* 1. create a new operation, set the buffer modified */
		up->oUndo.createOperation("KEY_DELETE/BACKSPACE");
		//up->setModified( true );

		/* 2. get the current line */
		EditLine *lineT=lines[ cursor.y() ];
		if (lineT==NULL) break;

		int backpos;
		/* 3.a this is a Delete event */
		if (e->key()==Key_Delete)
		{
			if (wasselected) break;

			/* are we at the end of the line ? */
			backpos=lineT->tabbedPosition( cursor.x(), up->tabsize );
			if (backpos==(int)lineT->length())
			{
				/* paste the next line to the actual */
				if ( cursor.y() < (int)lines.size()-1 )
				{
					eRepaintMode |= PM(REPAINT_FULL);
					
					up->oUndo.doAtomicOperation(AtomicTextOperation::INSERTSTRING, *new QString(*lines[ cursor.y()+1 ]),
							lineT->length(), cursor.y());
					//*lineT += *lines[ cursor.y()+1 ];
					up->oUndo.doAtomicOperation(AtomicTextOperation::DELETELINE, *new QString(*lines[cursor.y()+1]),
							0, cursor.y()+1);
				}
				break;
			}
		} else
		/* 3.b this is a Backspace event */
		{
			backpos=lineT->tabbedPosition( cursor.x()-1, up->tabsize );
			if ( backpos >= (int)lineT->length() )
				backpos=lineT->length()-1;
			if (backpos==lineT->tabbedPosition( cursor.x(), up->tabsize ))
				backpos--;                        // it's a tab
			/* 3.b.a check: are we at the beginning of a line */
			if (backpos<0)
			{                                     // moving this line upper
				if ( cursor.y() <= 0 ) break;
				EditLine *lineP=lines[ cursor.y()-1 ];
				/* move cursor to the end of the previous line */
				up->oUndo.doAtomicOperation(AtomicTextOperation::MOVECURSOR, QString::null,
					lineP->nontabbedPosition(lineP->length(), up->tabsize), cursor.y(), cursor.x(), cursor.y());
				up->oUndo.doAtomicOperation(AtomicTextOperation::INSERTSTRING, *new QString(*lineT),
						lineP->length(), cursor.y()-1);
				up->oUndo.doAtomicOperation(AtomicTextOperation::DELETELINE, *new QString(*lines[cursor.y()]),
						0, cursor.y());
				up->oUndo.doAtomicOperation(AtomicTextOperation::MOVECURSOR, QString::null,
					cursor.x(), cursor.y()-1, cursor.x(), cursor.y());
				eRepaintMode |= PM(REPAINT_FULL);
				break;
			}
		}
		/* 4.! remove the character */
		up->oUndo.doAtomicOperation(AtomicTextOperation::DELETESTRING, *new QString(lineT->at(backpos)),
				backpos, cursor.y());
		/* 5. move the cursor one character to the left */
		up->oUndo.doAtomicOperation(AtomicTextOperation::MOVECURSOR, QString::null,
				lineT->nontabbedPosition( backpos, up->tabsize ), cursor.y(), cursor.x(), cursor.y());
		/* 6. set repaint parameters */
		repaintfrom = cursor.y();
		repaintto = cursor.y();
		eRepaintMode |= PM(REPAINT_SCROLLTO);
		eRepaintMode |= PM(REPAINT_PART);

		/* 7. close operation */
		up->oUndo.closeOperation();
		break;
	  }
	  case Key_F12: {
		if ( e->type() != QEvent::KeyPress ) break;
		int pos=lines[ cursor.y() ]->tabbedPosition( cursor.x(), up->tabsize );
		printf("Hexcode: %d\n", lines[ cursor.y() ]->at(pos).latin1() );
		break;
	  }

	  default:
		// Tab - ident in or out
		if ((e->key()==Key_Tab) && (select[1].y()!=-1))
		{
			static char *tabulators[9] = { "\t", " ", "  ", "   ", "    ",
			                       "     ", "      ", "       ", "        " };
			/* 1. create a new operation, set the buffer modified */
			up->oUndo.createOperation("TAB-INDENT");

			up->oUndo.doAtomicOperation(AtomicTextOperation::SETSELECTION, *sCurrentSelectionColorName,
				select[0].x(), select[0].y(), select[1].x(), select[1].y());

			//up->setModified( true );
			int sel1y = select[selectsmaller].y(),
				sel2y = select[1-selectsmaller].y(),
				sel1x = select[selectsmaller].x(), i;

			for(i=sel1y; (i<=sel2y) && (i>=0); i++)
				if ((i<sel2y) || (sel1x>0))
				{
					if (e->state() & QKeyEvent::ControlButton)
					{
						int iTabMode = up->pTabCombo->currentItem();
						if (lines[i]->startsWith(tabulators[iTabMode]))
							up->oUndo.doAtomicOperation(AtomicTextOperation::DELETESTRING,
								*new QString(tabulators[iTabMode]), 0, i);
					} else 
					if (lines[i]->length() > 0)
					{
						int iTabMode = up->pTabCombo->currentItem();
						up->oUndo.doAtomicOperation(AtomicTextOperation::INSERTSTRING,
							*new QString(tabulators[iTabMode]), 0, i);
					}
				}

			up->oUndo.doAtomicOperation(AtomicTextOperation::SETSELECTION, *sCurrentSelectionColorName,
				select[0].x(), select[0].y(), select[1].x(), select[1].y());

			up->oUndo.closeOperation();
			eRepaintMode |= PM(REPAINT_FULL);
		} else
		// ANY printable key
		if (!(e->state()&ControlButton) && !(e->state()&AltButton))
		{
			if (e->text().length()>0)
			{
				up->oUndo.createOperation(QString("KEY_PRINTABLE_(")+e->text()+")");
				//up->setModified( true );
				cmd_text_insertchar( e->text() );
				e->accept();
				repaintfrom = repaintto = cursor.y();
				eRepaintMode |= PM(REPAINT_PART) | PM(REPAINT_SCROLLTO);
				up->oUndo.closeOperation();
			}
		}
	}

	iCurrLength = lines.size();
	iPrevLength += nLines;
	iCurrLength += nLines;
	for(; (iPrevLength != 0) || (iCurrLength !=0);
	       iPrevLength /= 10, iCurrLength /= 10)
	     if ((iPrevLength == 0) ^ (iCurrLength == 0))
	     {
	         up->reconfigure();
	         break;
	     }

	/* are we in the middle of a selection */
	if (selecting||mouseselecting)
	{
		/* YES - modify the borders of the selection */
		select[1].setX( cursor.x() );
		select[1].setY( cursor.y() );
		if ( (select[0].y()<select[1].y()) ||
			(select[0].y()==select[1].y()) && (select[0].x()<select[1].x()) )
		{
			selectsmaller = 0;
		} else
		{
			selectsmaller = 1;
		}

		up->oUndo.continueOperation();
		up->oUndo.doAtomicOperation(AtomicTextOperation::SETSELECTION, sSelectionColorName,
			select[0].x(), select[0].y(), select[1].x(), select[1].y());
		up->oUndo.closeOperation();
	}

	if (e->text().length() > 0)
		up->cmd_scrollToCursor();

	if (eRepaintMode & PM(REPAINT_FULL))
		repaintfrom = repaintto = -1;

	if (eRepaintMode & (PM(REPAINT_FULL) | PM(REPAINT_PART)))
		repaint(false);
	if (eRepaintMode & PM(REPAINT_SCROLLTO))
		up->cmd_scrollToCursor();
}

void EditBufferText::cmd_text_insertchar( const QString &oChar, bool bWrap )
{
	int iWrap = EditBufferText::properties->getNumericalValue("buffer.wrap.preferred_width", -1);
	bool bAutoWrap =  EditBufferText::properties->getBoolValue("buffer.wrap.autowrap", false);

	/* protecting the text from non-real characters */
	if ( oChar.length() == 0 ) return;

	cmd_text_filltocursor();

	/* insert the char into the string */
	EditLine *lineT=lines[ cursor.y() ];
	int iCursorCol = lineT->tabbedPosition( cursor.x(), up->tabsize );
	up->oUndo.doAtomicOperation(AtomicTextOperation::INSERTSTRING, oChar, iCursorCol, cursor.y());

	/* move the cursor to the next position */
	int iNewCursorCol = lineT->nontabbedPosition(iCursorCol+1, up->tabsize);
	up->oUndo.doAtomicOperation(AtomicTextOperation::MOVECURSOR, QString::null, iNewCursorCol,
	           cursor.y(), cursor.x(), cursor.y());
	repaintfrom = repaintto = cursor.y();

	/* word wrap  !!! fix the look of this code! */
	if ((bAutoWrap || bWrap) && (iWrap != -1) && (cursor.x() >= iWrap))
	{
		int iCursorPos = lineT->nontabbedPosition(cursor.x());
		if (iCursorPos >= (int)lineT->length()) iCursorPos --;
		int iSplitPos = lineT->findRev(' ', iCursorPos);
		printf("row: %s\n", lineT->latin1());
		printf("nt: %d\n", lineT->nontabbedPosition(cursor.x()));
		printf("wrap2, iPos:%d\n", iSplitPos);

		if (iSplitPos != -1)
		{               /* almost key event: Enter */
			/* split the line 1. get the second half */
			cursor.setX( lineT->nontabbedPosition(iSplitPos) );
			QString *sLineSecHalf = new QString(lineT->mid(lineT->nontabbedPosition(iSplitPos)));

			/* cut out the second half from the line */
			up->oUndo.doAtomicOperation(AtomicTextOperation::DELETESTRING, *sLineSecHalf,
					iSplitPos, cursor.y());

			up->oUndo.doAtomicOperation(AtomicTextOperation::CREATELINE, QString::null, 0, cursor.y()+1);
			up->oUndo.doAtomicOperation(AtomicTextOperation::INSERTSTRING, *new QString(*sLineSecHalf), 0, cursor.y()+1);
			EditLine *l = lines[ cursor.y()+1 ];
			if (l != NULL)
			up->oUndo.doAtomicOperation(AtomicTextOperation::MOVECURSOR, QString::null,
			           l->nontabbedPosition(l->length()), cursor.y()+1, cursor.x(), cursor.y());
		}
	}
}

// * COMMAND: text_fillToCursor ***
// inserts lines and spaces until the cursor
void EditBufferText::cmd_text_filltocursor( void )
{
    /* a. create lines (until the cursor) */
	while (cursor.y() >= (int) lines.size())
		up->oUndo.doAtomicOperation(AtomicTextOperation::CREATELINE, QString::null, 0, cursor.y());

	/* b. insert spaces until the cursor */
	EditLine *l = lines[ cursor.y() ];
	/* b.a create a string with spaces */
	QString sSpaces;
	for(int c=0, g=l->nontabbedPosition(l->length(), up->tabsize); 
				c < cursor.x()-g; c++) 
	{
		sSpaces.append( " " );
	}
	/* b.b insert the spaces after the string */
	up->oUndo.doAtomicOperation(AtomicTextOperation::INSERTSTRING, sSpaces, l->length(), cursor.y());
}

void EditBufferText::scroll( int first ) 
{
	int oldfirst = firstVisible;
	int lines = nLines - abs(first - oldfirst);

	firstVisible = first; 

	if (lines == nLines) return;

	if (lines <= 0) {
		repaint(false);
		return;
	}

	if (oldfirst < first)
		bitBlt(this, 0, 0, this, 0, (first-oldfirst)*fontheight, 
		width(), lines*fontheight);
	else bitBlt(this, 0, (oldfirst-first)*fontheight, 
	            this, 0, 0, width(),lines*fontheight);

	int nn = nLines-lines;
	if (oldfirst < first) {
		int l = first + nLines - 1;
		repaintfrom = l - nn + 1; repaintto = l;
		repaint( false );
	} else {
		int l = first;
		repaintfrom = l; repaintto = l + nn - 1;
		repaint( false );
	}
}

void EditBufferText::paintCursor( bool visible )
{
	if (cursorpainted == visible) return;

	int y = cursor.y();
	if ( (y < firstVisible) || (y >= firstVisible + nLines)) return;
	repaintfrom = repaintto = y;

	if ( ! visible ) {
		cursor.setY( -1 );
		cursorpainted = false;
	}

	repaint( false );
	cursor.setY( y );
}                                     

void EditBufferText::setSelecting(bool bSelecting)
{
	if (bSelecting)
	{
		select[0].setX(cursor.x());
		select[0].setY(cursor.y());

		select[1].setX(cursor.x());
		select[1].setY(cursor.y());

		selecting = true;
		up->oUndo.doAtomicOperation(AtomicTextOperation::SETSELECTION, sSelectionColorName,
			cursor.x(), cursor.y(), select[1].x(), select[1].y());

		repaint(true);
	} else selecting = false;
}

/* needed for the colorer interface */
class cRegion
{
  public:
	int iBegin, iLength, iColor;
	int iRegionType;           /* 0 - region, 1 - schemebegin, 2 - schemeend */
};
vector< cRegion * > *pRetVect = NULL;

void EditBufferText::paintEvent( QPaintEvent *pe )
{
	EditLine *s;
	int i;
	bool bFullRepaint = (repaintfrom < 0);
	QString sFilterFunction("filter.parseline");

	repaintfrom -= firstVisible;
	repaintto -= firstVisible;

	if (bFullRepaint)iMaxLineLength = 0;       /* needed for the horizontal scrollbar */

	if (repaintfrom < 0) repaintfrom = 0;
	if (repaintto < 0) repaintto = nLines - 1;
	if (repaintto < 0) repaintto = 0;

	QPainter p;

	// Drawing text
	//? i<nLines ??
	for(i=repaintfrom; i<nLines; i++)
	{
		// 1. get the current line
//		if ((firstVisible+i < 0) || (firstVisible+i > (int) lines.count())) continue;
		if (firstVisible+i < (int)lines.size())
			s = lines[firstVisible + i];
		else s = NULL;
		int length=( s!=NULL ? (int)s->length() : -1 );
		if ( i>repaintto ) break;


		// Double buffering (the text)
		if (dbuf)
		{
			p.begin(&pm, this);
		}
		else 
		{
			p.begin( this );
			p.translate( WOFFSET, i*fontheight );
		}


		// Selection
		int selbeg=-1, selend=-1;
		if ((select[0] != select[1]) &&
		    ( (select[selectsmaller].y() <= firstVisible+i) && 
			  (firstVisible+i <= select[1-selectsmaller].y()) ) )
		{
			if (select[selectsmaller].y()==firstVisible+i) 
				selbeg=select[selectsmaller].x() - up->hscrollValue; else selbeg=0;
			if (select[1-selectsmaller].y()==firstVisible+i) 
				selend=select[1-selectsmaller].x() - up->hscrollValue;
			else if (s != NULL)
			{
				if (*(properties->getValue(
					"buffer.selection.style", new QString("lineend")))
					== "wholeline" )
					selend = size().width()/fontwidth;
				else selend = s->nontabbedPosition( length, up->tabsize )  - up->hscrollValue;
			}
		} else { selbeg=-1; selend=0; }

		if (length == -1)
		{
			p.fillRect( -WOFFSET, 0, size().width(), fontheight, *bg );
		} else
		{
			p.fillRect( selend*fontwidth, 0,
				size().width(), fontheight,
				((*style == "darkerend") && (selbeg>-1) &&
					( firstVisible+i != select[1-selectsmaller].y() )?
#if QT_VERSION > 280
					selectcolor->dark(125): paletteBackgroundColor() ) );
#else
					selectcolor->dark(125): *textbg ) );
#endif
		}
		if (selbeg!=-1)
		{
			p.fillRect( selbeg*fontwidth, 0,
				(selend-selbeg)*fontwidth, fontheight,
				*selectcolor );
			p.fillRect( 0, 0,
				selbeg*fontwidth, fontheight,
#if QT_VERSION > 280
				(length == -1 ? *bg : paletteBackgroundColor() ) );
#else
				(length == -1 ? *bg : *textbg ) );
#endif
		}

		/* clear the left margin */
		p.fillRect( -WOFFSET, 0, WOFFSET, fontheight,
#if QT_VERSION > 280
				(length == -1 ? *bg : paletteBackgroundColor() ) );
#else
				(length == -1 ? *bg : *textbg ) );
#endif

		// The Cursor
		if ( firstVisible+i == cursor.y() )
		{
			if ( properties->getValue( "buffer.cursor.style",
					new QString("caret") )->lower() == "block" )
			{
				p.fillRect( (cursor.x() - up->hscrollValue)*fontwidth, 0,
				fontwidth, fontheight, *cursorpen );
			} else
			{
				int cursx = cursor.x();
				p.setPen( *cursorpen );
				p.drawLine( (cursx - up->hscrollValue)*fontwidth-1, 1,
				            (cursx - up->hscrollValue)*fontwidth-1, fontheight-1 );
				p.drawLine( (cursx - up->hscrollValue)*fontwidth, 1,
				            (cursx - up->hscrollValue)*fontwidth, fontheight-1 );
				p.drawLine( (cursx - up->hscrollValue)*fontwidth+1, 1,
				            (cursx - up->hscrollValue)*fontwidth+1, fontheight-1 );
				p.drawLine( (cursx - up->hscrollValue)*fontwidth-3, 1,
				            (cursx - up->hscrollValue)*fontwidth+3, 1);
				p.drawLine( (cursx - up->hscrollValue)*fontwidth-3, fontheight-1,
				            (cursx - up->hscrollValue)*fontwidth+3, fontheight-1);
			}
			cursorpainted = true;
		}

		// Drawing the content of ONE line
		if (length != -1) {
			int iPos;

			if ((!Editor::editor->moduleloader->pFilterModule) || (!bUseColorer))
			{
				syntax->setText( s, s->forward, up->bSyntaxHighlighting );
				for( iPos=0; syntax->next(); ) {
					QString *type = syntax->getType();
					int iSubStrPosition = syntax->getPosition();
					int iSubStrLength = syntax->getLength();

					/* set the paint color */
					if (type==NULL) p.setPen(*textpen);
					else if (*type=="String") p.setPen(*stringpen);
					else if (*type=="Comment") p.setPen(*commentpen);
					else if (*type=="Keyword") p.setPen(*keypen);
					else if (*type=="Preprocessor") p.setPen(*preprocpen);

					displayText(p, *s, iSubStrPosition, iSubStrLength);
				}
			} else
			{
				if (Editor::editor &&
				    Editor::editor->moduleloader &&
				    Editor::editor->moduleloader->pFilterModule)
				{
					int iSchemeColor;
					bool bInScheme = false;

					if ((bUseColorer) && (pFilterID == NULL))
					{
						pFilterID = Editor::editor->moduleloader->pFilterModule->pFunction(
							Editor::editor->moduleloader->pFilterModule, "new");
						if (pFilterID == NULL) bUseColorer = false;
					}

					// Filter call - call the colorer filter
					pRetVect = (vector<cRegion *> *)
						Editor::editor->moduleloader->pFilterModule->pFunction(
							Editor::editor->moduleloader->pFilterModule,
					               sFilterFunction, (QString *)pFilterID, NULL, firstVisible + i, 1);

					//! printf("-------------------------------------\nRESULT\n");
					int iPrevEnd = 0;
					int iBegin = 0, iLength = 0;
					static vector< cRegion *>::iterator itRegion;
					static QColor oActColor;

					if (pRetVect)
					{
						for(itRegion=pRetVect->begin(); itRegion!=pRetVect->end(); itRegion ++)
						{
							cRegion *pRegion = (*itRegion);
							if (pRegion != NULL)
							{
								cRegion &oRegion = *pRegion;

								iBegin = oRegion.iBegin;
								iLength = oRegion.iLength;
								/* print with the original color */
								if (iPrevEnd < iBegin)
								{
									if (bInScheme)
									{
										oActColor.setRgb(iSchemeColor);
										p.setPen( oActColor );
									}
									else p.setPen( *textpen );
									displayText(p, *s, iPrevEnd, iBegin - iPrevEnd);
								}

								/* is this region a scheme? */
								if (oRegion.iRegionType == 1)
								{
									iSchemeColor = oRegion.iColor;
									bInScheme = true;
									iPrevEnd = iBegin;
								}
								if (oRegion.iRegionType == 2)
								{
									bInScheme = false;
									iPrevEnd = iBegin;
								}

								/* print the highlighted area */
								if (bInScheme)
								{
									if (oRegion.iColor != -1)
									{
										//int iMasked = ((oRegion.iColor >> 24) / 3) << 24;
										//iMasked |= (((oRegion.iColor >> 16)&0xff) / 3) << 16;
										//iMasked |= (((oRegion.iColor >> 8)&0xff) / 3) << 8;
										//iMasked |= (oRegion.iColor &0xff) /3;
										//oActColor.setRgb(iMasked + iSchemeColor);
										oActColor.setRgb(oRegion.iColor);
									}
									else 
										oActColor.setRgb(iSchemeColor);
									p.setPen( oActColor );
								}
								else if (oRegion.iColor != -1)
								{
									oActColor.setRgb(oRegion.iColor);
									p.setPen( oActColor );
								}
								else p.setPen( *textpen );
								//! printf("%d:   %d - %d\n", jj, iBegin, iLength);
								displayText(p, *s, iBegin, iLength);
								iPrevEnd = iBegin + iLength;
							}
						}
					}                                               
					if ((itRegion==pRetVect->begin()) || (iBegin+iLength < (int)s->length()))
					{
						if (bInScheme)
						{
							oActColor.setRgb(iSchemeColor);
							p.setPen( oActColor );
						}
						else p.setPen( *textpen );
						displayText(p, *s, iBegin + iLength, s->length() - iBegin - iLength);
					}
				}
			}

			if (up->hscrollValue)
#if QT_VERSION > 280
				p.fillRect(-WOFFSET, 0, WOFFSET, fontheight, paletteBackgroundColor());
#else
				p.fillRect(-WOFFSET, 0, WOFFSET, fontheight, *textbg);
#endif
			if (bFullRepaint && (iPos >= iMaxLineLength))
				iMaxLineLength = iPos;
			if (firstVisible+i+1 < (int)lines.size())
			{
				lines[firstVisible+i+1]->forward = syntax->getForward();
			}
		}
		p.end();
		// End of double buffering: copying back the pixmap
		if (dbuf) bitBlt(this, WOFFSET, i*fontheight, &pm);
	}

	up->position_x->clear();
	up->position_x->setNum( cursor.x()+1 );
	up->position_y->clear();
	up->position_y->setNum( cursor.y()+1 );
	up->position->repaint(true);
	if (! (up->findbox ) ) setFocus();

	if (iMaxLineLength - 1 < cursor.x()) iMaxLineLength = cursor.x() + 1;
	if ((up->hscroll) && (up->hscroll->value() == 0) &&
	    (iMaxLineLength < size().width() / fontwidth)) iMaxLineLength = 0;
	else if ((up->hscroll) && bFullRepaint)
		iMaxLineLength = max(iMaxLineLength, up->hscroll->value() + up->hscroll->pageStep());

	up->cmd_scrollBar();

	repaintfrom=0; repaintto=-1;
}

// EditBufferText::DisplayText(....)
/**
 * Displays a given string. It can contain tabulator characters.
 */
void EditBufferText::displayText(QPainter &p, QString &sText, int iBegin, int iLength )
{
	static int iTab;           /* the previous tab - we search for the new tab after this */
	static int iNextTab;       /* the actually found new tab */
	static int iLen;           /* the length of the next substring */
	static int iPos;
	static QColor oOrigPen;

	oOrigPen = p.pen().color();
	iPos = EditLine::nontabbedPosition(sText, iBegin, up->tabsize);

	/* go through the string to print tabs / substrings*/
	for( iLen = 0, iTab = iBegin; iTab < iBegin + iLength; )
	{
		/* search for the next tab in the substring */
		iNextTab = sText.find('\t', iTab);
		/* chk: print until the next tab or until the end of the substring */
		if ((iNextTab == -1) || (iNextTab >= iBegin + iLength))
		{
			//!iLen = iBegin + iLength - iTab + 1;
			iLen = iBegin + iLength - iTab;
			iNextTab = -1;
		}
		else iLen = iNextTab - iTab;

		/* print the next portion */
		if (iLen > 0)
		{
			// don't change pen if it is not necessary
			if (iTab > iBegin) p.setPen(oOrigPen); 
//!			printf("drtext: %d %d %s\n", iTab, iLen, sText.latin1());
#if QT_VERSION > 280
			p.drawText( (iPos - up->hscrollValue) * fontwidth, fontascent, 
				sText, iTab, iLen);
#else
			p.drawText( (iPos - up->hscrollValue) * fontwidth, fontascent, 
				sText.mid(iTab, iLen));
#endif
			iPos += iLen;
		}

		/* print the next tab (if there is another) */
		if (iNextTab != -1)
		{
			p.setPen( *tabpen );
			int tabstop=(iPos/up->tabsize+1)*up->tabsize;
			p.drawLine( (iPos - up->hscrollValue)*fontwidth, fontheight-1,
				(tabstop - up->hscrollValue)*fontwidth-1, fontheight-1 );
			p.drawLine( (tabstop - up->hscrollValue)*fontwidth-1, fontheight-1,
				(tabstop - up->hscrollValue)*fontwidth-1, fontheight-5 );
			p.setPen( *textpen );
			iPos = tabstop;
			iTab = iNextTab + 1;
		} else break;
	}
}

