 /*
  * EditLineSyntax.h
  *
  * Syntax highlighting
  *
  * Author: Ferenc Moricz
  * $Revision: 1.18 $
  * $Date: 2003/12/26 15:06:09 $
  *
  */

#include "Editor.h"
#include "EditLineSyntax.h"
#include <stdio.h>

vector<SyntaxRule *, allocator<SyntaxRule *> > *EditLineSyntax::pRules = NULL;
vector<int *, allocator<int *> > *EditLineSyntax::pHits = NULL;

SyntaxRule::SyntaxRule( char *name, char *begstring, char *endstring )
{
	this->begstring = new QRegExp( *new QString(begstring) );
	if ( endstring == NULL ) this->endstring = NULL;
	else this->endstring = new QRegExp( *new QString(endstring) );
	this->name = new QString( name );
}

EditLineSyntax::EditLineSyntax()
{
	if (pRules == NULL) pRules = new vector<SyntaxRule *, allocator<SyntaxRule*> >();
	if (pHits == NULL) pHits = new vector<int *, allocator<int *> >();

	this->string = new QString( "" );

	pRules->push_back( new SyntaxRule( "Comment", "/\\*", "\\*/" ) );
	pRules->push_back( new SyntaxRule( "Comment", "//", "$" ) );
	pRules->push_back( new SyntaxRule( "String", "\"", "\"" ) );
	pRules->push_back( new SyntaxRule( "Preprocessor", "^#", "$" ) );
#if QT_VERSION > 280
	pRules->push_back( new SyntaxRule( "Keyword", "\\bvoid[\\s,\\*,)]", NULL ) );
	pRules->push_back( new SyntaxRule( "Keyword", "\\bchar[\\s,\\*,)]", NULL ) );
	pRules->push_back( new SyntaxRule( "Keyword", "\\bint[\\s,\\*,)]", NULL ) );
	pRules->push_back( new SyntaxRule( "Keyword", "\\blong[\\s,\\*,)]", NULL ) );
	pRules->push_back( new SyntaxRule( "Keyword", "\\bfor\\b", NULL ) );
	pRules->push_back( new SyntaxRule( "Keyword", "\\bwhile\\b", NULL ) );
	pRules->push_back( new SyntaxRule( "Keyword", "\\bdo\\b", NULL ) );
	pRules->push_back( new SyntaxRule( "Keyword", "\\bif\\b", NULL ) );
	pRules->push_back( new SyntaxRule( "Keyword", "\\belse\\b", NULL ) );
	pRules->push_back( new SyntaxRule( "Keyword", "\\bcontinue\\b", NULL ) );
	pRules->push_back( new SyntaxRule( "Keyword", "\\bbreak\\b", NULL ) );
	pRules->push_back( new SyntaxRule( "Keyword", "\\bnew\\b", NULL ) );
	pRules->push_back( new SyntaxRule( "Keyword", "\\bdelete\\b", NULL ) );
	pRules->push_back( new SyntaxRule( "Keyword", "\\bthis\\b", NULL ) );
	pRules->push_back( new SyntaxRule( "Keyword", "\\bclass\\b", NULL ) );
	pRules->push_back( new SyntaxRule( "Keyword", "\\bstruct\\b", NULL ) );
#else
	pRules->push_back( new SyntaxRule( "Keyword", "[\\s,^,(]?void[\\s,\\*,)]", NULL ) );
	pRules->push_back( new SyntaxRule( "Keyword", "[\\s,^,(]?char[\\s,\\*,)]", NULL ) );
	pRules->push_back( new SyntaxRule( "Keyword", "[\\s,^,(]?int[\\s,\\*,)]", NULL ) );
	pRules->push_back( new SyntaxRule( "Keyword", "[\\s,^,(]?long[\\s,\\*,)]", NULL ) );
	pRules->push_back( new SyntaxRule( "Keyword", "[\\s,^]?for[\\s,(,$]", NULL ) );
	pRules->push_back( new SyntaxRule( "Keyword", "[\\s,^]?while[\\s,(,$]", NULL ) );
	pRules->push_back( new SyntaxRule( "Keyword", "[\\s,^]?do[\\s,{,$]", NULL ) );
	pRules->push_back( new SyntaxRule( "Keyword", "[\\s,^]?if[\\s,(,$]", NULL ) );
	pRules->push_back( new SyntaxRule( "Keyword", "[\\s,^]?else[\\s,;,{,$]", NULL ) );
	pRules->push_back( new SyntaxRule( "Keyword", "[\\s,^]?continue[\\s,;,$]", NULL ) );
	pRules->push_back( new SyntaxRule( "Keyword", "[\\s,^]?break[\\s,;,$]", NULL ) );
	pRules->push_back( new SyntaxRule( "Keyword", "[\\s,^]?new[\\s,(,$]", NULL ) );
	pRules->push_back( new SyntaxRule( "Keyword", "[\\s,^]?delete[\\s,(,$]", NULL ) );
	pRules->push_back( new SyntaxRule( "Keyword", "[\\s,^]?this[\\s,(,$]", NULL ) );
	pRules->push_back( new SyntaxRule( "Keyword", "[\\s,^]?class[\\s,{,$]", NULL ) );
	pRules->push_back( new SyntaxRule( "Keyword", "[\\s,^]?struct[\\s,{,$]", NULL ) );
#endif

	for(int i = 0; i < 20; i++) pHits->push_back( new int(-1) );
}

void EditLineSyntax::setText( QString *string, int forward, bool bSyntaxHighlighting )
{
	*(this->string) = *string;
	for( unsigned int i = 0; i < pHits->size(); i++) {
		*(*pHits)[i] = -1;
	}
	this->type = NULL;
	iPosition = 0;
	length = 0;
	this->forward = forward;
	this->bSyntaxHighlighting = bSyntaxHighlighting;
}

bool EditLineSyntax::next()
{
	unsigned int iRulesCount = (bSyntaxHighlighting ? pHits->size() : 0);

	if ( length > 0 ) {
		string->remove( 0, length );
		for( unsigned int i=0; i < iRulesCount; i++) {
			int *value = (*pHits)[i];
			if (*value < length) {
				if (*value > -2) *value = -1;
			} else *(*pHits)[i] = *(*pHits)[i] - length;
		}
		iPosition += length;
	}
	if (string->length() == 0) return false;

	int wh=0, least = string->length();
	for( unsigned int i = 0; i < iRulesCount; i++) {
		int *value = (*pHits)[i];
		if ( *value == -1 ) {
#if QT_VERSION > 280
			*value = (*pRules)[i]->begstring->search(*string, 0);
#else
			*value = (*pRules)[i]->begstring->find(*string, 0);
#endif
		}
		if ( *value == -1) *value = -2;
		else if ( ( *value < least ) && ( *value >= 0 ) ) {
			least = *value;
			wh = i;
		}
	}

	if ( (least == (int)string->length()) && (forward == -1) ) {
		length = least;
		type = NULL;
	}
	else if ( (least > 0) && (forward == -1) ) {
		length = least;
		type = NULL;
	} else {
		if (forward > (int)pRules->size()) forward = -1;
		if (forward > -1) {
			wh = forward;
		}
		type = (*pRules)[wh]->name;
		if ( (*pRules)[wh]->endstring == NULL ) {
			int ml;
			(*pRules)[wh]->begstring->match( *string, 0, &ml );
			length = ml;
		} else {
			int l2, l = (*pRules)[wh]->endstring->match( 
					*string, (forward==-1? 1:0), &l2 );
			if (l != -1) {
				length = l+l2;
				forward = -1;
			} else {
				length = string->length();
				forward = wh;
			}
		}
	}
	return true;
}

