/**
 * @file EditSession.cpp
 * @brief A session or project - which is a set of file
 * 
 * @author Ferenc Moricz
 * $Revision: 1.20 $
 * $Date: 2004/02/22 00:12:50 $
 */

/**
 * @class EditSession
 * @brief Session is a set of file buffers, which is usable
 * if you works on more files at the same time.
 *
 */

#include "EditSession.h"
#include "Editor.h"
#include <qtabbar.h>

#include <algorithm>
using namespace std;

// EditSession(.)
/**
 * @brief Constructor - simply calls the parent's constructor
 */
EditSession::EditSession(QWidget *parent) : ExtQTabWidget(parent, "Tab")
{
//	setTabShape( QTabWidget::Triangular );
}

// .addBuffer(..)
/**
 * @brief Adds a new buffer to the session
 * @param name the name of the new buffer
 * @param bSkipPlacement don't look at the properties
 *
 * This function first looks for the value 'editor.openfile.tabindex'
 * and decides which will be the place of the new buffer's tab.
 * Possible values:
 *    - first - the first available position
 *    - last  - the last available position
 *    - afteractual  - the position after the current position
 *    - beforeactual - the position before teh actual position
 */
EditBuffer *EditSession::addBuffer( QString name, bool bSkipPlacement)
{
	int index = -1;             /* create new EditBuffer */
	EditBuffer *buffer = new EditBuffer(this, &name);

	if (!bSkipPlacement)        /* get current tab's index */
	{
		const QString *sProp =
		    EditBufferText::properties->getValue("editor.openfile.tabindex", NULL );
		if ((sProp == NULL) || (*sProp == "last"))
			index = -1;
		else if ((sProp == NULL) || (*sProp == "first"))
			index = 0;
		else if ((sProp == NULL) || (*sProp == "afteractual"))
			index = currentPageIndex() + 1;
		else if ((sProp == NULL) || (*sProp == "beforeactual"))
			index = currentPageIndex();
	}

#if QT_VERSION < 280
	if (index == -1)
		oBufferList.push_back( (QWidget *)buffer );
	else oBufferList.insert( oBufferList.begin()+index, (QWidget *)buffer);
#endif

                                /* insert new tab */
	insertTab(buffer, *buffer->name, index);
	if (index != -1) setCurrentPage(index);
	else setCurrentPage(tabBar()->count() - 1);

	buffer->reconfigure();

	return buffer;
}


/**
 * @brief Removes the actual buffer
 */
void EditSession::removeBuffer()
{
//	tabBar()->removeTab( tabBar()->tab( tabBar()->currentTab() );
	if (!tabBar()->count()) return;
	EditBuffer *buffer = (EditBuffer *)currentPage();
	if (buffer == NULL)
	{
		derr("session", "Remove null buffer!");
		return;
	} else
	{
		removePage(buffer);
	}

#if QT_VERSION < 280
	vector<QWidget*, allocator<QWidget *> >::iterator
		oPos = std::find( oBufferList.begin(), oBufferList.end(), (QWidget *)buffer);
	if (oPos != oBufferList.end())
		oBufferList.erase( oPos );
#endif

	buffer->close(true);
}

/**
 * @brief Transporter to the QTExtTabwidget class
 */
bool EditSession::event( QEvent *e )
{
	dmesg("6:session", "EVENT: EditSession type:%d", e->type()); 
	return ExtQTabWidget::event(e);
	//return false;
}

