/**
 * @file Editor.cpp
 * @brief Editor's main program
 * 
 * @author Ferenc Moricz
 * $Revision: 1.114 $
 * $Date: 2004/02/22 00:12:50 $
 */

/**
 * @class Editor
 * @brief The main class, the main widget of Freddy.
 *
 * <b>Commands:</b>\n\n
 * <b> filenew</b> - creates a new buffer to edit\n
 * <b> fileclose</b> - closes the actual buffer\n
 * 
 * <b> fileopen</b> - asks for a file in a dialog and opens this file\n
 * <b> filesave</b> - saves the actual buffer\n
 * <b> filesaveas</b> - asks for a name and saves the file under the given name\n

 * <b> sessionnext</b> - goes to the next session \n
 * <b> sessionprevious</b> - goes to the previous session \n
 * <b> buffernext</b> - goes to the next buffer in the actual session \n
 * <b> bufferprevious</b> - goes to the previous buffer in the actual session \n
 */
 
#include "Debug.h"
#include "Editor.h"
#include "EditProperties.h"

#ifdef _SYSTEM_WINDOWS_
char *INSTALL_ETC;
char *INSTALL_SHARE;
char *INSTALL_DOC;
#endif

#define QSTRING_ENDSWITH(qstr,str,n) ((qstr).mid((qstr).length()-(n))==(str))

static int FREDDY_BUILD_NUMBER= 645;
char FREDDY_VERSION[] = "0.80";
char FREDDY_BUILD[ 10 ];

const int FREDDYPORT = 7519;

Editor *Editor::editor = NULL;
EditorMainWindow *Editor::pMainWindow;

#include <algorithm>

#include <qapplication.h>
#include <qdialog.h>
#include <qgrid.h>
#include <qmenubar.h>
#include <qmessagebox.h>
#include <qmultilineedit.h>
#include <qpopupmenu.h>
#include <qpushbutton.h>

#include <stdlib.h>

#ifdef _SYSTEM_WINDOWS_
#include <winreg.h>
#endif

/* XPM */
const char * freddy_xpm[] = {
"35 35 218 2",
"  	c None",
". 	c #F90000",
"+ 	c #F90101",
"@ 	c #F70505",
"# 	c #F21313",
"$ 	c #F20E0E",
"% 	c #F60404",
"& 	c #F90202",
"* 	c #F70C0C",
"= 	c #F23636",
"- 	c #EE5353",
"; 	c #EC4C4C",
"> 	c #EF2828",
", 	c #F90303",
"' 	c #F70E0E",
") 	c #F32C2C",
"! 	c #EF4D4D",
"~ 	c #EB4F4F",
"{ 	c #EA4747",
"] 	c #F70404",
"^ 	c #F21C1C",
"/ 	c #F13636",
"( 	c #EE5E5E",
"_ 	c #E96F6F",
": 	c #F60909",
"< 	c #F41616",
"[ 	c #F23434",
"} 	c #F32626",
"| 	c #F41C1C",
"1 	c #EF5454",
"2 	c #EB7272",
"3 	c #F51515",
"4 	c #F33131",
"5 	c #F80808",
"6 	c #650000",
"7 	c #340000",
"8 	c #280000",
"9 	c #1A0000",
"0 	c #140000",
"a 	c #010000",
"b 	c #000000",
"c 	c #060000",
"d 	c #400F0F",
"e 	c #F51B1B",
"f 	c #F70D0D",
"g 	c #F52626",
"h 	c #F71A1A",
"i 	c #F51717",
"j 	c #F33636",
"k 	c #F70A0A",
"l 	c #2F0202",
"m 	c #370808",
"n 	c #F33030",
"o 	c #F90404",
"p 	c #F31D1D",
"q 	c #F23838",
"r 	c #F22020",
"s 	c #2F0C0C",
"t 	c #0C0000",
"u 	c #0D0000",
"v 	c #030000",
"w 	c #440707",
"x 	c #F90606",
"y 	c #F22424",
"z 	c #F24747",
"A 	c #F73434",
"B 	c #2F0707",
"C 	c #650202",
"D 	c #CB0202",
"E 	c #D10303",
"F 	c #D40000",
"G 	c #CB0000",
"H 	c #C20000",
"I 	c #C40E0E",
"J 	c #F24040",
"K 	c #F80A0A",
"L 	c #EF2C2C",
"M 	c #F22626",
"N 	c #F24848",
"O 	c #F90F0F",
"P 	c #2F0000",
"Q 	c #7E0606",
"R 	c #EE1F1F",
"S 	c #E84747",
"T 	c #ED2424",
"U 	c #F60606",
"V 	c #F60808",
"W 	c #F42929",
"X 	c #DB7272",
"Y 	c #EF4141",
"Z 	c #F14040",
"` 	c #F80707",
" .	c #7E0101",
"..	c #F41111",
"+.	c #EC4141",
"@.	c #EC4B4B",
"#.	c #F50F0F",
"$.	c #F80101",
"%.	c #F80404",
"&.	c #F02525",
"*.	c #F52C2C",
"=.	c #F31C1C",
"-.	c #F04545",
";.	c #F80B0B",
">.	c #020000",
",.	c #5A0000",
"'.	c #B20101",
").	c #B20000",
"!.	c #B30000",
"~.	c #C30000",
"{.	c #F00000",
"].	c #F80909",
"^.	c #F90909",
"/.	c #F50E0E",
"(.	c #F32D2D",
"_.	c #350000",
":.	c #F80000",
"<.	c #F80505",
"[.	c #330000",
"}.	c #D30000",
"|.	c #2E0000",
"1.	c #090000",
"2.	c #120808",
"3.	c #120202",
"4.	c #130000",
"5.	c #430000",
"6.	c #D40303",
"7.	c #ED3838",
"8.	c #F42020",
"9.	c #210000",
"0.	c #6A0000",
"a.	c #C93A3A",
"b.	c #D34545",
"c.	c #DD0202",
"d.	c #DE0000",
"e.	c #E40000",
"f.	c #F00E0E",
"g.	c #F23C3C",
"h.	c #F81010",
"i.	c #F32B2B",
"j.	c #F40F0F",
"k.	c #F60303",
"l.	c #F20A0A",
"m.	c #F60505",
"n.	c #180000",
"o.	c #7D0000",
"p.	c #E71818",
"q.	c #ED6464",
"r.	c #F70808",
"s.	c #F80D0D",
"t.	c #F62121",
"u.	c #F33D3D",
"v.	c #F04343",
"w.	c #F14848",
"x.	c #870505",
"y.	c #EE0D0D",
"z.	c #F04444",
"A.	c #F70303",
"B.	c #F60A0A",
"C.	c #F71515",
"D.	c #F61212",
"E.	c #EF4343",
"F.	c #110101",
"G.	c #881414",
"H.	c #F30F0F",
"I.	c #F31414",
"J.	c #F12424",
"K.	c #F02F2F",
"L.	c #F03737",
"M.	c #F61515",
"N.	c #F80202",
"O.	c #F03E3E",
"P.	c #820404",
"Q.	c #F40202",
"R.	c #F80303",
"S.	c #F22121",
"T.	c #F23D3D",
"U.	c #F43030",
"V.	c #F81818",
"W.	c #F22222",
"X.	c #661515",
"Y.	c #580101",
"Z.	c #480000",
"`.	c #310000",
" +	c #8B0000",
".+	c #F30000",
"++	c #F80606",
"@+	c #F70707",
"#+	c #F61313",
"$+	c #F71616",
"%+	c #F21A1A",
"&+	c #F13737",
"*+	c #F43A3A",
"=+	c #F81919",
"-+	c #F70000",
";+	c #F51D1D",
">+	c #F62222",
",+	c #F51C1C",
"'+	c #F14242",
")+	c #F26A6A",
"!+	c #F76B6B",
"~+	c #F83030",
"{+	c #F90A0A",
"]+	c #F90707",
"^+	c #F62727",
"/+	c #F45D5D",
"(+	c #F78D8D",
"_+	c #F88383",
":+	c #F83636",
"<+	c #F58989",
"[+	c #F88989",
"}+	c #F83737",
"|+	c #F13535",
"1+	c #F38787",
"2+	c #F57979",
"3+	c #F51313",
"4+	c #F03535",
"5+	c #F20000",
"                                                                      ",
"                                . .                                   ",
"                              . . . .                                 ",
"                            . . . . . .                               ",
"                          . . . . . . . .                             ",
"                        . . . . . + @ # $ %                           ",
"                      . . . . . . & * = - ; >                         ",
"                    . . . . . . . . . , ' ) ! ~ {                     ",
"                  . . . . . . . . . . . . ] ^ / ( _                   ",
"                . . : < , . . . . . . . . @ [ } | 1 2                 ",
"              . . . 3 4 5 6 7 8 9 0 a b b b c d e f g h               ",
"            . . . . i j k l b b b b b b b b b m n o & o +             ",
"          . . . . . p q r s b b b c t u u v b w q x . . . .           ",
"        . . . . . + y z A B b v C D E F F G H I J K . . . . .         ",
"      L + . . . . + M N O P b v Q R S T U . . V W 5 . . . . . .       ",
"    X Y + . . . . + y Z ` P b v  ...+.@.#.. . $.%.+ . . . . . . .     ",
"  $.&.*.. . . . . . =.-.;.P b >.,.'.'.).!.!.~.{.. . . . . . . . . .   ",
"  . ].^.. . . . . . /.(.].P b b b b b b b b _.F . :.:.. . . . . . . . ",
"  . . . . . . . . . $.<.$.P b b b b b b b b [.}.% ] . . . . . . . .   ",
"    . . . . . . . . . . . |.b b b 1.2.3.4.4.5.6.7.8.. . . . . . .     ",
"      . . . . @ @ . . . . 9.b b a 0.a.b.c.d.e.f.g.h.. . . . . .       ",
"        . . + 3 i.j.k.l.m.n.b b >.o.p.q.r.. :.<.s.& . . . . .         ",
"          . . <.t.u.v.w.y n.b b v x.y.z.` :.A.B.@ . . . . .           ",
"            . . + 5 C.D.E.F.b b v G.H.r.I.J.K.L.M.+ . . . .           ",
"              . . . . N.O.b b b >.P.Q.R.S.T.U.V.5 , . . .             ",
"                :.:.. . W.X.Y.Z.`. +.+. , ++@+..#+, . .               ",
"                  .+:.. @+j $++ . . $.N.V %+&+*+=++ .                 ",
"                    -+. $.;+>+, . N.K ,+'+)+!+~+{++                   ",
"                      . . , ]++ . + K ^+/+(+_+:+^.                    ",
"                        . . . . . . + ].j <+[+}+                      ",
"                          . . . . . . <.|+1+2+                        ",
"                            . . . . . & 3+4+                          ",
"                                . . . . 5+                            ",
"                                  . . .                               ",
"                                    .                                 "};

#define FIFO_FILE "/tmp/unix_npipe.pipe"

/* 
 * The main function. Creates an Editor object, and initializes it.
 */
int main(int argc, char *argv[])
{
	int retvalue = 1;

	dbeg();

	QApplication a( argc, argv );

	Editor *editor = new Editor(argc, argv, &a);

	retvalue = a.exec();

	editor->saveModifiedFiles();

	if ( EditBufferText::properties->getBoolValue("editor.startup.multi", true) )
		editor->saveState();

	dend();

	return retvalue;
}

EditorMainWindow::EditorMainWindow() : QMainWindow( 0, "Freddy", WDestructiveClose )
{
#ifdef _SYSTEM_LINUX_
//	debug_message("test",  "%s\n", __PRETTY_FUNCTION__);
#endif
}

/*
 *  - set icon
 *  - initialize properties
 *  - load config files
 *  - 

 *  - open files specified at the command line
 *  - initialize ModuleLoader
 *  - initialize menus
 *  - load previous state
 *  - open previuosly opened files
 */
Editor::Editor(int argc, char *argv[], QApplication *pApp)
{
	dbeg();

	debug_add_rule("-");
	debug_add_rule("3:");

	sprintf(FREDDY_BUILD, "%d", FREDDY_BUILD_NUMBER);
	editor = this;

	pMainWindow = NULL;
	pPreferences = NULL;

	pApplication = pApp;
	pTranslator = NULL;

	bSavedStateOpened = FALSE;
                                                       
	// Load properties
	EditBufferText::properties=new EditProperties();

	configGetDirectories();
	configLoadFile( COMMANDLINE );

	pEditorServer = NULL;
	pEditorClient = NULL;
	com = NULL;
	pCommandFileList = processCommandLine(argc, argv);

	if (EditBufferText::properties->getBoolValue("editor.server", true))
		EditorServer::locateServer( FREDDYPORT );
	else startUp();

	if (com) com->updateMenuItems();

	dend();
}

void Editor::startUp()
{
	dmesg("6:config", "Freddy server-client: pEServ: %p, pCmdL: %p, pEClient: %p\n",
	      pEditorServer, pCommandFileList, pEditorClient);

	// Create widgetstack - this is the central widget
	pMainWindow = new EditorMainWindow();
	pApplication->setMainWidget(pMainWindow);
	container = new QSplitter(Qt::Vertical, pMainWindow);
	stack = new QWidgetStack( container );
	pMainWindow->setCentralWidget( container );

	com = new Command();
	configLoadFile( FREDDY );
	configLoadFile( SHORTCUTS );
	configLoadFile( MENU );
	configLoadFile( SAVEDSTATE );

	/* startup in a transporter client mode */
	if ((!pEditorServer) && (pEditorClient))
	{
		pEditorClient->iState = 20;
		pEditorClient->slotWrote( 0 );
		return;
	}

/*? how to handle blank file lists ?
	} else
	if (pEditorClient)
	{
		dwarn("main.io", "No file to open!");
		exit(0);
	}*/

	/* sets up the main font */
	QFont oDefFont("Fixed", 10);
	QFont oFont = QFont(*EditBufferText::properties->getFontValue("editor.font", &oDefFont));
	pMainWindow->setFont(oFont);

	pMainWindow->setIcon( QPixmap( freddy_xpm ) );

	moduleloader = new ModuleLoader();
	EditBuffer::module = moduleloader;
	int x = savedstate.getNumericalValue("window.x", 100),
	    y = savedstate.getNumericalValue("window.y", 100),
		width = savedstate.getNumericalValue("window.width", 600),
		height = savedstate.getNumericalValue("window.height", 400);
	if (x < 0) x = 0;
	if (y < 0) y = 0;
	dprint("6:config", "Main Window: x:%d, y:%d, width:%d, height:%d\n", x, y, width, height);
	pMainWindow->setGeometry(x, y, width, height);

	// Creating menus
	QPopupMenu *p=new QPopupMenu( pMainWindow ); CHECK_PTR( p );
	QMenuBar *menu = pMainWindow->menuBar(); CHECK_PTR( menu );

	// "Session" menu
	sessionid = 100;

	// Open previously opened sessions and its files
	loadState();

//	if (pCommandFileList && !pEditorClient)
	if (pCommandFileList)
	{
		if (pCommandFileList->count() > 0)
			openCommandlineFile(pCommandFileList);

		delete(pCommandFileList);
	}

	dmesg("4:config", "-- Freddy init ended -----------------------\n\n");

	pMainWindow->show();

	pPreferences = new Preferences();
}

/*
 * Reconfigures the object and all its sessions/buffers
 */
void Editor::reconfigure(void)
{
	for(unsigned int iSes = 0; iSes < sessions.size(); iSes ++)
	{
		EditSession *pSession = (EditSession *)sessions[iSes];
		for(int iBuf = 0; iBuf < (int)pSession->myTabBar()->count(); iBuf++)
		{
#if QT_VERSION > 280
			EditBuffer *pBuffer = (EditBuffer *)pSession->page(iBuf);
#else
			pSession->setCurrentPage(iBuf);
			EditBuffer *pBuffer = (EditBuffer *)pSession->currentPage();
#endif
			if (pBuffer) pBuffer->reconfigure();
		}
	}
}

// loadConfigFiles()
/**
 * @brief Find and load the appropriate config files
 *
 * The first task of this function to find the place, where the
 * config files are stored. It searches for the \e freddy
 * directory in the following directories:
 *    - $HOME/etc/freddy
 *    - $HOME/.etc/freddy
 *    - $HOME/etc/.freddy
 *    - $HOME/.etc/.freddy
 *    - $HOME/.freddy
 *    - /etc/freddy
 *
 * If there isn't any result, then it cannot load any file.\n
 * Else it loads the following config files:
 *    - freddy.conf - the main config file (into EditBufferText::properties)
 *    - saved-state - the state in the end of the last execution
 *                    (into this->savedstate)
 */
void Editor::configGetDirectories(void)
{
	static char *userpath[] = { "etc/freddy", ".etc/freddy",
	                 "etc/.freddy", ".etc/.freddy", ".freddy"};

	dbeg();

#ifdef _SYSTEM_LINUX_
	if (!configdir_default.cd(INSTALL_ETC)) configdir_default = QDir::root();

	configdir_user = QDir::home();          // search for the config directory
	for(int which = 0; which < 5; which++)
		if (configdir_user.cd(userpath[which])) break;

	if (configdir_user == QDir::home()) configdir_user = QDir::root();

	/* finding user config directory */
	if (configdir_user.isRoot())    // WARNING: user config dir not found
	{
		QMessageBox mb("Freddy", tr("User config directory not found.<br>"
		               "<b>Create</b> one?"),
		               QMessageBox::Warning, QMessageBox::Yes|QMessageBox::Default,
		               QMessageBox::No|QMessageBox::Escape, QMessageBox::NoButton,
		               pMainWindow, tr("User config directory"), true);
		mb.setFocus();
		if ( mb.exec() == QMessageBox::Yes )
		{
			configdir_user = QDir::home();
			if (!configdir_user.mkdir(".freddy")||!configdir_user.cd(".freddy"))
				configdir_user = QDir::root();
		}
	}
	if (configdir_default.isRoot()) // ERROR: default config dir not found
	{
		derr("config", "No default configuration directory (%s) for Freddy!", INSTALL_ETC);
		derr("config", "Exiting...");
		QApplication::exit(0);
	}
#elif defined (_SYSTEM_WINDOWS_)
		/* Windows version - Getting the installation directory from the
		                     registry value HKEY_CLASSES_ROOT\Freddy\InstallPath */

	/* 1. open the registry key */
	HKEY hResult;
	long iReturn;
	if ((iReturn = RegOpenKeyEx(HKEY_CLASSES_ROOT, L"\\Freddy\\InstallPath", 0, KEY_ALL_ACCESS, &hResult)) == ERROR_SUCCESS)
	{
		/* 1.a registry key is succesfully opened */
		WCHAR sWBuf[256];
		DWORD iWBufLen = 512, iN;
		char sBuf[256], sBufTemp[256];

		/* 1.a.1 getting the value */
		RegQueryValueEx( hResult, NULL, NULL, NULL, (LPBYTE)&sWBuf[0], &iWBufLen);
		//! QMessageBox::information(NULL, "Freddy", "Hello");

		/* 1.a.2 conwert the wide string result to ANSI string */
		for(iN = 0; sWBuf[iN] != 0; iN ++)
		{
			sBuf[iN] = (char)sWBuf[iN];
		}
		sBuf[iN] = '\0';

		/* 1.a.3.1 the config dir is {app}\etc */
		strcpy(sBufTemp, sBuf);
		strcat(sBufTemp, "\\etc");
		INSTALL_ETC = strdup(sBufTemp);
		/* 1.a.3.2 the documentation dir is {app}\doc */
		strcpy(sBufTemp, sBuf);
		strcat(sBufTemp, "\\doc");
		INSTALL_DOC = strdup(sBufTemp);
		/* 1.a.3.3 the shared dir is {app}\share */
		strcpy(sBufTemp, sBuf);
		strcat(sBufTemp, "\\share");
		INSTALL_SHARE = strdup(sBufTemp);

		if (!configdir_default.cd(INSTALL_ETC))    // ERROR: default config dir not found
		{
			derr("config", "No default configuration directory (%s) for Freddy!", INSTALL_ETC);
			derr("config", "Exiting...");
			QApplication::exit(0);
		}
		configdir_user = QDir::root();
	} else
	{
		/* 1.b error during opening the registry key.
		       freddy is probably not installed correctly with inno setup */
		QMessageBox::information(NULL, "Freddy", "Freddy is not installed correctly");
	}

#endif

	dend();
}

void Editor::configLoadFile(Editor::WhichConfigFile eWhich, bool bReload)
{
	QFileInfo fi;
	static QDict<int> *pKeyList = NULL;

	dbeg();

	switch(eWhich)
	{
		case FREDDY:
			/* read freddy.conf */
			fi.setFile(configdir_user, "freddy.conf");
			if (!fi.exists()) fi.setFile(configdir_default, "freddy.conf");
			if (fi.exists())
			{
				EditBufferText::properties->loadConfig(fi.absFilePath().latin1(), bReload);
			}                               // ERROR: freddy.conf is not found!
			else derr("config", "%s (main config file) not found!", fi.absFilePath().latin1());
			break;

		case SHORTCUTS:
			pKeyList = new QDict<int>();

			/* load shortcuts.conf */
			fi.setFile(configdir_user, "shortcuts.conf");
			if (!fi.exists()) fi.setFile(configdir_default, "shortcuts.conf");
			if (fi.exists()) com->loadShortCuts(fi.absFilePath().latin1(), pKeyList);
			else dwarn("config", "%s (shortcuts config) not found!", fi.absFilePath().latin1());
			break;

		case MENU:
			/* load menu.conf */
			fi.setFile(configdir_user, "menu.conf");
			if (!fi.exists()) fi.setFile(configdir_default, "menu.conf");
			if (fi.exists())
			{
				sMenuPath = fi.absFilePath();
				com->loadMenu(fi.absFilePath().latin1(), "editor.menu", pKeyList);
		//		pContextMenu = NULL;
				pContextMenu = com->loadMenu(fi.absFilePath().latin1(), "editor.contextmenu", pKeyList);
				ExtQTabBar::pTabMenu = com->loadMenu(fi.absFilePath().latin1(), "editor.tabmenu", pKeyList);
			} else
			{
				sMenuPath = "";
				dwarn("config", "%s (menu config) not found!", fi.absFilePath().latin1());
			}
			if (pKeyList) delete(pKeyList);
			pKeyList = NULL;
			break;

		case COMMANDLINE:
			/* load commandline.conf */
			fi.setFile(configdir_user, "commandline.conf");     // read commandline.conf
			if (!fi.exists()) fi.setFile(configdir_default, "commandline.conf");
			if (fi.exists())
			{
				oPropCmdline.loadConfig(fi.absFilePath().latin1(), bReload);
			}
			else derr("config", "%s (command line config file) not found!", fi.absFilePath().latin1());
			break;

		case SAVEDSTATE:
			/* load saved-state */
			fi.setFile(configdir_user, "saved-state");      // read saved-state
			if (!fi.exists()) fi.setFile(configdir_default, "saved-state");
			if (fi.exists()) savedstate.loadConfig(fi.absFilePath().latin1());
			else dwarn("config", "%s (last state file) not found!", fi.absFilePath().latin1());
			break;
	}

	dend();
}

void Editor::showCommandLineHelp(void)
{
	dbeg();

	printf("Freddy  %s - (C) 2000-2002 by Ferenc Moricz\n\n", FREDDY_VERSION);
	printf("Options:\n\n");

	//vector<PropertyItem, allocator<PropertyItem> >::iterator oHelpIt;
	map< QString, tPropVect, less<QString> >::iterator oHelpIt;
	for(oHelpIt = oPropCmdline.begin(); oHelpIt != oPropCmdline.end(); oHelpIt++)
	{
		tPropVect oItem = (*oHelpIt).second;
		printf("  %s\t  %s\n", (*oHelpIt).first.latin1(),
		        (oItem[0].getComment() != NULL ? oItem[0].getComment()->latin1() : "")  );
	}

	dend();
}

// loadState()
/**
 * @brief Loads the previously saved state
 */
void Editor::loadState(void)
{
	static char filetag[64];

	dbeg();

	if(EditBufferText::properties->getBoolValue("editor.startup.multi", true))
	for(int sessionn = 0; ; sessionn++)
	{
		sprintf(filetag, "session%d.name", sessionn);
		const QString *sessionname = savedstate.getValue(filetag);
		if (sessionname != NULL)        // create new session
		{
			dmesg("5:config", "Loading SESSION: %s", sessionname->latin1());
			EditSession *session = newSession( sessionname );
			for(int filen = 0; ; filen ++)
			{
				EditBuffer *buffer = NULL;
				sprintf(filetag, "session%d.file%d", sessionn, filen);
				const QString *filename = savedstate.getValue(filetag);
                                        /* load the file */
				if (filename != NULL)
				{
					buffer = loadFile(new QFile(*filename));
					if (buffer != NULL)
					{
						int cursx, cursy, y;
						sprintf(filetag, "session%d.file%d.firstline",
						        sessionn, filen);
						y = savedstate.getNumericalValue(filetag, 0);
						y = max( 0, y );
						y = min( y, (int)buffer->text->lines.size()-1 );
						buffer->text->firstVisible = y;

						sprintf(filetag, "session%d.file%d.cursorline",
						        sessionn, filen);
						cursy = savedstate.getNumericalValue(filetag, 0);
						cursy = max( 0, cursy );
						cursy = min( cursy, (int)buffer->text->lines.size()-1 );
						sprintf(filetag, "session%d.file%d.cursorcolumn",
						        sessionn, filen);
						cursx = savedstate.getNumericalValue(filetag, 0);
						buffer->text->cursor.setX(cursx);
						buffer->text->cursor.setY(cursy);
					}
				}
				else break;
			}
			sprintf(filetag, "session%d.actualfile", sessionn);
			int act = savedstate.getNumericalValue(filetag, 0);
			if (act != -1)
			{
				int iSessionBufCount = session->myTabBar()->count();
				if (iSessionBufCount >= 1)
					session->setCurrentPage(min(act, iSessionBufCount-1));
			}
		} else break;
	}

	sessionMenu(100+savedstate.getNumericalValue("session.actual", 0));
	const QString *laststyle = savedstate.getValue("editor.style");
	if (laststyle != NULL) command("styleset", laststyle->latin1());

	dend();
}


// saveState()
/**
 * @brief Save the current state of the program
 * 
 * The following attributes will be saved:
 *   - the main window's size and position
 *   - the list of the loaded files
 */
void Editor::saveState(void)
{
	dbeg();
	int iOpenMode = (bSavedStateOpened ? IO_ReadWrite | IO_Append : IO_WriteOnly);

	if (pEditorServer == NULL) return;

	QFileInfo fi("/");
	QFile statefile("/");

	fi.setFile(configdir_user, "saved-state");
	statefile.setName(fi.absFilePath());
	if (!statefile.open(iOpenMode))
	{
		fi.setFile(configdir_default, "saved-state");
		statefile.setName(fi.absFilePath());
		if (!statefile.open(iOpenMode))
		{
			derr("config", "Could not create saved-state");
			return;
		}
	}

	dmesg("5:config", "saveState()");

	char prbuf[128];                    // save the actual state

	statefile.writeBlock("## saved-state: the state after the last session\n", 49);
	statefile.writeBlock("## Automatically generated - don't edit by hand!\n\n", 50);

                                        // window position and size
	sprintf(prbuf, "window.x %d\n", pMainWindow->geometry().x());
	statefile.writeBlock(prbuf, strlen(prbuf));
	sprintf(prbuf, "window.y %d\n", pMainWindow->geometry().y());
	statefile.writeBlock(prbuf, strlen(prbuf));
	sprintf(prbuf, "window.width %d\n", pMainWindow->geometry().width());
	statefile.writeBlock(prbuf, strlen(prbuf));
	sprintf(prbuf, "window.height %d\n", pMainWindow->geometry().height());
	statefile.writeBlock(prbuf, strlen(prbuf));

	// save the actual style (if any is selected)
	if (!style.isNull())
	{
		sprintf(prbuf, "editor.style %s\n\n", style.latin1());
		statefile.writeBlock(prbuf, strlen(prbuf));
	}

	int id, sess_n;                             // list of open sessions and files
	vector<QWidget*, allocator<QWidget *> >::iterator p;

	p = find(sessions.begin(), sessions.end(),
	                   stack->visibleWidget()); /* the actual file in the session */
	if (p != sessions.end())
	{
		id = p - sessions.begin();
		sprintf(prbuf, "session.actual %d\n\n", id);
		statefile.writeBlock(prbuf, strlen(prbuf));
	}

	for( id = sess_n = 0; id < (int)sessions.size(); id++ )
	{
		EditSession *session = (EditSession *)sessions[id];
		if ( session == NULL ) continue;

		sprintf(prbuf, "session%d.name ", sess_n);  // name of the session
		statefile.writeBlock(prbuf, strlen(prbuf));
		sprintf(prbuf, "\"%s\"\n", session->name.latin1());
		statefile.writeBlock(prbuf, strlen(prbuf));

		sprintf(prbuf, "session%d.actualfile %d\n",
			               sess_n, session->currentPageIndex());
		statefile.writeBlock(prbuf, strlen(prbuf));
		int iBufIdx, iFileIdx;
		for(iBufIdx = iFileIdx = 0; iBufIdx < (int)session->myTabBar()->count(); iBufIdx++)
		{
			/* get the 'iBufIdx'th file */
#if QT_VERSION > 280
			EditBuffer *buffer = (EditBuffer *)session->page(iBufIdx);
#else
			session->setCurrentPage(iBufIdx);
			EditBuffer *buffer = (EditBuffer *)session->currentPage();
#endif
			QFile *bufferfile = buffer->getFile();
			if (bufferfile == NULL) continue;

			sprintf(prbuf, "session%d.file%d ", sess_n, iFileIdx); // the list of its files
			statefile.writeBlock(prbuf, strlen(prbuf));
			fi.setFile(*buffer->getFile());
			strcpy(prbuf, fi.absFilePath().latin1());
			statefile.writeBlock(prbuf, strlen(prbuf));
			statefile.writeBlock("\n", 1);
                                        /* line numbas */
			sprintf(prbuf, "session%d.file%d.firstline %d\n",
			               sess_n, iFileIdx, buffer->text->firstVisible);
			statefile.writeBlock(prbuf, strlen(prbuf));
			sprintf(prbuf, "session%d.file%d.cursorline %d\n",
			               sess_n, iFileIdx, buffer->text->cursor.y());
			statefile.writeBlock(prbuf, strlen(prbuf));
			sprintf(prbuf, "session%d.file%d.cursorcolumn %d\n",
			               sess_n, iFileIdx, buffer->text->cursor.x());
			statefile.writeBlock(prbuf, strlen(prbuf));
			iFileIdx++;
		}
		statefile.writeBlock("\n", 1);

		sess_n++;
	}

	statefile.writeBlock("\n", 1);      // a clean line at the end

	statefile.close();                  // close file

	dend();
}

// saveModifiedFiles()
/**
 * @brief Ask the user about all of the modified files to save them or not
 */
void Editor::saveModifiedFiles(void)
{
	/* 1.ITER go through all of the sessions */
	for(int iSess = 0; iSess < sessions.size(); iSess++)
	{
		/* 1.1.COND get and check the pointer to the current session */
		EditSession *pSess = (EditSession *)sessions[iSess];
		if (!pSess) break;

		/* 1.2.ITER go through all of the buffers in the current session */
#if QT_VERSION > 280
		for(int iBuff = 0; iBuff < (int)(pSess->count()); iBuff++)
		{
			EditBuffer *pBuff = (EditBuffer *)pSess->page(iBuff);
#else
		for(int iBuff = 0; iBuff < (int)(pSess->oBufferList.size()); iBuff++)
		{
			EditBuffer *pBuff = (EditBuffer *)pSess->oBufferList[iBuff];
#endif
			/* 1.2.1. COND check if the buffer exists and modified */
			if (pBuff && pBuff->isModified())
			{
				/* 1.2.1.1 display the buffer of the unsaved file */
				stack->raiseWidget(stack->id(pSess));
#ifdef _SYSTEM_LINUX_
				pSess->setCurrentPage( iBuff );
#else
				pSess->setCurrentPageIndex( iBuff );
#endif

				/* 1.2.1.2 display the warning */
				printf("modified: %s\n", (pBuff->getFile() ? pBuff->getFile()->name().latin1() : "Untitled"));
				QString sUnsavedMsg;
				sUnsavedMsg.sprintf("File '%s' is modified but not saved!",
					(pBuff->getFile() ? pBuff->getFile()->name().latin1() : "Untitled"));
				int iSaveOrNot = QMessageBox::warning(NULL, "Freddy - Unsaved file", sUnsavedMsg, "Save it", "Discard changes");

				/* 1.2.1.3 COND if user said "Save", then save this file */
				if (iSaveOrNot == 0)
					saveFile();
			}
		}
	}
}


/* Process command line */
QStringList *Editor::processCommandLine(int argc, char *argv[])
{
	dbeg();

	if (argc == 1) return NULL;

	QStringList *filelist = new QStringList();
	QDir *dir = new QDir();

	/* (*) all arguments */
	for(int pi = 1; pi < argc; pi ++)
	{
		if ((!strcmp(argv[pi], "-h")) || (!strcmp(argv[pi], "--help")) ||
		         (!strcmp(argv[pi], "/?")))
		{
			showCommandLineHelp();
			exit(0);
		} else if ((argv[pi][0] == '-')&&(argv[pi][1] == '-'))
		{
			/* look for this argument in the config file */
			char *pToken = argv[pi];
			char *pNo = strstr(argv[pi], "no-");
			if ((pNo == NULL) || (pNo - pToken > 2)) pNo = NULL;

			if (pNo != NULL)
			{
				pToken += 3;
				(*pToken) = '-';
			}

			char *pArg = strstr(pToken, "=");
			if (pArg) (*pArg++) = '\0';

			if (oPropCmdline.find(pToken) != oPropCmdline.end())
			{
				vector<PropertyItem, allocator<PropertyItem> > *pValue = &oPropCmdline[pToken];
				for(int i = 0; i < (int)pValue->size(); i++)
				{
					bool bBoolValue = (pNo == NULL);
					QString stRes(*(*pValue)[i].getString());
					if (stRes.at(0) == '-')
					{
						stRes.remove(0, 1);
						bBoolValue = !bBoolValue;
					}
					if (!pArg)
						EditBufferText::properties->setBoolValue(stRes, bBoolValue);
					else EditBufferText::properties->setValue(stRes, new QString(pArg));
					dmesg("4:config", "commandline: %d %s", bBoolValue, stRes.latin1());
				}
			} else
			{
				derr("config", "No such flag or property '%s'\n", pToken);
				exit(1);
			}
		} else 
		{
			filelist->append(dir->absFilePath(argv[pi]));
		}
	}

	dend();

	return filelist;
}

// Create a new file buffer
void Editor::newFile()
{
	dbeg();

	EditSession *session = (EditSession *)stack->visibleWidget();
                                // If there is no session, create one
	if (session == NULL) session = newSession( );

	session->addBuffer( tr("Untitled") );
	session->show();

	dend();
}


#include <qpainter.h>
// .Editor::loadFile(.)
/**
 *   @brief loads a file, and creates a buffer for it
 *   @param file the file to load
 */                                  
QPixmap *pPixHeader = NULL, *pPixSource = NULL;

EditBuffer * Editor::loadFile(QFile *file, bool bSkipPlacement)
{
	dbeg();

	if (pPixHeader == NULL)
	{
		pPixHeader = new QPixmap( 6, 10);
		pPixHeader->fill( QColor( 195, 195, 80 ) );
		QPainter painter;
		painter.begin( pPixHeader );
		painter.setPen( QColor( 145, 145, 35) );
		painter.drawRect( 0, 0, 6, 10 );
		painter.end();
		pPixSource = new QPixmap( 6, 10);
		pPixSource->fill( QColor( 192, 131, 0) );
		painter.begin( pPixSource );
		painter.setPen( QColor( 142, 81, 0) );
		painter.drawRect( 0, 0, 6, 10 );
		painter.end();
	}

	EditSession *session = (EditSession *)stack->visibleWidget();
	if (session == NULL) session = newSession();

	if (!file->exists()) return NULL;
	QFileInfo fileinfo( *file );
                                        // is it opened now?
	for(int i = 0; i < (int)session->myTabBar()->count(); i++)
	{
#if QT_VERSION > 280
		EditBuffer *buffer = (EditBuffer *)session->page(i);
#else
		session->setCurrentPage(i);
		EditBuffer *buffer = (EditBuffer *)session->currentPage();
#endif

		if (buffer->getFile() == NULL) continue;

		if (file->name() == buffer->getFile()->name()) 
		{
			session->setCurrentPage( i );
			session->show();
			return buffer;
		}
	}
                                        // load the text
	EditBuffer *buffer = session->addBuffer( fileinfo.fileName(), bSkipPlacement );

	/* configure the tab: set pixmap ... */
#if QT_VERSION > 280
	if ((file->name().endsWith( ".h")) || (file->name().endsWith( ".hpp")) ||
	    (file->name().endsWith( ".hxx")))
		session->setTabIconSet( buffer, QIconSet( *pPixHeader ) );
	if ((file->name().endsWith( ".c")) || (file->name().endsWith( ".cpp")) ||
	    (file->name().endsWith( ".cxx")))
		session->setTabIconSet( buffer, QIconSet( *pPixSource ) );
#else
	if (QSTRING_ENDSWITH(file->name(),".h",2) || QSTRING_ENDSWITH(file->name(),".hpp",4) ||
	    QSTRING_ENDSWITH(file->name(),".hxx",4))
		session->changeTab(buffer, QIconSet( *pPixHeader ), session->tabLabel(buffer));
	if (QSTRING_ENDSWITH(file->name(),".c",2) || QSTRING_ENDSWITH(file->name(),".cpp",4) ||
	    QSTRING_ENDSWITH(file->name(),".cxx",4))
		session->changeTab(buffer, QIconSet( *pPixSource ), session->tabLabel(buffer));
#endif
	               /*p ... set tooltip */
	QString tip;
	tip = fileinfo.dirPath( true );
	tip.append("/<b>");
	tip.append( fileinfo.fileName() );
	tip.append("</b>");
#if QT_VERSION > 280
	session->setTabToolTip( buffer, tip ); /*?*/
	if ( !(file->name().endsWith(".c")) && !(file->name().endsWith(".h")) &&
	     !(file->name().endsWith(".c++")) && !(file->name().endsWith(".cpp")) &&
	     !(file->name().endsWith(".cxx")) )
#else
	if ( !QSTRING_ENDSWITH(file->name(),".c",2) && !QSTRING_ENDSWITH(file->name(),".h",2) &&
	     !QSTRING_ENDSWITH(file->name(),".c++",4) && !QSTRING_ENDSWITH(file->name(),".cpp",4) &&
	     !QSTRING_ENDSWITH(file->name(),".cxx",4) )
#endif
	{
		buffer->bSyntaxHighlighting = false;
	}

#ifdef COLORER_FILTER
	buffer->bSyntaxHighlighting = false;
#endif

	buffer->loadFile(*file);

	session->show();

	dend();

	return buffer;
}

// File/Open
#include <qfiledialog.h>
#include <qfileinfo.h>
#include <qmessagebox.h>
#include <stdio.h>

void Editor::openFile(QStringList *paramlist)
{
	bool bReturn = false;

	dbeg();

	/* 1. COND is the parameter list clear? */
	if (paramlist == NULL)
	{
		QFileInfo *fileinfo = new QFileInfo();
		EditSession *session = (EditSession *)stack->visibleWidget();
	                                // If there is no session, create one
		if (session == NULL) session = newSession( );

		QString *currentdir = NULL;         // get the current buffer's dir
		if (session->myTabBar()->count()) 
		{
			EditBuffer *buffer = (EditBuffer *)session->currentPage();
			if ((buffer != NULL) && (buffer->getFile() != NULL))
			{
				fileinfo->setFile( *(buffer->getFile()) );
				currentdir = new QString( fileinfo->dirPath() );
			}
		}
		if (currentdir == NULL) currentdir = new QString( "" );

                                        // open the files
		QStringList filelist = QFileDialog::getOpenFileNames(
		           tr("C/C++ sources (*.c *.cpp *.h);;All files (*)"),
	               *currentdir, pMainWindow, tr("Open file..."), tr("Open File: Select one or more files") );

		for(QStringList::Iterator it = filelist.begin(); it != filelist.end(); ++it)
		{
			QFileInfo fi = QFileInfo(*it);
			QFile *file=new QFile(*it);
			loadFile(file);
		}
	} else
	{
		/* 1n we got at least 1 parameter
		     - ITER go through the parameters */
		for(QStringList::Iterator it = paramlist->begin(); it != paramlist->end(); ++it)
		{
			/* 1n.1 get the name of the n-th file */
			QFile oFile(*it);
			QFileInfo oFI(oFile);
			QString sFilePath = oFI.absFilePath();

#ifdef _SYSTEM_WINDOWS_
			/* 1n.1. (WIN) convert the short file name to long file name
			           (e.g "C:\Progra~1\text.txt"  ->  "C:\Program Files\text.txt) */
			WCHAR sWLongPath[1024], sWShortPath[1024];
			int iN;

			/* 1n.1.2 sFilePath -> sWShortPath - copy the file path string (QString -> windows wide string (WCHAR *)) */
			for(iN = 0; iN < sFilePath.length(); iN++)
			{
				sWShortPath[iN] = sFilePath.at(iN).unicode();
			}
			sWShortPath[iN] = 0;

			/* 1n.1.3 get the long name of the path */
			int iRet;
			if (GetLongPathName(sWShortPath, sWLongPath, 256) == 0)
			{
				int iLastErr = GetLastError();
				if (iLastErr != ERROR_PATH_NOT_FOUND)
					QMessageBox::warning(this->pMainWindow, "Freddy problem",
							"Internal problem (converting file name string)",
						   QMessageBox::Yes|QMessageBox::Default, 0);
				bReturn = true;
			}

			/* 1n.1.4 sWLongPath->sFilePath - copy the file path string (windows wide string (WCHAR *) -> QString) */
			if (!bReturn)
			{
				sFilePath = "";
				for(iN = 0; sWLongPath[iN] != 0; iN ++)
				{
					sFilePath.append(QChar(sWLongPath[iN]));
				}
			}

			/* 1n.1.5 reopen the file with the long name */
			if (!bReturn)
			{
				oFile.setName(sFilePath);
				oFI.setFile(oFile);
			}
#endif

			if (!bReturn)
			{
				/* 1n.2 COND does the file exist? (and is it a real file? (not directory)) */
				if (!oFile.exists() || ((!oFI.isFile()) && (!oFI.isSymLink())))
				{
					/* 1n.2n file does not exist*/
					/* 1n.2n. ask the user if he/she wants to create the file */
					char sCannotOpen[512];
					sprintf(sCannotOpen, "Cannot open '%s'! <br><b>Create</b> it?", (*it).latin1());
					QMessageBox mb("Freddy", sCannotOpen,
								   QMessageBox::Warning, QMessageBox::Yes|QMessageBox::Default,
								   QMessageBox::No|QMessageBox::Escape, QMessageBox::NoButton,
								   pMainWindow, tr("New file"), true);
					mb.setMinimumWidth(400);
					if (mb.exec() == QMessageBox::Yes)
					{
						/* 1n.2n.y - create the file */
						oFile.open( IO_WriteOnly );
						oFile.close();
					}
					/* 1n.2n.n - user refused to create the file -> do nothing */
					else continue;
				}

				/* 1n.3 load the file */
				loadFile(&oFile);
			}
		}
	}

	dend();
}

void Editor::openCommandlineFile(QStringList *pFileList)
{
	const QString *pCProp;
	bool bLoad = false;

	dbeg();

	pCProp = EditBufferText::properties->getValue("editor.commandline.opento", NULL );

	if ( (pCProp == NULL) || (*pCProp == "actualsession") )
	{
		if (!sessions.size()) newSession();
		bLoad = true;
	}
	else if ( *pCProp == "newsession" )
	{
		newSession();
		bLoad = true;
	} else if ( pCProp->startsWith("session ") )
	{
		char *sProp = strdup(pCProp->latin1()), *sName;
		int i;
		strtok(sProp, " \t\n\r");
		sName = strtok(NULL, " \t\n\r");

		if (sName != NULL)
		{
			for(i = 0; i < (signed)sessions.size(); i++)
				if (((EditSession *)sessions[i])->name == sName)
				{
					stack->raiseWidget(sessions[i]);
					break;
				}
		}
		if ((sName == NULL) || (i >= (signed)sessions.size()))
			newSession();

		bLoad = true;
	}

	if (bLoad && (pEditorServer || !pEditorClient)) openFile(pFileList);

	dend();
}

void Editor::saveFile()
{
	dbeg();

	EditSession * es = (EditSession *)stack->visibleWidget();
	if ((es == NULL) || ( !es->myTabBar()->count() )) return;
	EditBuffer *buffer = (EditBuffer *)es->currentPage();
	if (buffer->getFile() == NULL)              // is it untitled yet?
		saveasFile();
	else if (! buffer->getFile()->exists())
		saveasFile();
	else buffer->saveFile();

	dend();
}

void Editor::saveasFile(const char *sNewName)
{
	bool bReturn = false;
	EditSession *pSession = NULL;
	EditBuffer *pBuffer = NULL;
	QString sFileName;

	dbeg();

	/* 1. get the current buffer */
	pSession = (EditSession *)stack->visibleWidget();
	if (!pSession || ( !pSession->myTabBar()->count() ))
		bReturn = true;
	if (!bReturn)
	{
		if ((pBuffer = (EditBuffer *)pSession->currentPage()) == NULL)
			bReturn = true;
	}

	/* 2. "Save as..." dialog - query for the new file's name */
	if (!bReturn)
	{
		QString sOldFileName = QString::null;
		if (pBuffer->getFile() != NULL)
		{
			QFileInfo fi(*(pBuffer->getFile()));
			sOldFileName = fi.absFilePath();
		}
		QString sFileType = "All files, latin1 (*)";
		sFileType.append(";;All files, UTF-8 (*)");

		if (!sNewName)
			sFileName = QFileDialog::getSaveFileName(sOldFileName, sFileType,
				pMainWindow, tr("Save file as..."), tr("Save file as...") );
		else sFileName = sNewName;
		if (sFileName == QString::null) bReturn = true;
	}

	/* 3. create an object for the file */
	QFile *pFile;
	if (!bReturn)
		pFile=new QFile( sFileName );

	/* 4. COND if this file exists query for whattodo (yes - overwrite, no - load it, 0 - cancel) */
	int iWhatto = 0;
	if (!bReturn && pFile->exists()) 
		iWhatto = QMessageBox::warning(NULL, "Freddy", 
		               "This file already exists.<br>" "<b>Overwrite</b> it?", "Overwrite",
		               "Load it", "Cancel", 0, 2);

	/* 4y. write to the file */
	if (!bReturn && (iWhatto == 0))
	{                                                                   
		pFile->open(IO_WriteOnly);

		if (pFile->exists())
		{
			pFile->close();
			QFileInfo fi(*pFile);
			pSession->changeTab(pBuffer, fi.fileName());
			pBuffer->setFile(pFile);
			pBuffer->saveFile();
		}
	}

	/* 4n. load this file */
	else if (!bReturn && (iWhatto == 1))
	{
		QFileInfo fi(*pFile);
		QStringList oOpenFiles;
		oOpenFiles.append(fi.absFilePath());
		openFile(&oOpenFiles);
		return;
	}

	dend();
}

void Editor::closeFile()
{
	dbeg();

	EditSession *session = (EditSession *)stack->visibleWidget();
	session->removeBuffer();

	dend();
}

/* SESSION MANEGEMENT */
//  . Editor :: newSession (.)
/**
 * @brief Creates a new session
 * @param name the session's name
 * @return the pointer to the new session
 *    
 * This function:
 * - creates a new session
 * - put it in the Session menu
 */
EditSession *Editor::newSession(const QString *name)
{
	dbeg();

	EditSession *session = new EditSession(stack);  // create new session
	if (name != NULL) session->name = *name;
	else session->name = tr("Untitled");
	sessions.push_back(session);

	if (sessionmenu != NULL)                // insert to the menu
		sessionmenu->insertItem( session->name, sessionid,
		                         sessionmenuindex+sessions.size()-1 );

	stack->addWidget( session, sessionid ); // put to the stack
	sessionMenu( sessionid );

	session->show();

	sessionid++;

	dend();

	return session;
}

//  Editor :: closeSession (.)
/**
 * @brief Closes the session
 * @param session a pointer to the session
 *    
 * This function:
 * - closes the session
 * - removes the session from the Session menu
 */
void Editor::closeSession( EditSession *session )
{
	dbeg();
                                        /* default: actual session */
	if ( session == NULL) session = (EditSession *) stack->visibleWidget();

	if (sessionmenu != NULL)            /* remove session from the menu */
	dmesg("5:session", "In Editor::closeSession session is %p",session);
		sessionmenu->removeItem( stack->id(session) );

	/* cleanup buffers ( by Timofej Dod ) */
#if QT_VERSION > 280
	for(int i = ((int)(session->count()) - 1); i >= 0; i--)
	{
		EditBuffer *buffer = (EditBuffer *)session->page(i);
#else
	for(int i = ((int)(session->oBufferList.size()) - 1); i >= 0; i--)
	{
		session->setCurrentPage(i);
		EditBuffer *buffer = (EditBuffer *)session->currentPage();
#endif
		dmesg("5:session", "In Editor::closeSession deleting buffer %p",buffer);
		session->removePage(buffer);
		buffer->close(true);
	}

	vector<QWidget*, allocator<QWidget *> >::iterator p;
	p = find( sessions.begin(), sessions.end(), session);
	int n = p - sessions.begin();
	if (p != sessions.end()) sessions.erase( p ); /* delete session */
	stack->removeWidget( session );
	session->close(true);

	if (sessions.size())                /* bring another session to front */
	{
		n %= sessions.size();
		sessionMenu( stack->id(sessions[n]) );
	}

	dend();
}

//  Editor :: renameSession (..)
/**
 * @brief Renames a session
 * @param name the new name
 * @param session a pointer to the session (if it is NULL
 *        the actual session will be used)
 */
#include <qinputdialog.h>
void Editor::renameSession( QString *name, EditSession *session )
{
	dbeg();

	if ( session == NULL ) session = (EditSession *) stack->visibleWidget();
	if ( session == NULL ) return;

	if (name == NULL)
	{
		bool ok = false; 

		name = new QString(QInputDialog::getText(tr("Rename session"),  //! mem loss?
		          tr("Please enter new name of this session:"), QLineEdit::Normal,
		          (session!=NULL?session->name : QString::null), &ok, pMainWindow));
		if (!ok || name->isEmpty()) return;
	}

	session->name = *name;
	if (sessionmenu != NULL)
		sessionmenu->changeItem(stack->id(session), *name);

	dend();
}

//  Editor :: checkActualSession()
/**
 * @brief Simply checks the actual session
 */
void Editor::checkActualSession()
{
	dbeg();

	int nid, i, actid = stack->id( stack->visibleWidget() );

	if (sessionmenu != NULL)
		for(i = 0; i < (int)sessions.size(); i++)
		{
			nid = stack->id( sessions[i] );
			sessionmenu->setItemChecked(nid, (nid == actid ? true : false));
		}

	dend();
}

//  Editor :: sessionMenu (.)
/**
 * @brief The Session menu's event function
 * @param id the ID of the activated menu
 *
 * This function handles the functions of the Session menu.\n
 * The possible values of the \em id variable:
 *   - 0 : "New session" menu activated
 *   - 1 : "Close session" menu activated
 *   - 2 : "Rename session" menu activated
 *   - greater than 100: a session with the given ID activated
 */
void Editor::sessionMenu(int id)
{
	dbeg();

	if (id>=100) stack->raiseWidget( id );
	checkActualSession();
	if (com) com->updateMenuItems();

	dend();
}

EditorServer::EditorServer( int iPort ) : QServerSocket( iPort )
{
	dbeg(); 
	/* Unable to bind - network configuration error */
	if (!ok()) derr("freddy", "Unable to bind to port %d - network error!", iPort);
	dend();
}

EditorSocket::EditorSocket()
{
	dbeg();

	QObject::connect( this, SIGNAL(connected( void )),
						this, SLOT( slotConnected( void ) ) );
	QObject::connect( this, SIGNAL(connectionClosed( void )),
						this, SLOT( slotClosed( void ) ) );
	QObject::connect( this, SIGNAL(readyRead( void )),
						this, SLOT( slotRead( void ) ) );
	QObject::connect( this, SIGNAL(bytesWritten( int )),
						this, SLOT( slotWrote( int ) ) );
	QObject::connect( this, SIGNAL(error( int )),
						this, SLOT( slotError( int ) ) );

	dend();
}

void EditorServer::newConnection(int iSocket)
{
	dbeg();

	EditorSocket *pSocket = new EditorSocket();
	pSocket->iState = 0;
	pSocket->setSocket( iSocket );

	dend();
}

QSocket *EditorServer::locateServer( int iPort )
{
	dbeg();

	Editor::editor->pEditorClient = new EditorSocket();
	Editor::editor->pEditorClient->iState = 10;
	Editor::editor->pEditorClient->connectToHost( "127.0.0.1", iPort );

	dend();

	return NULL;
}

void EditorSocket::slotConnected( void )
{
	dbeg();

	dmesg("5:freddy", "Freddy server found! (=> running in client mode)");

	if ((Editor::editor->pCommandFileList != NULL) &&
	    (Editor::editor->pCommandFileList->count() > 0))
		Editor::editor->startUp();
	else {
		derr("startup", "No files specified to load in the existing session!");
		exit(0);
	}

	dend();
}

void EditorSocket::slotClosed( void )
{
	dbeg();

	dmesg("startup", "Communication slot is Closed!");
	delete(this);
	if (Editor::editor->pEditorServer)
	{
//		Editor::editor->setActiveWindow();
/*
		Editor::editor->hide();
		Editor::editor->show();*/
	}

	dend();
}

void EditorSocket::slotRead( void )
{
	dbeg();

	if (canReadLine())
	{
		QString oLine = readLine();
#if QT_VERSION > 280
		while( oLine.endsWith("\n") || oLine.endsWith("\r") )
#else
		while( QSTRING_ENDSWITH(oLine,"\n",1) || QSTRING_ENDSWITH(oLine,"\r",1) )
#endif
			oLine.remove(oLine.length()-1, 1);
		oLine.simplifyWhiteSpace();
		QString oContent = oLine.mid( 6 );
		dmesg("startup", "< %s", oLine.latin1());
		if (oLine.startsWith("OPEN"))
		{
			QStringList *filelist = new QStringList();
			filelist->append( oContent );
			Editor::editor->openCommandlineFile( filelist );
			delete( filelist );
//			Editor::editor->setActiveWindow();
		}
	}

	dend();
}

void EditorSocket::slotWrote( int iWroteN )
{
	dbeg();

	if (iState == 20)
	{
		if (Editor::editor->pCommandFileList->count() > 0)
		{
			QString &pFileName = Editor::editor->pCommandFileList->first();
			QString oComLine("OPEN  ");
#if defined(_SYSTEM_LINUX_)
			if (pFileName.at(0) != '/')
#elif defined(_SYSTEM_WINDOWS_)
			if ((pFileName.at(0) != '\\') && (pFileName.at(1) != ':'))
#endif
			{				           /* not absolute path */
				oComLine.append(QDir::current().absPath());
				oComLine.append("/");
			}
			oComLine.append(pFileName);
			QStringList::Iterator it = Editor::editor->pCommandFileList->begin();
			Editor::editor->pCommandFileList->remove(it);
			oComLine.append("\n");
			writeBlock(oComLine.latin1(), oComLine.length());
		} else
		QApplication::exit(0);
	}

	dend();
}

void EditorSocket::slotError( int iError )
{
	dbeg();

	if (iState == 10)
	{
		dmesg("startup:", "Freddy server not found! (=> running in server mode)");

		if (Editor::editor->pEditorServer == NULL)
		{
			Editor::editor->pEditorServer = new EditorServer( FREDDYPORT );
		}
		iState = 0;

		Editor::editor->startUp();
	}

	dend();
}

// . Editor :: command (....)
/**
 * @brief Editor's command function
 * @param commandstr the name of the command
 * @param arg1-arg3 the 1-3. argument
 * @return the result of the command
 */
#if QT_VERSION > 280
#include <qstylefactory.h>
#else
#include <qwindowsstyle.h>
#include <qcdestyle.h>
#include <qmotifstyle.h>
#include <qplatinumstyle.h>
#endif

char *Editor::command(const char *commandstr,
			const char *arg1, const char *arg2, const char *arg3 )
{
	char **command;
	int comnum;
	bool bCheck = false;

	dbeg();

	if (strstr(commandstr, "CHECK ") == commandstr)
	{
		commandstr += 6;
		bCheck = true;
	}

	EditSession *session;
	EditBuffer *buffer;

                                /* get actual session and buffer */
	session = (EditSession *)Editor::editor->getStack()->visibleWidget();
	if (session != NULL) buffer = (EditBuffer *)session->currentPage();
	else buffer = NULL;

	enum { CM_ADD_RULE, CM_PRINT_RULES, CM_PRINT_TREE,
	       CM_FILENEW, CM_FILEOPEN, CM_FILESAVE, CM_FILESAVEAS, CM_FILECLOSE,
	       CM_STYLESET,
	       CM_SESSIONNEW, CM_SESSIONRENAME, CM_SESSIONCLOSE,
	       CM_SESSIONNEXT, CM_SESSIONPREV, CM_BUFFERNEXT, CM_BUFFERPREV,
	       CM_LANGUAGESET,
	       CM_SHOW,
	       CM_PREFERENCES,
	       CM_EXIT };
	static char *comcache[]= {
	       "add.rule", "print.rules", "print.tree",
	       "filenew", "fileopen", "filesave", "filesaveas", "fileclose",
	       "styleset",
	       "sessionnew", "sessionrename", "sessionclose",
	       "sessionnext", "sessionprevious", "buffernext", "bufferprevious",
	       "languageset",
	       "show",
	       "preferences",
	       "exit",
	       NULL };

	if (!bCheck)
		dmesg("6:editor.command", "%s(%s,%s,%s)", commandstr, arg1, arg2, arg3);

	//if (strstr(commandstr, "CHECK ") != commandstr) com->updateMenuItems();

	                                    // Search for the command
	for(command=comcache, comnum=0; (*command)!=NULL; command++, comnum++)
		if (0 == strcmp( commandstr, *command )) break;

	if (*command == NULL)               // Invalid command
	{
		dwarn("4:command", "Invalid command: %s!", commandstr);
		return NULL;
	}

	switch( comnum )                    // Execute the command
	{
		case CM_ADD_RULE:
			debug_add_rule(arg1);
			break;
		case CM_PRINT_RULES:
			debug_print_rules();
			break;
		case CM_PRINT_TREE:
			debug_print_tree();
			break;
		case CM_FILENEW:
			if (bCheck) return (char *)1;
			newFile();
			break;
		case CM_FILEOPEN:
			if (bCheck) return (char *)1;
			openFile(NULL);
			break;
		case CM_FILESAVE:
			if (bCheck) return (buffer ? (char *)1: NULL);
			saveFile();
			break;
		case CM_FILESAVEAS:
			if (bCheck) return (buffer ? (char *)1: NULL);
			if (bCheck) return NULL;
			saveasFile(arg1);
			break;
		case CM_FILECLOSE:
			if (bCheck) return (buffer ? (char *)1: NULL);
			if (bCheck) return NULL;
			closeFile();
			break;
		case CM_STYLESET:
			if (bCheck) return (char *)1;
#if QT_VERSION > 280 /* !*/
			dmesg("5:system", "Available library's path: %s\n", QApplication::libraryPaths().join(", ").latin1());
			dmesg("5:system", "Available styles: %s\n", QStyleFactory::keys().join( ", " ).latin1());
			/* check the current style in the menu */
			if (com->pStyleListMenu)
				for(int i = 0; 1; i++)
				{
					int iID = com->pStyleListMenu->idAt( i );
					if (iID == -1) break;
					com->pStyleListMenu->setItemChecked( iID,
						com->pStyleListMenu->text( iID ) == arg1);
				}
			QApplication::setStyle( QStyleFactory::create( arg1 ) );
			style = arg1;
#else
			if (com->pStyleListMenu)
				for(int i = 0; 1; i++)
				{
					int iID = com->pStyleListMenu->idAt( i );
					if (iID == -1) break;
					com->pStyleListMenu->setItemChecked( iID,
						com->pStyleListMenu->text( iID ) == arg1);
				}
			{
				QStyle *pStyle;
				if (!strcmp(arg1, "Windows")) pStyle = new QWindowsStyle();
				else if (!strcmp(arg1, "CDE")) pStyle = new QCDEStyle();
				else if (!strcmp(arg1, "Motif")) pStyle = new QMotifStyle();
				else if (!strcmp(arg1, "Platinum")) pStyle = new QPlatinumStyle();
				if (pStyle) QApplication::setStyle( pStyle );
				if (pStyle) delete(pStyle);
				style = arg1;
			}
#endif
			break;
		case CM_SESSIONNEW:
			if (bCheck) return (char *)1;
			newSession();
			break;
		case CM_SESSIONCLOSE:
			if (bCheck) return (session ? (char *)1: NULL);
			closeSession();
			break;
		case CM_SESSIONRENAME:
			if (bCheck) return (session ? (char *)1: NULL);
			renameSession();
			break;
		case CM_SESSIONNEXT:            // go to the next / prev session
		case CM_SESSIONPREV:
			if (bCheck) return (session ? (char *)1: NULL);

			{
				vector<QWidget*, allocator<QWidget *> >::iterator p;
				p = find(sessions.begin(), sessions.end(), stack->visibleWidget());
				if (p != sessions.end())
				{
					int n = p - sessions.begin();
					n += sessions.size() + (comnum == CM_SESSIONNEXT ? 1 : -1);
					n %= sessions.size();
					sessionMenu( stack->id(sessions[n]) );
				}
			}
			break;
		case CM_BUFFERNEXT:             // go to the next / prev buffer
		case CM_BUFFERPREV:
			if (bCheck) return (buffer ? (char *)1: NULL);

			{
				int n;
				EditSession *session = (EditSession *)stack->visibleWidget();
				n = session->currentPageIndex() + session->myTabBar()->count();
				n = (n + (comnum == CM_BUFFERNEXT ? 1 : -1));
				n %= session->myTabBar()->count();
#ifdef _SYSTEM_LINUX_
				session->setCurrentPage( n );
#else
				session->setCurrentPageIndex( n );
#endif
				session->show();
#ifdef _SYSTEM_WINDOWS_
/*????*/		((EditBuffer *)session->currentPage())->text->setFocus();
#endif
			}
			break;
		case CM_LANGUAGESET:            /* load the appropriate translator file */
			if (bCheck) return (char *)1;

			{
				QDict<int> keylist;
				QString sFile, sDir;
				if (pTranslator != NULL)
				{
					pApplication->removeTranslator(pTranslator);
					pTranslator = NULL;
				}

				if (strcmp(arg1, "en")) // english? -> don't load, use orig. texts
				{
					sFile.append("freddy_");
					sFile.append(arg1);
					sFile.append(".qm");
					if (true)
					{
						sDir = INSTALL_SHARE;
						sDir.append("/i18n");
					}
					//?? else sDir.append("share/i18n");
					pTranslator = new QTranslator(NULL);
					if (pTranslator->load(sFile, sDir))
						pApplication->installTranslator(pTranslator);
					else derr("config", "Translator file (%s/%s) not found!",
					                 sDir.latin1(), sFile.latin1());
				}

				/* reload the main menu */
				if (sMenuPath != "")
				{
					com->loadMenu(sMenuPath.latin1(), "editor.menu", &keylist);
					/* reload the context menu */
					pContextMenu = com->loadMenu(sMenuPath.latin1(), "editor.contextmenu", &keylist);
					/* reload the tab menu */
					ExtQTabBar::pTabMenu = com->loadMenu(sMenuPath.latin1(), "editor.tabmenu", &keylist);
				}
				com->updateLabels();
			}
			break;
		case CM_SHOW:
			if (bCheck) return (char *)1;

			{
				if (!strcmp(arg1, "about"))
				{
					char sAboutStr[512];
					sprintf(sAboutStr, "<center><b>Freddy</b> "
					     "<i>(<b>Fr</b>anc's <b>ed</b>itor), </i>"
					     "version %s_b%d</center><br>"
					  "<center>ascii text editor "
					     "for Linux/Win32</center><br>"
					  "<center>Homepage: http://freddy.sf.net</center>"
					  "<center>(C) 2000-2003 Ferenc Moricz</center>",
					  FREDDY_VERSION, FREDDY_BUILD_NUMBER);
					QMessageBox mb("About Freddy",
					               tr(sAboutStr),
					               QMessageBox::Information, QMessageBox::Ok|QMessageBox::Default,
					               QMessageBox::NoButton, QMessageBox::NoButton,
					               pMainWindow, tr("About Freddy"), false);
					mb.setFocus();
					mb.resize(350, 170);
					mb.exec();
				}
			}
			break;
		case CM_PREFERENCES:
			if (bCheck) return (char *)1;

			{
				pPreferences->show();

				//EditBufferText::properties->preferences();
//p				ConfiguratorItem::pPreferences = new Preferences();
//p				ConfiguratorItem::pPreferences->show();
			}
			break;
		case CM_EXIT:
			if (bCheck) return (char *)1;

			QApplication::exit(0);
			break;

		default: dwarn("command", "Unimplemented command: %s", commandstr);
	}

	dend();

	return NULL;
}

