/*
 * Modules.cpp
 *
 * Loadable modules support for Freddy
 * 
 * Author: Ferenc Moricz
 * $Revision: 1.31 $
 * $Date: 2004/02/22 00:12:50 $
 *
 */

#include "Debug.h"
#include "Module.h"
#include "EditBufferText.h"
#include "Editor.h"

#if defined(_SYSTEM_LINUX_)
#include <dlfcn.h>
#include <qdir.h>
#include <qstringlist.h>
#elif defined(_SYSTEM_WINDOWS_)
#include <windows.h>
#include <winbase.h>

#pragma warning( disable: 4660 )
#pragma warning( disable: 4786 )
#endif

Module *ModuleLoader::pFilterModule = NULL;

ModuleLoader::ModuleLoader()
{
	pFilterModule = NULL;
	init();
}

/*
 * ModuleLoader::init()
 *
 * Initializes the module loader, and loads the modules
 * -RESTR- Only one initialization is possible
 */

#include <stdio.h>
void ModuleLoader::init()
{
	const QString *path;

	if (!(path = EditBufferText::properties->getValue("module.path", NULL)))
	{
		path = new QString(QString(INSTALL_SHARE)+QString("/modules"));
	}

	dmesg("4:plugin", "home directory: %s", path->latin1());
	loadModules(path->latin1(), 0);

	dmesg("4:plugin", "Loading modules..."); fflush(stdout);
}

void ModuleLoader::loadModules(const char *sFile, int iLevel)
{
	static QFileInfo *pFInfo = NULL;

	if (pFInfo == NULL) pFInfo = new QFileInfo();

	pFInfo->setFile(sFile);

	if ((!pFInfo->exists() || !pFInfo->isDir()) && (iLevel == 0))
	{
		derr("plugin", "Couldn't find module directory! (%s)", pFInfo->absFilePath().latin1());
		return;
	}

	if (!pFInfo->exists()) return;
	else if ((pFInfo->fileName() == ".") || (pFInfo->fileName() == "..")) return;

	if (pFInfo->isDir())
	{
		QDir dir(pFInfo->absFilePath(), "*");
		QStringList list = dir.entryList();

		dir.setMatchAllDirs( true );
		for(int n=0; n < (int)list.count(); n++)
		{
			QValueListIterator<QString> sStr = list.at(n);
			if (sStr != NULL)
				loadModules(dir.absFilePath((*sStr).latin1()), iLevel+1);
		}
	}
#ifdef _SYSTEM_LINUX_
#if QT_VERSION > 280
	else if (pFInfo->isFile() && pFInfo->fileName().endsWith(".so"))
#else
	else if (pFInfo->isFile() &&
	         pFInfo->fileName().mid(pFInfo->fileName().length()-3) == ".so")
#endif
	{
		void *handle;
		char *error;
		Module *pModule;
	                                                  // open dynamic library
		handle = dlopen (sFile, RTLD_NOW|RTLD_GLOBAL);
		dmesg("5:plugin", "Opening plugin: %s", sFile);

		if (!handle)
		{
			derr("plugin", "%s - unable to open plugin (error: %s)", sFile, dlerror());
			return;
		}

		/* load symbols from the ld module */
		pModule = new Module();
		pModule->pQuery = moduleQuery;

		pModule->pInit = (pModuleInit)dlsym(handle, "module_init");

		if ((error = dlerror()) != NULL)
		{
			derr("plugin", "%s - module_init() not found (error: %s)\n"
			        "Possibly corrupt plugin binary", sFile, dlerror());
			return;
		}

		pModule->pFunction = (pModuleFunction)dlsym(handle, "module_function");

		if ((error = dlerror()) != NULL)
		{
			derr("plugin", "%s - module_function() not found (error: %s)"
			        "Possibly corrupt plugin binary", sFile, dlerror());
			return;
		}

		/* run module_init() and put module into the modules' array */
		pModule = pModule->pInit(pModule);
		if (pModule->qsName) dmesg("4:plugin", "Plugin is loaded. Name: %s", pModule->qsName->latin1());
		if (pModule == NULL) return;

		modules.push_back(pModule);
	}

#elif defined (_SYSTEM_WINDOWS_)
#if QT_VERSION > 280
	else if (pFInfo->isFile() && pFInfo->fileName().endsWith(".dll"))
#else
	else if (pFInfo->isFile() &&
	         pFInfo->fileName().mid(pFInfo->fileName().length()-4) == ".dll")
#endif
	{
		/* load module */
		unsigned short sWStr[100];
		for(int i = 0; ((i==0) ||(sWStr[i-1]!='\0'))&&(i<100); i++)
			if (sFile[i] != '/') sWStr[i] = sFile[i];
			else sWStr[i] = '\\';

//		HINSTANCE iModuleInstance = LoadLibrary(sWStr);
		HINSTANCE iModuleInstance = LoadLibraryEx(sWStr, NULL, 0);
		if (iModuleInstance == NULL)
		{
			printf("Unable to load module: %s (error: %d)\n", sFile, GetLastError());
			return;
		} else printf("Opened lib file: %s\n", sFile);

		/* load symbols from the DLL module */
		Module *pModule;
		pModule = new Module();
		pModule->pQuery = moduleQuery;

		pModule->pInit =
			(pModuleInit)GetProcAddress(iModuleInstance, "module_init");

		if ((pModule->pInit == NULL) && (GetLastError() != 0))
		{
			printf("%s - module_init() not found (error: %d)",
				sFile, GetLastError());
			return;
		}

		pModule->pFunction =
			(pModuleFunction)GetProcAddress(iModuleInstance, "module_function");

		if ((pModule->pFunction == NULL) && (GetLastError() != 0))
		{
			printf("%s - module_function() not found (error: %d)\n",
				sFile, GetLastError());
			return;
		}

		/* run module_init() and put module into the modules' array */

		pModule = pModule->pInit(pModule);
		printf("modcont: %d\n", modules.size());
		modules.push_back(pModule);
		printf("modcont: %d\n", modules.size());
		fflush(stdout);
	}
#endif

	/* searching for the filter plugin (there can be only 1 present) */
	if (iLevel == 0)
	{
		int iFiltModN;
		bool bFound = false;
		for(iFiltModN = 0; iFiltModN < (signed)modules.size(); iFiltModN++)
		{
			void *pRet;
			if ((pRet = modules[iFiltModN]->pFunction(modules[iFiltModN],
			          "module.isfilter", NULL, NULL, -1, -1)))
			{
				modules[iFiltModN]->pFunction(modules[iFiltModN],
				           "filter.init", NULL, NULL, -1, -1);
				bFound = true;
				break;
			}
		}
		pFilterModule = (bFound ? modules[iFiltModN] : NULL);

		if (!bFound)
			dwarn("plugin", "Filter module (colorer plug-in) not found!");
	}
}

void ModuleLoader::broadcastEvent(const QString &eventName,
                  QString *qsParam1, QString *qsParam2,
                  int iParam1, int iParam2)
{
	/* broadcast this event to every module */
	for(unsigned int i = 0; i < modules.size(); i++)
		modules[i]->pFunction(modules[i], eventName, qsParam1, qsParam2,
		                                        iParam1,  iParam2);
}

// .ModuleLoader::command(....)
/**
 * Command interface for calling module functions
 * @param plugin the called plugin's name
 * @param command the command to execute
 * @param arg1 first argument
 * @param arg2 second argument
 * @param arg3 third argument
 * @return the result of the command
 */
char *ModuleLoader::command( const char *plugin, const char *command, const char *arg1,
				   const char *arg2, const char *arg3 )
{
	int iN;
	void *pRETURN;

	if (command != NULL)
	{
		/* search for the appropriate module (full search by name) */
		for(iN = 0; iN < ((int)modules.size()); iN++)
		{
			if (*(modules[iN]->qsName) == plugin)
			{
				dmesg("5:plugin", "command: Executing module %s, function %s\n", plugin, command);
				QString qsCommand(command);
				Module *pModule = modules[iN];
				pRETURN = pModule->pFunction(pModule, qsCommand, NULL, NULL, -1, -1);
				break;
			}
		}
	}

	return (char *)pRETURN;
}


#include <stdlib.h>
// .moduleQuery(....)
/**
 * Function which provide access to the main app for the modules.
 */
void *ModuleLoader::moduleQuery( const char *sCommand,
         const char *sArg1, const char *sArg2, const void *pArg )
{
	static void *pReturn;
	static EditSession *pSession;
	static EditBuffer *pBuffer;

	pReturn = NULL;

	if (strstr(sCommand, "cb.") == sCommand)
	{
		static QWidgetStack *pStack;
		pStack = Editor::editor->getStack();

		if (pStack != NULL)
		{
			pSession = (EditSession *)pStack->visibleWidget();
			if (pSession != NULL)
				pBuffer = (EditBuffer *)pSession->currentPage();
		}

		if (!strcmp(sCommand+3, "get.buffer.lines.at"))
		{
			int iLine = atoi( sArg1 );
			if (pBuffer && pBuffer->text)
			{
				if (iLine < (int)pBuffer->text->lines.size()) 
					pReturn = (void *)pBuffer->text->lines[iLine];
			}
		}
		else if (!strcmp(sCommand+3, "get.ptr.buffer"))
			pReturn = pBuffer;
		else if (!strcmp(sCommand+3, "set.widget.mapholder"))
		{
			if ((pArg != NULL) && (pBuffer != NULL) && (pBuffer->childs.size() >= 2))
			{
				QWidget *pHolder, *pWidget = (QWidget *)pArg;
				pHolder = pBuffer->childs[1];
				pWidget->reparent(pHolder, 0, QPoint(0,0), true);
				pHolder->setMaximumWidth(pWidget->sizeHint().width());
				pWidget->show();
			}
		}
		else if (!strcmp(sCommand+3, "set.buffer.cursor.position"))
		{
			if (pBuffer == NULL) return NULL;

			/* checking the coordinates */
			int iX = atoi(sArg1), iY = atoi(sArg2);
			if ( (iX >= 0) &&
			     (iY >= 0) && (iY < (int)(pBuffer->text->lines.size())))
			{
				/* go to the column/line */
				EditLine *pLine = pBuffer->text->lines[iY];
				if (pLine != NULL)
				{
					iX = pLine->nontabbedPosition(iX, pBuffer->tabsize);
					pBuffer->text->cursor.setX( iX );
					pBuffer->text->cursor.setY( iY );
					pBuffer->cmd_scrollToCursor();
					pBuffer->text->repaint(false);
				}
			}
		}
		else if (!strcmp(sCommand+3, "do.buffer.find.cursor"))
		{
			/* checking the coordinates */
			int iX = atoi(sArg1), iY = atoi(sArg2);
			if ( (iX >= 0) &&
			     (iY >= 0) && (iY < (int)(pBuffer->text->lines.size())))
			{
				/* go to the column/line */
				EditLine *pLine = pBuffer->text->lines[iY];
				if (pLine != NULL)
				{
					iX = pLine->nontabbedPosition(iX, pBuffer->tabsize);
					pBuffer->text->cursor.setX( iX );
					pBuffer->text->cursor.setY( iY );
					pBuffer->cmd_scrollTo(iY, 35, true);
					pBuffer->text->repaint(false);
				}
			}
		}
		else if (!strcmp(sCommand+3, "get.editor.havescrollbar"))
		{
			pReturn = (pBuffer->childs.size()>2?pBuffer:NULL);
		}
		else if (!strcmp(sCommand+3, "get.buffer.havescrollbar"))
		{
			pReturn = (pBuffer->childs.size()>2?pBuffer:NULL);
		}
		else if (!strcmp(sCommand+3, "get.buffer.getscrollbar"))
		{
			if (pBuffer->childs.size()>2)
				pReturn = pBuffer->vscroll;
		}
		else if (!strcmp(sCommand+3, "get.buffer.text.height"))
		{
			pReturn = (void *)pBuffer->size().height();
		}
		else if (!strcmp(sCommand+3, "get.buffer.visiblelines.count"))
		{
			pReturn = (void *)pBuffer->text->nLines;
		}
		else if (!strcmp(sCommand+3, "get.properties.color"))
		{
			pReturn = (void *)pBuffer->text->properties->getColorValue(
			          sArg1, (QColor *)sArg2);
		}
		else if (!strcmp(sCommand+3, "get.properties.string"))
		{
			pReturn = (void *)pBuffer->text->properties->getValue(
			          sArg1, (QString *)sArg2);
		}
		else if (!strcmp(sCommand+3, "set.properties.string"))
		{
			pReturn = (void *)pBuffer->text->properties->setValue(
			          sArg1, (QString *)sArg2);
		}
		else if (!strcmp(sCommand+3, "get.properties.number"))
		{
			pReturn = (void *)pBuffer->text->properties->getNumericalValue(
			          sArg1, (int)sArg2);
		}
		else if (!strcmp(sCommand+3, "get.buffer.lines.count"))
		{
			if (pBuffer && pBuffer->text)
				pReturn = (void *)pBuffer->text->lines.size();
		}
		else if (!strcmp(sCommand+3, "get.buffer.filepath"))
		{
			if (pBuffer) pReturn = (void *)pBuffer->getFile();
		}
		else if (!strcmp(sCommand+3, "show.editor.filebypath"))
		{
			bool bFound = false;
			if (pSession == NULL) return NULL;
			QFileInfo fi( NULL );
			QString realfilepath(sArg1);

#if QT_VERSION >= 280
			for(int i = 0; i < pSession->count(); i++)
			{
				EditBuffer *pIBuf = (EditBuffer *)pSession->page( i );
				fi.setFile( *pIBuf->getFile() );
				if (fi.absFilePath() == realfilepath)
				{
					pSession->setCurrentPage( i );
					bFound = true;
					break;
				}
			}
#else
			for(int i = 0; i < (int)pSession->oBufferList.size(); i++)
			{
				EditBuffer *pIBuf = (EditBuffer *)pSession->oBufferList.at( i );
				fi.setFile( *pIBuf->getFile() );
				if (fi.absFilePath() == realfilepath)
				{
					pSession->showPage( pSession->oBufferList.at( i ) );
					bFound = true;
					break;
				}
			}
#endif

			return (bFound ? Editor::editor : NULL);
		}
		/* set the current state of the buffer - we are selecting a block or not */
		else if (!strcmp(sCommand+3, "set.buffer.selectmode"))
		{
			pBuffer->text->setSelecting(!strcmp(sArg1, "true") || !strcmp(sArg1, "1"));
		}
		else if (!strcmp(sCommand+3, "get.editor.downcontainer"))
		{
			return Editor::editor->getContainer();
		}
		else if (!strcmp(sCommand+3, "editor.openfile"))
		{
				QStringList files( sArg1 );
				Editor::editor->openFile( &files );
		}
		else if (!strcmp(sCommand+3, "get.buffer.filename"))
		{
			if (pBuffer != NULL)
				pReturn = (void *)pBuffer->name;
		}
		else if (!strcmp(sCommand+3, "buffer.currentword"))
		{
			if (pBuffer != NULL)
				return pBuffer->command(sCommand+10, NULL, NULL, NULL);
		}
		else if (!strcmp(sCommand+3, "get.buffer"))
		{
			return pBuffer;
		}
		else if (!strcmp(sCommand+3, "module.find"))
		{
			for(unsigned int i = 0; i < Editor::editor->moduleloader->modules.size(); i++)
				if (*(Editor::editor->moduleloader->modules[i]->qsName) == sArg1)
					return  Editor::editor->moduleloader->modules[i];
			return NULL;
		}
		else if (strstr(sCommand+3, "module.command.") == sCommand+3)
		{
			bool bFound = false;
			unsigned int iModule;
			vector<Module *, allocator<Module *> > *pModList =
				&(Editor::editor->moduleloader->modules);
			for(iModule = 0; iModule < pModList->size(); iModule++)
				if (*( (*pModList)[iModule]->qsName) == sCommand+18)
					{ bFound = true; break; }
			QString *qsArg1, *qsArg2;
			qsArg1 = (sArg2 != NULL ? new QString(sArg2) : NULL);
			qsArg2 = (pArg != NULL ? new QString((char *)pArg) : NULL);
			if (bFound) (*pModList)[iModule]->pFunction( (*pModList)[iModule],
			       sArg1, qsArg1, qsArg2, -1, -1);
			if (qsArg1) delete(qsArg1);
			if (qsArg2) delete(qsArg2);
		}
	}

	return pReturn;
}

